/*
 *   (C) Copyright ECMWF http://www.ecmwf.int
 *   This software is licensed under the terms of the Apache Licence Version 2.0
 *
 *   Products are expected to be verified before instantiating this widget
 *   and all required properties are assumed be present.
 *
 *   emitted signals:
 *     - chart-product-availability
 *         the chart is expected to emit chart-product-availability under these conditions:
 *           1- product swap. (can be optimised (may not always be needed))
 *           2- instance availability changes (new cycle).
 *           3- base_time is changed.
 *     - product-swapped
 *     - product-swapped-and-loaded
 *     - chart-content-update
 *     - size-changed
 *     - chart-time-select
 *
 *   promises:
 *     internal:
 *       (a new product_swap promise is created for each product swap)
 *       - product_swap.product_swap
 *       - product_swap.plot
 *
 */
define(['config', 'jquery', 'bus', 'ajax_queue', 'browser', 'utils', './model/product', './epsgram', './tcyclone', 'chart/animator', 'chart/spinner', './basic_chart.css', 'tooltip/tipsy'], function(ECMWF, $, Bus, Ajaxqueue, Browser, Utils, BasicInstance, Epsgram, Tcyclone, Animator, Spinner) {

$.widget('ui.basicChart', {
	options: {
		instance: [],
		params: {},
		plugins: [],               // mixin items should be objects with the properties: mixin, options.
		active: true,
		initially_hidden: false
	},

	_init: function(){
		var self = this, o = self.options;
		o.instance = Array.isArray(o.instance) ? o.instance[0] : o.instance; // passing as an array prevents jquery ui cloning the instance.

		self.bus = o.messageBus = self.element.messageBus();
		self.on = {};

		Bus().send('event-start', ['basic chart load']);

		self.element.addClass('basic chart');

		self.container = self.element;
		self.img_wrapper = self.element.find('.www-chart-image-wrapper')
			.orelse(function(){ return self.element; }); // this fallback is a temporary fix for non-www pages.
		self.clip = self.element.div('.animation-frames'); // needed by Animator

		if(o.initially_hidden && o.instance.data.img && o.instance.data.img.width < 400){ // 400 is for case where dashboard thumbnail is not completed updating following previous preview close
			self._original_image = o.instance.data.img.url;
		}

		(function sizing() {
			var viewport = {};

			// ensure that resizing is not done before the time_navigator is loaded.
			self._loader = new (function loader(){
				this.checks   = { tnav: new $.Deferred(), img_load: new $.Deferred() };
				this.deferred = $.when
					.apply(self, [this.checks.img_load, this.checks.tnav])
					.done(set_tnav_size);
			})();

			self.img = self.element
				.find('img')
				.removeAttr('width').removeAttr('height')
				.orelse(function(){ self._image_added = true; return self.element.img()[o.initially_hidden ? 'hide' : 'show'](); })
				.css({width: 'auto', height: 'auto'})
				.on('load', function set_size(){
					self.img.css({opacity: 1.0});
					var padding = 0;

					if(!self.product.data.img){
						// TODO check that the image that has loaded is for the current product.
						return console.warn('product has no image');
					}

					var size = self.product.data.img.width
						? {width: self.product.data.img.width, height: self.product.data.img.height}
						: {width: parseInt(self.img.width(), 10) + 2 * padding, height: parseInt(self.img.height(), 10)};

					var _viewport = {}; // basic images cannot be assumed to be the same size.
					size.width && (_viewport.width = size.width);
					size.height && (_viewport.height = size.height);

					self._set_container_size(size.height);

					self._loader.checks.img_load.resolve(_viewport);

					((!self.element.is(':visible')) || (size.width === viewport.width && size.width === viewport.height)) || self.bus.send('size-changed', size);

					viewport = size;
				});

			self._set_container_size(self.img.height()); // for initial image when unchanged.
		})();

		/* -- disabled because it is degrading page performance
		require(['menu-pane', 'messages'], function() {
			self.element.div().insertBefore(self.img_wrapper).messages({messageBus: self.element.messageBus(), enable_popup: true});
		});
		*/

		// plugins are added before the product is set. To get product details, plugins should subscribe to product-swapped.
		self.plugins = $.map(o.plugins, function(mixin) {
			return mixin.mixin
				? new mixin.mixin(self, mixin.options || {})
				: new mixin(self, {});
		});

		self._connect = self.__connect;
		o.active && self._set_product();

		// set up a message that is sent each time the product is swapped, but not until the plot is available
		self.bus.listen('product-swapped', function() {
			$.Deferred(function(dfd){
				self.instance._loaded && dfd.resolve() ||
					self.bus
						.clear('chart-content-update')
						.listen_once(this, 'chart-content-update', function(_, msg) {
							dfd.resolve(msg);
						});
			}).then(function(msg) {
				self.bus.send('product-swapped-and-loaded', msg);
			});
		});

		var width;
		function set_tnav_size(size){
			// TODO finish moving this to basic_chart/time_navigator (only done here now for initialisation)
			var _width;
			if(self.time_navigator && (_width = size.width) !== width){
				self.time_navigator.trigger('layout', {width: width = _width});
			}
		}

		this.spinner = Spinner(this.element)
	},

	_set_product: function() {
		var self = this, o = self.options;

		var product_count = self.product_count = self.product_count ? ++self.product_count : 1;

		// invalidate the instance image if it is obsolete
		// -note that we must only freeze if we are absolutely sure that a request
		//  will be made otherwise the product will not be displayed.

		/* this compares the new products values with the latest request - it does not work if the product values have already been set to the new values!
		$.map(o.instance.variables, function(variable, name){
			if(self.instance && self.instance.__prev_request && self.instance.__prev_request.values[name] &&
				(JSON.stringify(self.instance.__prev_request.values[name]) !== JSON.stringify(variable.for_json())))
					return name;
		}).length && (console.log('_set_product: freezing...'), o.instance.on.image.value = null);
		*/

		// for this 'pending' test to work, it is required that any parameter changes must have been set previously.
		$.isEmptyObject(o.instance.pending) || (console.log('_set_product: delaying product display due to pending changes...'), o.instance.on.image.value = null);

		// time must be set imediately otherwise can be out of sync with instance.availability -- check this.
		if(o.instance.load){
			self.time = o.instance.variables.time && o.instance.variables.time.value; // TODO remove chart.time, just use product.time
		}

		self.instance && (
			self.instance.events.unbind('parameters-changed'),
			$.each(self.instance.variables, function(i, observable) {
				observable.unsubscribe(self);
			}),
			self._disconnect()
		);

		self.instance = o.instance.load ? o.instance : new BasicInstance.get_instance(o.instance);

		// ensure that chart-content-changed is always emitted, and is _after_ product-swapped
		// The timeout is for the case where the product is unchanged, however if it is set long
		// enough to allow for slow plots, it is too slow for the user.
		if(self.on.product_swap){
			if(self.on.product_swap.promise.state() !== 'resolved') self.on.product_swap.cancelled = true;
		}
		self.on.product_swap = when_with_timeout.call(self,
			20000,
			['product_swap', 'plot'],
			function() {
				self.bus.send('chart-content-update', {product: self.instance});
			},
			function() {
				console.warn('timeout waiting for product load');
			}
		);

		self.instance.is_from_cache && self.on.product_swap.plot.resolve();

		self.product = self.instance                                 // needed by Animator
		self.name = self.instance.data.name;                         // used by Animator for cancellations.

		if(self.instance.has_widget('epsgram')){
			$.extend(self, new Epsgram());
			self.epsgram_init();
		}else{
			self.epsgram_destroy && self.epsgram_destroy();
		}

		$.each({tcyclone: 'Tcyclone'}, function(plugin, N) {
			if(self.instance.data.product.package === 'tcyclone'){
				if(!self.tcyclone_init){
					$.extend(self, new Tcyclone());
					self.tcyclone_init();
				}
			}else{
				if(self.tcyclone_init){
					// TODO remove tcyclone plugin
					true;
				}
			}
		});

		// variables.time is created in set_parameters, so if exists must have already been called.
		//self.instance.variables.time || self._set_parameters(o.params); // initiate plot request for new products. old products are assumed to have had the correct parameters already set.

		// TODO check that instance.set_parameters does nothing if the parameters are the same
		//if($.isEmptyObject(self.instance.variables && !o.instance.pending){
		self._set_parameters(o.params);
		//}

		(product_count === 1) && self.instance.have.availability.done(function() {
			self
				._listen_once('tool-animation-toggle', function() {
					self.animation = new Animator.Controller(self, true).animation();
					self.animation.toggle();
				})
				._listen('chart-animation-stop', function() {
					// during animation, the chart time is different from the instance time.
					// when the animation stops, the instance needs to be resynced to the chart.

					//console.log('basiChart.on_stop: chart.time=', [chart.time.base_time.format('mmdd:HH'), chart.time.step, chart.time.valid_time.format('mmdd:HH')]);
					//console.log('basiChart.on_stop: instance.time=', [chart.instance.time.base_time.format('mmdd:HH'), chart.instance.time.step, chart.instance.time.valid_time.format('mmdd:HH')]);
					self.instance.variables.time &&
						(self.time.key !== self.instance.variables.time.value.key) &&
						self.instance.variables.time.set(self.time);
				});
		});

		self._min_height = 0;

		(product_count === 1) && self._step_change_listen([
			['prev-step', 'previous'],
			['next-step', 'next'],
			['first-step', 'first'],
			['last-step', 'last'],
			['next-base-time', 'nextBaseTime'],
			['prev-base-time', 'previousBaseTime']
		]);

		self.instance.events
			.bind('parameters-changed', function(changed){
				if(self.element.is(':visible')){ // test not needed now the event is unbound
					console.log('basicChart: parameters-changed', changed);
					self._clear();
				}
				if(changed.area){
					self._min_height = 0;
				}
			});

		self.instance.variables.time.subscribe(self, function(){
			console.log('--> basicChart: got time change');
			self.time = this.value;
		});

		self._connect();

		// the availability may not have actually changed on a product swap, but as we dont keep the old product, we cannot compare.
		// *** for new instances, this will mean chart-product-availability is emitted twice ?
		self.instance.have.availability.done(function(){

			self.time = self.instance.variables.time.value; // TODO remove chart.time, just use product.time

			self.instance.availability(self.time, function(avail){
				self.bus.send('chart-product-availability', avail); // sending specially formatted availability provided by the base class
			});
		});

		self.bus.send('product-swapped', {}); // dummy value needed to trigger use of bus Past feature.
		self.on.product_swap.product_swap.resolve();
	},

	_create: function(){
	},

	_destroy: function(){
		this.options.messageBus.forget_me(this);
		Bus().forget_me(this);
		this._disconnect();
		this.img.hide();
		this._image_added && this.img.remove();
		this.animator && this.animator().stop_listening();

		$.each(this.plugins, function(i, plugin) {
			plugin.destroy && plugin.destroy();
		});
	},

	_render: function(url) {
		var self = this;

		(url === self._original_image) || self.img.show();

		if(url === self.img.attr('src')){
			self.first || console.log('basicChart._render: image unchanged');
			self.img.css({opacity: 1.0});
		}else{
			self.img.attr({src: url || Utils.blank_gif, name: self.instance.name()});
		}
	},

	/*
	 *   args.instance should be a fully hydrated basic chart product, or a collection object
	 */
	setProduct: function(args) {
		this.instance && this.instance.cancel();

		if(args.instance.has_widget && !args.instance.load){
			args.instance = BasicInstance.get_instance({product: args.instance.data})
		}

		if(__DEV__ && this.instance){
			this.instance.validate([
				['has instance', function(){ return args.instance; }],
				['is hydrated', function(){ return args.instance.load; }],
			]);
		}

		this.options.instance = args.instance;
		this._set_product();
		return this;
	},

	blank: function() {
		this.img.attr('src', Utils.blank_gif);
	},

	grey: function() {
		this.img.css({opacity: 0.3});
	},

	__connect: function() {
		var self = this, product = self.instance;

		self.instance.base_time.subscribe(self, function(){
			console.log('--> basicChart: got basetime change');
			self._clear();
			self.time = self.instance.variables.time.value; // FIXME this is also done in the main time handler, but unfortunately this handler is called first.

			// although the availability has not changed, specially formatted emitted object depends on the base_time
			self.instance.availability(self.time, function(avail){
				self.bus.send('chart-product-availability', avail); // sending specially formatted availability provided by the base class
			});
		});

		self.instance.valid_time.subscribe(self, function() {
			var param = this;
			self.instance.have.availability.done(function(){
				self.bus.send('chart-time-select', param.value);
			});
		});

		self.instance.on.availability.subscribe(self.uuid, function(availability){
			self.time = self.instance.variables.time.value; // FIXME this is also done in the main time handler, but unfortunately this handler is called first.

			console.log('--> basicChart: got availability change');
			self.instance.availability(self.time, function(avail){
				self.bus.send('chart-product-availability', avail); // sending specially formatted availability provided by the base class
			});
		});

		self.instance.on.image.subscribe(self.uuid, function(url){
			// if the image has a value, we get a callback imediately (before load).
			// note that size is not yet available. must wait until img.load

			self._render(url);

			self._loader.checks.img_load.done(function(size) {
				if(!self.on.product_swap || self.on.product_swap.plot.state() === 'resolved'){
					self.bus.send('chart-content-update', {product: self.instance});
				}
			});

			self.first = false;
		});

		product.on.load_start.subscribe(self.uuid, function(){
			self.bus.send('log-clear');
			self._start_spinner();
		});

		product.on.load_end.subscribe(self.uuid, function(){
			false && self.instance.error_msg && self.bus.send('log-error', {title: 'Plot failed', message: self.instance.error_msg});

			self.on.product_swap.plot.resolve();
		});

		self._connect = $.noop;
	},

	_disconnect: function() {
		var self = this;

		$.each(['base_time', 'valid_time'], function(i, name) {
			self.instance[name].unsubscribe(self);
		});

		$.each(['load_start', 'load_end', 'availability', 'image'], function(i, name) {
			self.instance.on[name].unsubscribe(self.uuid);
		});

		self._connect = self.__connect;
	},

	/*
	 *  A chart can be 'deactivated' if it is not visible and does not need to respond to product changes.
	 */
	deactivate: function() {
		this.options.active = false;
		this._stop_spinner()._disconnect();
	},

	/*
	 *  Following deactivation, activate will update the display to be in sync with the product.
	 */
	activate: function() {
		this.options.active = true;
		this.product_count || this._set_product();
		// TODO make sure connecting is noop if already connected
		this._connect();
	},

	_set_container_size: function(height) {
		this.img_wrapper.css({height: this._min_height = Math.max(this._min_height, height || this.img.height())}); // needed where chart img is position:absolute.
	},

	_show_frame: function(img) {
		var self = this;

		this.frame_prev = this.frame_prev || this.img;
		this.frame_prev.hide();
		this.frame_prev = img.show();

		self.title && self.title.html(self.product.title(self.time));
	},

	showFrame: function(time, img) {
		var self = this, o = self.options;

		self.time = time;
		self._show_frame(img);
		o.messageBus.send('chart-time-select', self.time);
	},

	_clear: function () {
		this.animation && this.animation.clear();
		return this;
	},

	_set_parameters_done: function(error){
		error && console.warn('basicChart._set_parameters_done: error: ', error);
	},

	_set_parameter: function(name, value){
		var self = this;

		self._clear();
		self.instance.set_parameter(name, value, self._bind(self._set_parameters_done));

		return this;
	},

	_set_parameters: function(parameters){
		var self = this;

		self._clear();
		self.instance.set_parameters(parameters, self._bind(self._set_parameters_done));

		return this;
	},

	set_date: function(time){
		this._set_date(time);
	},

	_set_date: function(time){
		if(time){
			this._clear();
			this.instance.variables.time.set(time);
		}
		return this;
	},

	_start_spinner: function(){
		var self = this;

		this.product.on.load_end.once(function(){
			self._stop_spinner();
		});

		this.spinner('queued')
	},

	_stop_spinner: function(){
		this.spinner('complete')

		return this;
	},

	_print: function(options, args) {
		var self = this;

		self.on.product_swap.promise.done(function(size) {

			self.product.print($.extend({
					json: self.product.as_json(size, null, self.mail, 'pdf'), // TODO arg2 (time)
					success: function(result){
						console.log('print: success.', result, self.time);
						if(result){
							if(result.result && result.result.url){
								location.href = result.result.url;
							}

							// TEMPORARY
							// currently print-chart does not support basic charts and will return a 'redirect'.
							else if(result.redirect){
								Ajaxqueue($.extend({
										url: result.redirect,
										priority: 'very-high',
										json: self.product.as_json(size, null, self.mail, 'pdf'),
										success: function(result) {
											if(result.result.url){
												location.href = result.result.url;
											}
										}
									},
									args
								));
							}
						}
					}
				},
				args)
			);
		});
	},

	_listen: function (name, fn) {
		var self = this, o = self.options;

		o.messageBus.listen(self, name, fn);
		return self;
	},

	_listen_once: function (name, fn) {
		this.options.messageBus.listen_once(this, name, fn);
		return this;
	},

	// global listener, eg for keyboard shortcuts.
	_step_change_listen: function (items) {
		var self = this;

		var bus = Bus();
		$.each(items, function(i, item){
			bus.listen(this, items[i][0], function(){
				self.element.parents('#main').length && self.time && self.set_date(self.time[items[i][1]](self.instance.variables.time.availability));
			});
		});
		return this;
	},

	_bind: function (fn) {
		var self = this;
		return function(){ return fn.apply(self, arguments); };
	},

	/*
	 *  Instantiate a plugin on a running chart instance
	 */
	add_plugin: function(name, plugin) {
		if(!$.findWhere(this.plugins, {name: name})){
			var plugin = new chart_plugins[name](this);
			plugin && (plugin.name = name, this.plugins.push(plugin));
			return true;
		}
		return false;
	}
});

var chart_plugins = {};
$.chartRegisterPlugin = function(name, plugin){
	chart_plugins[name] = plugin;
};

function fade(Spinner){
	var self = Spinner;
	var stop = Spinner.stop;
	var spin = Spinner.spin;
	Spinner.spin = function(){
		var a = spin.apply(self, arguments);
		$(this.el).css({opacity: 0.5});
		return a;
	}
	Spinner.stop = function(){
		$(self.el).css({opacity: 0}).delayed(1000, function(){
			stop.apply(self, arguments);
		});
	}
	return Spinner;
}

$.widget('ui.locationMap', {
	options: {
	},

	_init: function(){
		this.element.div().text('locationMap');
	}
});

/*
 *  Return a new promise with an array of sub-promises. The promise will resolve
 *  following either the completion of the subpromises or on a timeout.
 */
function when_with_timeout(timeout, promise_names, done, on_error) {
	var self = this;

	var _promises = $.map(promise_names, function(i, v){ return $.Deferred(); });

	var promises = {};
	for(var i in _promises){
		promises[promise_names[i]] = _promises[i];
	}

	var promise = $.Deferred(function(df) {
		$.when.apply(self, _promises).done(df.resolve);
	}).done(done);

	setTimeout(function() {
		if(!promises.cancelled){
			if(promise.state() !== 'resolved') on_error && on_error();
			promise.resolve();
		}
	}, timeout);

	promises.promise = promise;
	return promises;
} 

});
