
/*
 *  Subclass of the model object used by the ECMWF Chart widget.
 *
 *  This module exports 1) the BasicInstance type, and 2) a factory method.
 *
 *  Events:
 *    parameters-changed - the selected value of one or more parameters has changed
 *
 *  Observables:
 *    parameters: to subscribe to changes to an individual parameter, for example: my_product.variables.time.subscribe(fn)
 *    on.image
 *    on.load_start
 *    on.load_end
 *
 *  Availability:
 *    to reliably get the availability, use the have.availability promise.
 */
define(['config', 'jquery', 'utils', 'ajax_queue', 'time', 'chart/model/base_product', 'emitter', 'observable', 'basic_chart/mixins', 'chart/model/parameters'], function(ECMWF, $, Utils, Ajaxqueue, Time, BaseProduct, Emitter, Observable, mixins, Parameters){

	function BasicInstance(args) {
		// parameter values: parameters can only be considered 'provisional' until the authoritative information is returned from get_plot.

		// products are identified by their name (args.data.name)
		// -if args.data.product exists, this is the product that the current product is derived _from_.
		//  A complication arises because the backend doesnt know the location of the derived products.

		__DEV__ && $.each(args, function(i, v){ $.isFunction(v) && console.error('Attempt to create BasicInstance from another object ', i) });

		$.extend(this, mixins);

		this.options = {
			values: ECMWF.remove(args, 'values') || {},
			priority: ECMWF.remove(args, 'priority')
		}

		this.data = $.extend({product: {}}, args);
		delete this.data.values;

		this.data.package && (this.data.product.package = this.data.package, delete this.data.package);
		if(!this.is_instance()){
			if(this.data.name){
				if(this.data.name.indexOf('uid') === 0) console.warn('BasicInstance: unexpected name format');
			}
		}
		this.uuid = ++ECMWF.uuid;
		this.request_id = 0;
		this._changed = {}, // _changed is similar to pending, but pending is not reset until _after_ the load is complete.
		this.pending = {},
		this.variables = {}; // this must _always_ be itererable
		this.plots = {}; // plots keyed by width (or 0 for default plot)
		this.events = new Emitter();
		this.on = {
			image: new Observable({state: true}),
			load_start: new Observable(),
			load_end: new Observable()
		};
		this.have = {
			availability: new $.Deferred()
		};
		this.url = ECMWF.options.ajax.plots.get_map; // TODO now there is only 1 url this can be removed
	}

	var _prototype = new BaseProduct();
	BasicInstance.prototype = $.extend({}, _prototype, {
			_prototype: _prototype, // give access to overridden methods of the base class.

			/*
			 *  set_parameters is used by clients to set the complete state of the instance.
			 *  It _must_ be called, even if the instance is only using default values. // FIXME
			 *  To use the default values, the parameters argument can be unset or an empty object.
			 *  Setting the state is not done at instantiation time in order to give clients more control over when the changed signals are sent.
			 *
			 *  Note that it is allowable for parameters specified here to be invalid for this product.
			 */
			set_parameters: function(parameters, done){
				var self = this;
				done = done || $.noop;
				parameters = parameters || {};

				var _changed = false;
				//if($.isEmptyObject(parameters) && (!self.variables.time)) _changed = true; // TODO check this is what we want.

				// the time variable is unique in that it is not in the variables list.
				// we dont want to set it before we have the availability, but we need it to make the first plot request.
				if(!self.variables.time){
					_changed = true; // always emit the first time

					self.variables.time = new Parameters.time(parameters.time);

					self.variables.time.subscribe(function(time){
						self.changed({time: true}/*, done*/);
					});

					self.valid_time = self.variables.time.filter(function(){
						var param = this;
						return !param.prev || (param.value.valid_time.getTime() !== param.prev.valid_time.getTime());
					});

					self.base_time = self.variables.time.filter(function(){
						var param = this;
						if(!param.prev || (param.value.base_time.getTime() !== param.prev.base_time.getTime())) console.log('BASETIME CHANGE', [param.value.base_time.format('mmdd:HH'), param.value.step, param.value.valid_time.format('mmdd:HH')]);
						return !param.prev || (param.value.base_time.getTime() !== param.prev.base_time.getTime());
					});

					self.on.availability = self.variables.time
						.filter(function(){
							return this.availability_changed;
						})
						.map(function(value){
							return self.variables.time.availability;
						});
				}

				$.each(parameters, function(i, v) {
					if(self.variables[i]){
						if(v !== self.variables[i].to_string()){
							// TODO check this condition is correct
							self.variables[i].set(v) && (_changed = true);
						}
					}
				});

				if(self.options.values){
					$.extend(self.options.values, parameters);
				}

				if(_changed || self._failed){

					var vals = {};
					$.each(self.variables, function(i, v){
						(typeof v.value === 'undefined') || (vals[i] = true);
					});
					self.changed(vals, done);
				}else{
					console.log('basicInstance.set_parameters: unchanged');
					done();
				}

				return this;
			},

			set_size: function(viewport) {
				this.data.viewport = viewport;
				return this.changed({size: true});
			},

			load: function(done) {
				var self = this, o = self.options;
				done = done || $.noop;

				if(!self.data) return done();

				if(self._loaded) return done(self.data.img.url);

				if(__DEV__) (self.data.product && self.data.product.package) || console.log('BasicProduct.load: package not specified (default implied)', self.data);

				var json = self.as_json();

				if(!compare_requests(json.product, self.__prev_request)){
					self.pending = {};
					return console.log('%cBasicProduct.load: ignoring duplicate request', 'color:#c63'), done(self.data.img && self.data.img.url);
				}
				self.__prev_request = json.product;

				var request_id = ++self.request_id;
				self._failed = false;

				self.on.load_start.emit();

				Ajaxqueue($.extend({}, self._req_defaults, {
					json: json,
					url: self.url,
					cache: 120000,
					priority: o.priority,
					tag: self.uuid,
					complete: function(result){
						console.log('%cBasicProduct.load: result=', 'color:#39f', $.extend(true, {}, result));
						self.error_msg = null;

						if(result.exception){
							self.error_msg = result.status +
								(result.error
									? '<br/>' + new DjangoBacktrace(result.error).concise()
									: '');
							return failed();
						}

						if(request_id !== self.request_id) return;

						self.pending = {};

						if(!self.validate_plot(result, json)) return failed();

						self._loaded = true;

						var img = self.data.img = result.result;

						// record value substitutions to prevent duplicate requests.
						self.__prev_request.values = $.extend({}, result.values);

						if(result.availability && !self.variables.time){
							self.variables.time = new Parameters.time(new Time(result.product.time));
							if(ECMWF.query){
								var q = ECMWF.query();
								q.time && self.variables.time.set(q.time);
							}
						}

						if(result.availability){
							var avail = $.map(result.availability.availability, function (a) { return new Time(a); });
							self._availability = avail; // Do not use this (it is however needed by the base Product class).
							result.availability.format && (self.variables.time.format = result.availability.format);

							self.variables.time.set(new Time(result.availability.closest), avail);

							self.__prev_request.values.time = self.variables.time.for_json();
						}

						result.axis && $.each(result.axis, function(i, v){
							if(!self.variables[v.name]){
								if(result.values[v.name]) v.value = result.values[v.name];
								self.variables[v.name] = new Parameters[(v.type === 'dynamic') ? 'dynamic' : 'string'](v)
									.subscribe(function(param) { // TODO unsubscribe
										console.log('--> ' + this.name + '=', param);
										var val = {};
										val[this.name] = true;
										self.changed(val);
									});
							}
						});

						(self.data.metadata = result.result.magics.metadata) && (
							delete result.result.magics.metadata
						);

						delete self.options.values; // provisional values no longer needed.
						self.data.is_fall_back && (self.data.is_fall_back = result.is_fall_back);

						self.plots[json.product.viewport ? self.data.img.width : 0] = self.data.img;

						self.on.image.set(self.data.img.url);
						self.has_availability() && self.have.availability.resolve();
						self.on.load_end.emit();
						done(img.url);
					}
				}));

				function failed() {
					self._failed = true;
					self.on.load_end.emit();
					return done();
				}
			},

			has_availability: function(){
				var p = this.variables.time;
				return p && p.availability && p.availability.length;
			},

			/*
			 *  Internal. Used by parent class.
			 */
			get_chart: function(req){
				// TODO Eccharts appears to already have an instance name, eg "map_1".
				//      -perhaps the chart widget should use the existing instance name for web2013 products.
				req = $.extend({}, req);
				delete req.name;

				Ajaxqueue($.extend(true,
					{
						url: ECMWF.options.ajax.plots.get_instance_plot,
						json: $.extend({}, defaults, this.request)
					},
					req
				));
			},

			cancel: function() {
				this._changed = {};
				this.pending = {};
				Ajaxqueue.cancelByTag(this.uuid);
			},

			save: function(done){
				var self = this;

				// item can be either a product, or an existing instance
				var is_new = (!this.data.kind) || (this.data.kind !== 'instance');

				var json = $.extend({}, this.data);

				json.package = json.product.package;
				if(is_new){
					json.package || (json.package = 'medium');
					json.name = json.product.name; // TODO change backend so this is not needed.
				}

				// keep request size down. jsonp has a hard limit.
				delete json.variables;
				delete json.axis;
				delete json.product;
				delete json.widgets;
				delete json.batch;
				delete json.availability;

				json.values && (json.values = Utils.map(json.values, function(v){ return v.replace('=', '%3D'); }));

				$.extend(json, {size: this.size});

				Ajaxqueue({
					url: is_new ? ECMWF.options.ajax.add_instance : ECMWF.options.ajax.update_instance,
					json: json,
					success: function(result){
						var error = !result
							? 'no result'
							: result.error
								? result.error
								: (!result.instance || !result.instance.name)
									? 'no instance'
									: null;
						console.log('basicProduct.save.complete: instance=', result ? result.instance : null);
						self.data.name = result.instance.name;
						done && done(result.instance, error);
					},
					error: function(a, b){
						console.log('basicProduct.save.error:', a, b);
						done && done(null, {error: 'error (TODO)'})
					}
				});
			},

			/*
			 *  TODO decide which thumbnail type to use; either this one or the one used by the dashboard,
			 *       and remove the other.
			 *       Perhaps categorise as 'static' and 'dynamic'. Static thumbnails might be best considered
			 *       as part of the Package.
			 */
			thumbnail: function(done){
				if(typeof done !== 'function') done = arguments[1];

				var item = this.data;

				function _fail(){
					done && done(null, 'fail');
				}

				item ? Ajaxqueue({
					url        : '/backend/product-thumbnail/' + item.product.name + '/',
					data       : {},
					log_errors : 'Get Basic thumbail (' +  item.title + ')',
					success: function(result){
						done && done(result ? result.url : null, result ? null : 'No result');
					},
					error: _fail
				}) : _fail();
			},

			as_json: function (viewport, time, mail, format) {

				var values = this.options.values
					? this.options.values // first time only
					: Utils.map(this.variables, function(v){ return v.for_json(); });

				if(this.options.values && this.options.values.time){
					values.time = new Parameters.time(this.options.values.time).for_json();
				}

				if(time) values.time = new Parameters.time(time).for_json();

				return {
					product: $.extend({
						product: {
							name    : this.name(),
							package : this.data.product.package
						},
						values: values
					},
					(this.data.viewport ? {viewport: this.data.viewport} : {}),
					{
						viewport   : viewport,
						format     : format
					}
				)};
			},

			// TODO remove in favour of _is_new
			is_instance: function () {
				return this.data.kind && this.data.kind === 'instance';
			},

			name: function(){
				// TODO products should not pretend to be instances.
				// note that this fn has been changed to return the product name if there is no instance name.
				return this._is_new()
					? (this.data.product.name || this.data.product)
					: this.data.name;
			},

			title_: function(){
				return this.data.title || this.data.product.title || '';
			},

			keywords: function() {
				return this.data.product.keywords || '';
			},

			availability: function(time, ready) {
				var self = this;

				return this._prototype.availability.call(this, time, function(data) {
					ready($.extend(data, {format: self.variables.time.format}));
				});
			},

			canonical_url: function() {
				return ECMWF.fix_host.to_frontend(window.location.protocol + '//' + window.location.host)
					+ '/en/forecasts/charts/' + this.data.product.package + '/' + this.name() + '?' + $.map(this.variables, function(v) { return v.name + '=' + v.to_string(); }).join('&');
			},

			_is_new: function() {
				return (!this.data.kind) || (this.data.kind !== 'instance');
			},

			get_save_state: function() {
				var derived_name = this.name() + '_' +  Math.floor(Math.random()*10000000000);
				var state = this.as_json().product;
				return $.extend(true, state, {kind: 'basic_chart', name: derived_name});
			},

			add_to_dashboard: function(options, done) {
				console.log('basicProduct.add_to_dashboard');

				if(!this.in_dashboard){ // TODO
					$.extend(options, {
						title: this.title_()
					});

					return this._prototype.add_to_dashboard.apply(this, arguments);
				}
			},

			/*
			 *  Note that there is no check for 'name'. Plots do not have names, and the plotting calls do not understand instances so instance.name is non-sensical.
			 */
			validate_plot: function(result, request){
				var self = this;
				var img = result ? result.result : null;
				var product = result ? result.product : null;
				return self.validate([
					['has result', function(){ return result; }],
					['no backend error', function(){ result.error && (self.error_msg = result.error, console.warn('result.error=', result.error)); return !result.error && (typeof result.error !== 'function'); }],
					['has product', function(){ return product; }],
					['has product name', function(){ return product.name; }],
					['has product title', function(){ return product.title; }],
					['viewport ok', function(){ return product.viewport ? (product.viewport.width && product.viewport.height) : true; }],
					['has axes', function(){ return result.axis; }],
					['has time', function(){ return result.product.time && !$.isEmptyObject(result.product.time); }],
					['step is numeric', function(){ return typeof result.product.time.step === 'number'; }],
					['has image', function(){ return img && img.length; }],
					['image has size', function(){ if(img && (!img.width || !img.height)) console.warn('image size missing'); return true; }],
					['has correct size', function(){
						if(request && request.product && request.product.viewport){
							var viewport = request.product.viewport;
							if((viewport.height && viewport.height > 0 && Math.abs(img.height - viewport.height) > 1) || (viewport.width && viewport.width > 0 && Math.abs(img.width - viewport.width) > 1)) console.warn('plot is incorrect size');
						}
						return true;
					}],
					['has magics', function(){ return img.magics || (img.magics = {}); }], // some products do not have this property
					['consistent param value types', function(){
						// -----------------------------------------------------------
						// FIXME acceptance of bad data will likely cause code failure.
						// -----------------------------------------------------------
						$.map(result.axis, function(axis){
							if(axis.values){
								var val_type = typeof axis.values[0];
								if($.map(axis.values, function(v){
									if(typeof v !== val_type) return true;
								}).length) return true;
							}
						}).length && console.warn('bad data: inconsistent param value types');
						return true;
					}],
					['has availability', function(){
						//201510 _all_ products now have availability. It is an error otherwise.
						return result.availability && result.availability.availability && result.availability.availability.length;
						//return (!self.has_availability()) || (result.availability && result.availability.availability && result.availability.availability.length);
					}],
					['availability is valid', function(){
						if(result.availability && result.availability.availability && result.availability.availability.length){
							return !$.map(result.availability.availability, function(v){
								var t = new Time(v);
								var vt = new Date(t.base_time.getTime());
								vt.setTime(vt.getTime() + (v[1] * 60 * 60 * 1000));
								if(vt.getTime() !== t.valid_time.getTime()){
									console.error("!!!!");
									return false;
								}
							}).length;
						}
						return true;
					}],
					['value has variable', function(){
						// TODO this test only warns because the backend returns all requested values even if invalid for the current product.
						// -invalid parameters are currently removed. This can be changed once the backend is fixed.
						var variables = result.axis;

						var valid = {};
						if(variables) for(var i in result.values){
							$.findWhere(variables, {name: i}) || console.warn('bad data: missing variable for value \'' + i + '\'. variables=', $.map(variables, function(a){ return a.name; }));
							$.findWhere(variables, {name: i}) && (valid[i] = result.values[i]);
						}
						result.values = valid;
						return true;
					}],
					['no integer values', function(){
						// handling integers would require special attention. As there is no need to handle integers, they are converted to strings.
						// TODO move conversion to server.
						if(result.axis){
							for(var p in result.axis){
								var param = result.axis[p];
								if(typeof result.values[param.name] === 'number'){
									result.values[param.name] = result.values[param.name] + '';
									for(var v in param.values){
										param.values[v] = param.values[v] + '';
									}
									console.log('validate: found integer type! ', param.values[v], param);
								}
							}
						}
						return true;
					}],
					['relative archive date value is string', function(){
						return (result.values && result.values.relative_archive_date ) ? (typeof result.values.relative_archive_date === 'string') : true;
					}],
					['no BT in axes', function(){
						if(result.axis){
							if($.findWhere(result.axis, {name: 'base_time'})){
								console.warn('product error: axes should not contain an item for base_time');
								for(var i in result.axis) if(result.axis[i].name === 'base_time') result.axis.splice(i, 1);
								return false;
							}
						}
						return true;
					}],
					['no step in axes', function(){
						if(result.axis){
							if($.findWhere(result.axis, {name: 'step'})){
								console.warn('product error: axes should not contain an item for step');
								for(var i in result.axis) if(result.axis[i].name === 'step') result.axis.splice(i, 1);
								return false;
							}
						}
						return true;
					}],
					['forecast time', function(){
						if(!self.has_availability()) return true;

						var availability = result.availability.availability;
						var warned1 = 0;
						var warned2 = 0;
						return (!availability) || !$.map(availability, function(v, i){
							typeof v[0] === 'string' || (
								warned1++ || console.error('availability validation: first value must be string. Applying workaround...'),
								availability[i][0] = v[0] + ''
							);
							typeof v[2] === 'string' || (
								warned2++ || console.error('availability validation: third value must be string. Applying workaround...'),
								availability[i][2] = v[2] + ''
							);
						}).length;
					}],
					['is_fall_back', function(){
						// apparently this flag is unreliable except in the case of epsgrams
						if(!result.is_fall_back) delete result.is_fall_back;
						return true;
					}],
					['widgets valid', function(){
						if(result.widgets && result.widgets.time_navigator){
							return false;
						}
						return true;
					}],
					['layers', function(){
						if(product.layers){
							console.warn('bad data: product should not contain layers');
							delete product.layers;
						}
						return true;
					}]/*
// this test should be re-enabled once the products are fixed.
					['labels is object ', function(){
						var axes = result.axis;
						for(var i in axes){
							var param = axes[i];
							if(param.labels){
								if(Array.isArray(param.labels)) return false;
							}
						}
						return true;
					}]
					*/
				]);
			},

			// TODO this method causes a load() so needs to be renamed to avoid misunderstanding.
			changed: function(changed, done) {
				var self = this;
				done = done || $.noop;

				self._loaded = false; // TODO use $.isEmptyObject(pending) instead?
				if(!changed || !$.isEmptyObject(ECMWF.remove($.extend({}, changed), 'size'))) self.plots = {};
				self.on.image.unset();

				$.extend(this._changed, changed);
				$.extend(this.pending, changed);

				self._timer || (self._timer = setTimeout(function(){
					self.events.send('parameters-changed', self._changed);
					self._changed = {};
					self._timer = null;

					self.load(function(url, error){
						done(error);
					});
				}, 10));

				return this;
			}

		/*
		copy: function() {
			var self = this;

			var copy = $.extend(true, {}, this);
			$.each($.map(copy.variables, function(v, i) {return i}), function(i, v) {
				copy.variables[v] = self.variables[v].copy();
			});
			return copy;
		}
		*/
	});

	var defaults = {
		name: 'ecmwf_sample_product_snow_cover_ecchart',
		package: 'sample_product',
		axis: '{}'
	}

	var instances = [];

	// TODO move into ajaxqueue
	function DjangoBacktrace(error) {
		this.error = error.length
			? $.map(error, function(a){ if(a) return a; }) // remove empty entries
			: error;

		this.concise = function() {
			return this.error.length ? this.error[this.error.length - 1] : this.error;
		}
	}

		var validate_product = function(result){

			var instance = result.instance;
			var product = instance ? instance.product : null;

			return mixins.validate([
				['has instance', function(){ return instance; }],
				['has product', function(){ return product; }],
				['has widgets', function(){ return product.widgets; }],
				['axes is array', function(){ return Array.isArray(product.variables); }],
				['axes array not empty', function(){ return instance.product.variables.length; }],
				//['axis has name', function(){ for(var i in instance.product.variables) console.log('  axis=', instance.product.variables[i]); return true; }],
				['axis has name', function(){ for(var i in instance.product.variables) if(!instance.product.variables[i].name) return false; return true; }],
				['axis has title', function(){ for(var i in instance.product.variables) if(!instance.product.variables[i].title) return false; return true; }],
				['axis title is string', function(){ for(var i in instance.product.variables) if(typeof product.variables[i].title !== 'string') return false; return true; }]
			]);
		}

	function compare_requests(a, b) {
		// compare plain objects. object keys must be plain objects, arrays or strings

		if(!a || !b) return 1;

		for(var i in a){
			if(typeof a[i] === 'object'){
				if(compare_requests(a[i], b[i])) return 1;
			}
			else if(a[i] !== b[i]) return 1;
		}

		for(i in b){
			if(typeof b[i] === 'object'){
				if(compare_requests(b[i], a[i])) return 1;
			}
			else{
				if(b[i] !== a[i]) return 1;
			}
		}

		return false; // equal
	}

	return {
		Instance: BasicInstance,

		/*
		 *   Do not pass a collection object, instead use {product: object.data}
		 */
		get_instance: function(args) {

			// TODO args.name is used for Dashboard where we have derived products and in this case is the name of the derived product
			var name = args.name || args.product.name;
			var instance = $.findWhere(instances, {name: name});

			if(instance){
				instances.push(instances.splice($.inArray(instance, instances), 1)[0]); // re-order the array to reflect new access

				if(!$.isEmptyObject(instance.instance.pending)){
					__DEV__ && console.warn('FIXME cached product should not have pending changes');
					// this can happen with 2 of the same charts on the dashboard
					instance.instance.pending = {};
				}
				instance.instance.is_from_cache = true; // temporary - find a better test based on an existing property

				if(!instance.instance.on.image.value){ // handle failed products
					console.warn('cached instance has no image');
					instance = new BasicInstance(args);
					instances.push({name: name, instance: instance});
					return instance;
				}

				return instance.instance;
			}else{
				if(instances.length > 20) instances.shift();

				instance = new BasicInstance(args);
				instances.push({name: name, instance: instance});
			}
			return instance;
		},

		/*
		 * This has been modified to convert an instance into a product to instantiate a BasicInstance.
		 * TODO needs refactoring
		 */
		get_info: function(args, params, done) {
			// the args contains, at minimum, only the id.
			// Use this fn to merge back full product details.
			var self = this;

			if(args.variables){ done(args); return; }

			Ajaxqueue({
				url: ECMWF.options.ajax.get_instance,
				data: JSON.stringify({id: args.name}),
				success: function(result){
					if(!result.error){
						if(!validate_product(result))
							result.error = self.error_msg;
					}
					if(result.error){ done && done(result, result.error); return; }

					//if(!self._validate_instance(result)){ done(result, self.error_msg); return; }
					$.extend(result.instance.product.values, params);

					var instance = new BasicInstance({product: result.instance.product});
					done(instance, result.error);
				},
				error: function(a){
					console.log('get_instance.error:', arguments);
					// TODO dont return a message if there is no information. Should be left to the presentation layer.
					done(null, a.status === 'timeout' ? 'No response' : a.exception ? 'error: ' + a.exception : 'No error message was returned');
				}
			});
		}
	}
});

