
/*
 *  (C) 2015 Copyright ECWMF http://www.ecmwf.int
 *
 *  This software is licensed under the terms of the Apache Licence Version 2.0
 *  which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 *  In applying this licence, ECMWF does not waive the privileges and immunities
 *  granted to it by virtue of its status as an intergovernmental organisation nor
 *  does it submit to any jurisdiction.
 *
 *  (C) Visionmedia https://visionmedia.github.io/page.js/
 *  MIT License
 *
 */
define(['config', 'jquery', 'bus', 'page/query', 'browser'], function(ECMWF, $, Bus, qs, Browser) {
	/**
	 * To work properly with the URL
	 * history.location generated polyfill in https://github.com/devote/HTML5-History-API
	 */
	var location = ('undefined' !== typeof window) && (window.history.location || window.location);

	/**
	 * Perform initial dispatch.
	 */
	var dispatch = true;

	/**
	 * Decode URL components (query string, pathname, hash).
	 * Accommodates both regular percent encoding and x-www-form-urlencoded format.
	 */
	var decodeURLComponents = true;

	/**
	 * Base path.
	 */
	var base = '';

	/**
	 * Running flag.
	 */
	var running;

	/**
	 * HashBang option
	 */
	var hashbang = false;

	/**
	 * Previous context, for capturing
	 * page exit events.
	 */
	var prevContext;

	/**
	 * Register `path` with callback `fn()`,
	 * or route `path`, or redirection,
	 * or `page.start()`.
	 *
	 *   page(fn);
	 *   page('*', fn);
	 *   page('/user/:id', load, user);
	 *   page('/user/' + user.id, { some: 'thing' });
	 *   page('/user/' + user.id);
	 *   page('/from', '/to')
	 *   page();
	 *
	 * @param {String|Function} path
	 * @param {Function} fn...
	 * @api public
	 */
	function page(path, fn) {
		if (typeof path === 'function') {
			// <callback>
			return page('*', path);
		}

		if (typeof fn === 'function') {
			// route <path> to <callback ...>
			var route = new Route(path);
			for (var i = 1; i < arguments.length; ++i) {
				page.callbacks.push(route.middleware(arguments[i]));
			}
		} else if (typeof path === 'string') {
			// show <path> with [state]
			page[typeof fn === 'string' ? 'redirect' : 'show'](path, fn);
		} else {
			// start [options]
			page.start(path);
		}
	}

	/**
	 * Callback functions.
	 */
	page.callbacks = [];
	page.exits = [];

	/**
	* Current path being processed
	* @type {String}
	*/
	page.current = {
		path: '',
		state_id: 0
	};
	page._path = {};
	page.q = {}; // allow different clients to add their own independent set of q parameters, eg {facets: {facets: 'extended'}}

	page._state_id = 0; // next state. used to group changes together in a single history item.

	/**
	 * Number of pages navigated to.
	 * @type {number}
	 *
	 *     page.len == 0;
	 *     page('/login');
	 *     page.len == 1;
	 */
	page.len = 0;

	/**
	 * Get or set basepath to `path`.
	 *
	 * @param {String} path
	 * @api public
	 */
	page.base = function(path) {
		if (0 === arguments.length) return base;
		base = path;
	};

	/**
	 * Bind with the given `options`.
	 *
	 * Options:
	 *
	 *    - `click` bind to click events [true]
	 *    - `popstate` bind to popstate [true]
	 *    - `dispatch` perform initial dispatch [true]
	 *
	 * @param {Object} options
	 * @api public
	 */
	page.start = function(options) {
		options = options || {};
		if (running) return;
		running = true;
		if (false === options.dispatch) dispatch = false;
		if (false === options.decodeURLComponents) decodeURLComponents = false;
		if (false !== options.popstate) window.addEventListener('popstate', onpopstate, false);
		if (false !== options.click) window.addEventListener('click', onclick, false);
		if (true === options.hashbang) hashbang = true;
		if (!dispatch) return;
		var url = (hashbang && ~location.hash.indexOf('#!')) ? location.hash.substr(2) + location.search : location.pathname + location.search + location.hash;

		this.path = location.href.replace(location.protocol + '//' + location.host + this.base_path, '');
		var c = this.parse_path();

		var _page = c[1];
		var dir = c[0].join('/');

		if(ECMWF.browser.msie9){
			var path = this.original_path.split('#')[0].replace(this.base_path + this.app_path, '');
			this.path = '/' + this.app_path.replace('/', '') + path;
			c = [[this.path.split('/')[0]], '', undefined];
			dir = path.split('/')[0];
			_page = '';
		}

		// FIXME setting page title from url is wrong, it needs to come from the Product title
		this.set_path(dir, _page, c[2], _page ? _page : ECMWF.ucfirst(dir));

		Bus().listen('add-page-state', function() {
			page.add_state();
		});

		return this;
	};

	/*
	 *   Return an array with 3 elements
	 *     0- array of path tokens up to but not including the leaf name
	 *     1- leaf (product) name
	 *     2- query string with leading '?'
	 */
	page.parse_path = function() {
		var d = location.href.replace(location.protocol + '//' + location.host + this.base_path + this.app_path, '').split('/');
		var c = d.pop().split('?');
		if(c[1]) c[1] = '?' + c[1];

		return [d, c[0], c[1]];
	}

	/*
	 *  Parse the URL to determine BASE, APP, PACKAGE and PRODUCT.
	 *  The path must now include the base_path
	 */
	page.app_tokens_by_path = function(path) {
		var r = {
			app: 'charts',
			app_path: '/',
			base_path: '',
			chart_package: ECMWF.options.default_package, // landing page should show ALL packages, but alternatively a default package can be shown instead
			tokens: path.replace(/\/$/, '').split('/')
		}
		r.tokens.shift();

		var i = -1;
		if(ECMWF.chart_app && ECMWF.chart_app.base_url){
			i = ECMWF.chart_app.base_url.split('/').length - 1;
		}else if(r.tokens.length > 1 && r.tokens[1] === 'tcyclone'){
			i = 1;
		}else{
			// parse url without config (deprecated, but still useful if we want to support multiple apps on the same site)
			var match = $.map(r.tokens, function(v, i){
				if(!!~$.inArray(v, ['charts', 'charts-beta', 'browser', 'tabs'])){
					return i;
				}
			});
			i = match.length ? match[0] : -1;
		}

		if(i > -1){
			r.app = r.tokens[i];
			r.app_path = '/' + r.app + '/';

			if(!r.chart_package && (r.tokens.length > i + 1 && (r.tokens[i + 1] !== 'all')))
				r.chart_package = r.tokens[i + 1]

			r.base_path = '/' + r.tokens.slice(0, i).join('/');
			if(r.base_path == '/') r.base_path = '';
		}

		return r;
	}

	/**
	 * Unbind click and popstate event handlers.
	 *
	 * @api public
	 */
	page.stop = function() {
		if (!running) return;
		page.current.path = '';
		page.len = 0;
		running = false;
		window.removeEventListener('click', onclick, false);
		window.removeEventListener('popstate', onpopstate, false);
	};

	/**
	 * Show `path` with optional `state` object.
	 *
	 * @param {String} path
	 * @param {Object} state
	 * @param {Boolean} dispatch
	 * @return {Context}
	 * @api public
	 */
	page.show = function(path, state, dispatch, push) {
		var ctx = new Context(path, state);
		page.current = {path: ctx.path, state_id: this._state_id};
		if (dispatch !== false) page.dispatch(ctx);
		if (ctx.handled !== false && push !== false) ctx.pushState();
		return ctx;
	};

	/**
	 * Goes back in the history
	 * Back should always let the current route push state and then go back.
	 *
	 * @param {String} path - fallback path to go back if no more history exists, if undefined defaults to page.base
	 * @param {Object} [state]
	 * @api public
	 */
	page.back = function(path, state) {
		if (page.len > 0) {
			// this may need more testing to see if all browsers
			// wait for the next tick to go back in history
			history.back();
			page.len--;
		} else if (path) {
			setTimeout(function() {
				page.show(path, state);
			});
		}else{
			setTimeout(function() {
				page.show(base, state);
			});
		}
	};

	/**
	 * Register route to redirect from one path to other
	 * or just redirect to another route
	 *
	 * @param {String} from - if param 'to' is undefined redirects to 'from'
	 * @param {String} [to]
	 * @api public
	 */
	page.redirect = function(from, to) {
		// Define route from a path to another
		if ('string' === typeof from && 'string' === typeof to) {
			page(from, function(e) {
				setTimeout(function() {
					page.replace(to);
				}, 0);
			});
		}

		// Wait for the push state and replace it with another
		if ('string' === typeof from && 'undefined' === typeof to) {
			setTimeout(function() {
				page.replace(from);
			}, 0);
		}
	};

	/**
	 * Replace `path` with optional `state` object.
	 *
	 * @param {String} path
	 * @param {Object} state
	 * @return {Context}
	 * @api public
	 */
	page.replace = function(path, state, init, dispatch) {
		var ctx = new Context(path, state);
		page.current.path = ctx.path;
		ctx.init = init;
		ctx.save(); // save before dispatching, which may redirect
		if (dispatch !==  false) page.dispatch(ctx);
		return ctx;
	};

	page.set_base_path = function(path) {
		page.base(this.base_path = path);
	}

	page.route = function(path, fn) {
		page(path, fn);
		return this;
	}

	/*
	 * set_path is the preferred way of initiating a route change.
	 * This is because specialised knowledge of the product is needed to set the title.
	 * Using HTML links will bypass this and hence the title will not be set.
	 *
	 * Note that passing null for a parameter will reset the current value but passing undefined will preserve it.
	 */
	page.set_path = function(dir, _page, q, title) {
		// TODO check behaviour when regular links are clicked - is page state properly set?

		if(typeof q === 'object') q = qs.get_string(q);

		var p = $.extend({}, this._new_path, {dir: dir, page: _page, q: q});

		var a = [p.page];
		p.dir && a.unshift(p.dir);
		var path = this.app_path + (a.join('/') + (q || (this._new_path && this._new_path.q) || '')).replace(/%2C/g, ',');

		(/*(this._first === false) && */(path === this.path)) || console.log('router.set_path: ', [dir, _page, q]);

		(path !== this.path) && (
			console.log('router.set_path: path has changed:'),
			console.log('       ', this.path),
			console.log('       ', path)
		);

		((this._first !== false) || (path !== this.path)) && (
			this._new_path = p, // set a 'local' variable for comparisons before the dispatch has finished.
			this._first = false,
			page[this._state_id > this.current.state_id ? 'show' : 'replace'](this.path = path),
			this._path = p,

			this.set_title(title)
		);
	}

	// this preserves everything except the directory
	page.set_dir = function(dir) {
		if((dir !== this._path.dir) || this._path.page){
			this.set_path(dir);
		}
	},

	page.set_product = function(product_name, title) {
		this.set_path(undefined, product_name, undefined, title);
	},

	// query parameters are merged with the existing ones
	page.set_parameter = function(params) {
		var _q = this._path.q;
		var q = $.extend(qs.parse((_q || '?').split('?')[1]), params);
		this.set_path(undefined, undefined, q);
	},

	page.goto_list_page = function(dir) {
		dir = dir || this._path.dir;
		if((dir !== this._path.dir) || this._path.page){
			this.set_path(dir, null, undefined, ECMWF.ucfirst(this._path.dir));
		}
	},

	page.set_title = function(_title) {
		var split = this.path.split('?');
		var prefix = ECMWF.options.title_prefix || 'ECMWF'
		if(split.length === 2){
			var query = qs.parse(split[1]);
			var params = (query.area ? query.area + ' ' : '') + (query.time || '');
			var title = [prefix, ECMWF.truncate(_title || this._path.page || 'Charts', 32), params];
		}else{
			title = [prefix, _title];
		}
		title = title.join(' | ');

		try {
			document.getElementsByTagName('title')[0].innerHTML = title.replace('<', '&lt;').replace('>', '&gt;').replace(' & ', ' &amp; ');
		}
		catch (Exception) {}
	}

	/*
	 *  Add a set of parameters with a namespace
	 *  The purpose of this is so that parameters can be cleared for one namespace (eg chart) without affecting the other namespaces
	 *  Currently this is only used for facets, but it should be used for chart parameters also
	 */
	page.add_parameter = function(name) {
		Object.defineProperty(page.q, name, {
			get: function() {
				return this.value
			},
			set: function(val) {
				this.value = val
				var p = {}
				p[name] = val
				page.set_parameter(p)
			}
		});
	}

	page.add_state = function() {
		this._state_id++;
		console.log('* router.add_state:', this._state_id);

		setTimeout(function() {
			console.log('tracker update: ga=', window.ga);
			window.ga && ga('send', 'pageview', document.location.href);
		}, 100)

		return this;
	}

	/**
	 * Dispatch the given `ctx`.
	 *
	 * @param {Object} ctx
	 * @api private
	 */
	page.dispatch = function(ctx) {
		var prev = prevContext,
		i = 0,
		j = 0;

		prevContext = ctx;

		function nextExit() {
			var fn = page.exits[j++];
			if (!fn) {
				if($.map(page._exit_routes, function(v, i) {
					if(v === ctx.canonicalPath) console.log('dispatch: found exit match: ', ctx.canonicalPath);
					if(v === ctx.canonicalPath) return v;
				}).length) {
					return unhandled(ctx);
				} else
					return nextEnter();
			}
			fn(prev, nextExit);
		}

		function nextEnter() {
			var fn = page.callbacks[i++];

			if (ctx.path !== page.current.path) {
				ctx.handled = false;
				return;
			}
			if (!fn) return unhandled(ctx);
			fn(ctx, nextEnter);
		}

		if (prev) {
			nextExit();
		} else {
			nextEnter();
		}
	};

	/**
	 * Unhandled `ctx`. When it's not the initial
	 * popstate then redirect. If you wish to handle
	 * 404s on your own use `page('*', callback)`.
	 *
	 * @param {Context} ctx
	 * @api private
	 */
	function unhandled(ctx) {
		if (ctx.handled) return;
		var current;

		if (hashbang) {
			current = base + location.hash.replace('#!', '');
		} else {
			current = location.pathname + location.search;
		}

		if (current === ctx.canonicalPath) return;
		page.stop();
		ctx.handled = false;
		location.href = ctx.canonicalPath;
	}

	/**
	 * Register an exit route on `path` with
	 * callback `fn()`, which will be called
	 * on the previous context when a new
	 * page is visited.
	 *
	 * Note that this is not a way to exit from the application
	 */
	page.exit = function(path, fn) {
		if (typeof path === 'function') {
			return page.exit('*', path);
		}

		var route = new Route(path);
		for (var i = 1; i < arguments.length; ++i) {
			page.exits.push(route.middleware(arguments[i]));
		}
	};

	page._exit_routes = [];

	page.exit_route = function(path) {
		if(Array.isArray(path)){
			page._exit_routes = page._exit_routes.concat(path);
		}else{
			page._exit_routes.push(path);
		}
	};

  /**
   * Remove URL encoding from the given `str`.
   * Accommodates whitespace in both x-www-form-urlencoded
   * and regular percent-encoded form.
   *
   * @param {str} URL component to decode
   */
  function decodeURLEncodedURIComponent(val) {
    if (typeof val !== 'string') { return val; }
    return decodeURLComponents ? decodeURIComponent(val.replace(/\+/g, ' ')) : val;
  }

  /**
   * Initialize a new "request" `Context`
   * with the given `path` and optional initial `state`.
   *
   * @param {String} path
   * @param {Object} state
   * @api public
   */
  function Context(path, state) {
    /*
     *   This feature of appending the base_url is confusing when mixing internal and external links
     *   Perhaps it is more trouble than it is worth - should we insist on using the full path?
     */
    if ('/' === path[0] && 0 !== path.indexOf(base)) path = base + (hashbang ? '#!' : '') + path;
    var i = path.indexOf('?');

    this.canonicalPath = path;
    this.path = path.replace(base, '') || '/';
    if (hashbang) this.path = this.path.replace('#!', '') || '/';

    this.title = document.title;
    this.state = state || {};
    this.state.path = path;
    this.querystring = ~i ? decodeURLEncodedURIComponent(path.slice(i + 1)) : '';
    this.pathname = decodeURLEncodedURIComponent(~i ? path.slice(0, i) : path);
    this.params = {};

    // fragment
    this.hash = '';
    if (!hashbang) {
      if (!~this.path.indexOf('#')) return;
      var parts = this.path.split('#');
      this.path = parts[0];
      this.hash = decodeURLEncodedURIComponent(parts[1]) || '';
      this.querystring = this.querystring.split('#')[0];
    }
  }

  /**
   * Expose `Context`.
   */
  page.Context = Context;

  /**
   * Push state.
   *
   * @api private
   */
  Context.prototype.pushState = function() {
    page.len++;
    history.pushState(this.state, this.title, hashbang && this.path !== '/' ? '#!' + this.path : this.canonicalPath);
  };

  /**
   * Save the context state.
   *
   * @api public
   */
  Context.prototype.save = function() {
    history.replaceState(this.state, this.title, hashbang && this.path !== '/' ? '#!' + this.path : this.canonicalPath);
  };

  /**
   * Initialize `Route` with the given HTTP `path`,
   * and an array of `callbacks` and `options`.
   *
   * Options:
   *
   *   - `sensitive`    enable case-sensitive routes
   *   - `strict`       enable strict matching for trailing slashes
   *
   * @param {String} path
   * @param {Object} options.
   * @api private
   */
  function Route(path, options) {
    options = options || {};
    this.path = (path === '*') ? '(.*)' : path;
    this.method = 'GET';
    this.regexp = pathtoRegexp(this.path,
      this.keys = [],
      options.sensitive,
      options.strict);
  }

  /**
   * Expose `Route`.
   */
  page.Route = Route;

  /**
   * Return route middleware with
   * the given callback `fn()`.
   *
   * @param {Function} fn
   * @return {Function}
   * @api public
   */
  Route.prototype.middleware = function(fn) {
    var self = this;
    return function(ctx, next) {
      if (self.match(ctx.path, ctx.params)) return fn(ctx, next);
      next();
    };
  };

  /**
   * Check if this route matches `path`, if so
   * populate `params`.
   *
   * @param {String} path
   * @param {Object} params
   * @return {Boolean}
   * @api private
   */
  Route.prototype.match = function(path, params) {
    var keys = this.keys,
      qsIndex = path.indexOf('?'),
      pathname = ~qsIndex ? path.slice(0, qsIndex) : path,
      m = this.regexp.exec(decodeURIComponent(pathname));

    if (!m) return false;

    for (var i = 1, len = m.length; i < len; ++i) {
      var key = keys[i - 1];
      var val = decodeURLEncodedURIComponent(m[i]);
      if (val !== undefined || !(hasOwnProperty.call(params, key.name))) {
        params[key.name] = val;
      }
    }

    return true;
  };

	/**
	 * Handle "populate" events.
	 */
	function onpopstate(e) {
		if (e.state) {
			var path = e.state.path;

			var c = page.parse_path();
			var dir = c[0].join('/');
			var p = $.extend({}, page._path, {dir: dir, page: c[1], q: c[2]});
			page._new_path = p;

			var ctx = page.replace(path, e.state);

			if(!!~$.inArray(ctx.canonicalPath, page._exit_routes)){
				location.href = ctx.canonicalPath;
				return;
			}

			page._path = p;
			page.path = path; // TODO replace has set page.current.path instead. why are we not using page.current.path in set_path?
		} else {
			page.show(location.pathname + location.hash, undefined, undefined, false);
		}
	}

  /**
   * Handle "click" events.
   */
  function onclick(e) {

    if (1 !== which(e)) return;

    if (e.metaKey || e.ctrlKey || e.shiftKey) return;
    if (e.defaultPrevented) return;


    // ensure link
    var el = e.target;
    while (el && 'A' !== el.nodeName) el = el.parentNode;
    if (!el || 'A' !== el.nodeName) return;


    // Ignore if tag has
    // 1. "download" attribute
    // 2. rel="external" attribute
    if (el.getAttribute('download') || el.getAttribute('rel') === 'external') return;

    // ensure non-hash for the same path
    var link = el.getAttribute('href');
    if (!hashbang && el.pathname === location.pathname && (el.hash || '#' === link)) return;

    // Check for mailto: in the href
    if (link && link.indexOf('mailto:') > -1) return;

    // check target
    if (el.target) return;

    // x-origin
    if (!sameOrigin(el.href)) return;

    // rebuild path
    var path = el.pathname + el.search + (el.hash || '');


    // same page
    var orig = path;

    path = path.replace(base, '');
    if (hashbang) path = path.replace('#!', '');

    if (base && orig === path) return;

    e.preventDefault();
    page.show(orig);
  }

  /**
   * Event button.
   */
  function which(e) {
    e = e || window.event;
    return null === e.which ? e.button : e.which;
  }

  /**
   * Check if `href` is the same origin.
   */
  function sameOrigin(href) {
    var origin = location.protocol + '//' + location.hostname;
    if (location.port) origin += ':' + location.port;
    return (href && (0 === href.indexOf(origin)));
  }

  page.sameOrigin = sameOrigin;

//};//,{"path-to-regexp":2}],2:[function(require,module,exports){
//var isArray = require('isarray');

/**
 * Expose `pathtoRegexp`.
 */
//module.exports = pathtoRegexp;

/**
 * The main path matching regexp utility.
 *
 * @type {RegExp}
 */
var PATH_REGEXP = new RegExp([
  // Match already escaped characters that would otherwise incorrectly appear
  // in future matches. This allows the user to escape special characters that
  // shouldn't be transformed.
  '(\\\\.)',
  // Match Express-style parameters and un-named parameters with a prefix
  // and optional suffixes. Matches appear as:
  //
  // "/:test(\\d+)?" => ["/", "test", "\d+", undefined, "?"]
  // "/route(\\d+)" => [undefined, undefined, undefined, "\d+", undefined]
  '([\\/.])?(?:\\:(\\w+)(?:\\(((?:\\\\.|[^)])*)\\))?|\\(((?:\\\\.|[^)])*)\\))([+*?])?',
  // Match regexp special characters that should always be escaped.
  '([.+*?=^!:${}()[\\]|\\/])'
].join('|'), 'g');

/**
 * Escape the capturing group by escaping special characters and meaning.
 *
 * @param  {String} group
 * @return {String}
 */
function escapeGroup (group) {
  return group.replace(/([=!:$\/()])/g, '\\$1');
}

/**
 * Attach the keys as a property of the regexp.
 *
 * @param  {RegExp} re
 * @param  {Array}  keys
 * @return {RegExp}
 */
function attachKeys (re, keys) {
  re.keys = keys;

  return re;
}

/**
 * Normalize the given path string, returning a regular expression.
 *
 * An empty array should be passed in, which will contain the placeholder key
 * names. For example `/user/:id` will then contain `["id"]`.
 *
 * @param  {(String|RegExp|Array)} path
 * @param  {Array}                 keys
 * @param  {Object}                options
 * @return {RegExp}
 */
function pathtoRegexp (path, keys, options) {
  if (!Array.isArray(keys)) {
    options = keys;
    keys = null;
  }

  keys = keys || [];
  options = options || {};

  var strict = options.strict;
  var end = options.end !== false;
  var flags = options.sensitive ? '' : 'i';
  var index = 0;

  if (path instanceof RegExp) {
    // Match all capturing groups of a regexp.
    var groups = path.source.match(/\((?!\?)/g);

    // Map all the matches to their numeric indexes and push into the keys.
    if (groups) {
      for (var i = 0; i < groups.length; i++) {
        keys.push({
          name:      i,
          delimiter: null,
          optional:  false,
          repeat:    false
        });
      }
    }

    // Return the source back to the user.
    return attachKeys(path, keys);
  }

  // Map array parts into regexps and return their source. We also pass
  // the same keys and options instance into every generation to get
  // consistent matching groups before we join the sources together.
  if (Array.isArray(path)) {
    var parts = [];

    for (i = 0; i < path.length; i++) {
      parts.push(pathtoRegexp(path[i], keys, options).source);
    }
    // Generate a new regexp instance by joining all the parts together.
    return attachKeys(new RegExp('(?:' + parts.join('|') + ')', flags), keys);
  }

  // Alter the path string into a usable regexp.
  path = path.replace(PATH_REGEXP, function (match, escaped, prefix, key, capture, group, suffix, escape) {
    // Avoiding re-escaping escaped characters.
    if (escaped) {
      return escaped;
    }

    // Escape regexp special characters.
    if (escape) {
      return '\\' + escape;
    }

    var repeat   = suffix === '+' || suffix === '*';
    var optional = suffix === '?' || suffix === '*';

    keys.push({
      name:      key || index++,
      delimiter: prefix || '/',
      optional:  optional,
      repeat:    repeat
    });

    // Escape the prefix character.
    prefix = prefix ? '\\' + prefix : '';

    // Match using the custom capturing group, or fallback to capturing
    // everything up to the next slash (or next period if the param was
    // prefixed with a period).
    capture = escapeGroup(capture || group || '[^' + (prefix || '\\/') + ']+?');

    // Allow parameters to be repeated more than once.
    if (repeat) {
      capture = capture + '(?:' + prefix + capture + ')*';
    }

    // Allow a parameter to be optional.
    if (optional) {
      return '(?:' + prefix + '(' + capture + '))?';
    }

    // Basic parameter support.
    return prefix + '(' + capture + ')';
  });

  // Check whether the path ends in a slash as it alters some match behaviour.
  var endsWithSlash = path[path.length - 1] === '/';

  // In non-strict mode we allow an optional trailing slash in the match. If
  // the path to match already ended with a slash, we need to remove it for
  // consistency. The slash is only valid at the very end of a path match, not
  // anywhere in the middle. This is important for non-ending mode, otherwise
  // "/test/" will match "/test//route".
  if (!strict) {
    path = (endsWithSlash ? path.slice(0, -2) : path) + '(?:\\/(?=$))?';
  }

  // In non-ending mode, we need prompt the capturing groups to match as much
  // as possible by using a positive lookahead for the end or next path segment.
  if (!end) {
    path += strict && endsWithSlash ? '' : '(?=\\/|$)';
  }

  return attachKeys(new RegExp('^' + path + (end ? '$' : ''), flags), keys);
}

return page;

});
