/*
 *  Messages sent:
 *      - chart-product-availability          availability here refers to the object that contains the actual availability. message is sent even if the actual availability is unchanged, but the selected date has changed.
 *      - chart-product-new-availability      sent after we learn that the server has new availability data. We dont yet have the data (check this, probably we now do have the data)
 *      - chart-availability-invalidated      indicates to observers that the chart has no availability data.
 *      - chart-content-update                sent following *any* change to the chart. Includes a 'changed' property describing the type of change.
 *                                            Because the backend may change anything upon redraw, all subscribers must listen to this signal, making other signals somewhat redundant.
 *      - chart-time-select                   notifies clients that the chart has been changed to show a new time (though the image may still be being retrieved).
 *      - chart-product-availability-mismatch currently nobody listens to this message
 *      - chart-mode
 *
 *    via the animation plugin:
 *      - chart-animation-stop                sent when changing to Pause state
 *      - chart-animation-load-complete
 *      - chart-frame-loaded
 *
 *    via chartAnimationRangeController
 *      - chart-animation-selection-change    sent when the range has changed.
 *
 *  Incoming dom events:
 *      [ none ? ]
 */
define(['config', 'jquery', 'ajax_queue', 'bus', 'time', 'matrix', './product', './model/parameters', 'undo', 'utils', './animator', './spinner', './utils', 'jquery-ui/mouse', 'throttle-debounce', /*'ajax-loader', */'./chart.css'], function(ECMWF, $, Ajaxqueue, Bus, Time, sylvester, Product, Parameters, Undo, Utils, Animator, Spinner){

    'use strict';

    ECMWF.Chart || (ECMWF.Chart = {});

    ECMWF.Chart.Prototype = {
        options: {
            product: 'empty', // Either the name of a product or a conformant object.
            projection: undefined,
            time: null,
            distance: 10, // For ui-mouse
            messageBus: Bus(),
            mousewheel: null, // Boolean
            touch: touch,
            gesture: gesture,
            frame_count: 3, // Number of frames to load simultaneously
            expand_on_resize: true, // On resize, the image is expanded instead of being resized
            title: $(), // Dom element to hold the title.
            availability: true, // If set, availability data will be periodically updated.
            priority: 'very-high',
            editable: true,
            show_latlong: true, // Will not be shown if it is disabled in user_prefs
            show_title_overlay: false,
            preload_animations: true,
            use_outline: false,
            enable_undo: true,
            show_marker: true,
            messages: true,
            zoom_level: 5,
            use_product_projection: false,
            animation_loader: null,
            format: undefined, // default is PNG, or user preference
            values: undefined, // Preset values to be passed to product
            plugins: []        // list of mixins objects with the properties: mixin, options.
        },

        _create: function() {
            this.plugins = [];
            for (var i in chart_plugins) {
                var plugin = new chart_plugins[i](this);
                plugin && (plugin.name = i, this.plugins.push(plugin));
            }
        },

        _init: function() {
            var self = this,
                o = self.options;

            if (!o.touch) {
                this._mouseInit();
            }

            var name = 'map_' + this.uuid;

            this.zoom_level = o.zoom_level;

            this.image_pool = [];
            self.in_destroy = false;
            self.bus = o.messageBus;

            this.element.addClass('chart');

            (o.mousewheel === null) && (o.mousewheel = $.getUserPreference('chart__enable_scrollwheel'));

            self.title = o.title || null;

            var clip = $('<div class="chart-clip">').prependTo(this.element)
                .css({
                    position: 'absolute'
                })
                .mousedown(function(event) {
                    event.preventDefault();
                });

            var img = $(new Image()).addClass('chart-image')
                .attr({
                    id: name + '_image'
                })
                .css({
                    position: 'absolute',
                    cursor: 'move'
                })
                .appendTo(clip);

            self.name = name;

            self.product = ProductFactory({
                owner: name,
                product: o.product,
                projection: o.projection,
                use_product_projection: o.use_product_projection,
                values: o.values
            });

            self._promises = {size: new $.Deferred()};

            self.renderer = new DomChartRenderer(
                img,
                function() { return self.product.size(); },
                function() { return self._viewport(); }
            );

            self.plugins_v2 = $.map(o.plugins, function(mixin) {
                if(typeof mixin === 'function') mixin = {mixin: mixin}
                return new mixin.mixin(self, mixin.options);
            });

            self.img = img;
            self.clip = clip;
            self.current_tool = o.editable ? 'pan' : 'select';
            self.zoom = {};
            self.drag = {};
            self.time = o.time;

            Animator && new Animator.Controller(self, o.preload_animations, o.animation_loader);

            self.matrix = (function() {

                var _apply = $.debounce(500, function(){
                    this.transformMap(this.matrix.magics);
                    this._pinch_gesture = false;
                    this._drag_gesture = false;
                });

                return {
                    magics: null,
                    screen: null,

                    reset: function() {
                        this.magics = sylvester.Matrix.I(3);
                        this.screen = sylvester.Matrix.I(3);
                    },

                    transform: function(x, y, a) {

                        // zoom
                        var mm = $M([
                            [1 / a,     0, 0],
                            [0,     1 / a, 0],
                            [0,         0, 1]
                        ]);

                        var ms = $M([
                            [a, 0, 0],
                            [0, a, 0],
                            [0, 0, 1]
                        ]);

                        // Translate
                        var t1 = $M([
                            [1, 0, -x],
                            [0, 1, -y],
                            [0, 0,  1]
                        ]);

                        var t2 = $M([
                            [1, 0, x],
                            [0, 1, y],
                            [0, 0, 1]
                        ]);

                        this.magics = this.pinch_magics.multiply(t2.x(mm).x(t1));
                        this.screen = this.pinch_screen.multiply(t2.x(ms).x(t1));

                        this.apply();

                        //self.transformations.zoom = self.zoom.view.width / port.width;
                    },

                    apply: function() {
                        var self = this.self;
                        var rect = self._rect();

                        var tl = {
                            x: rect.left,
                            y: rect.top
                        };
                        var br = {
                            x: rect.right,
                            y: rect.bottom
                        };

                        tl = $V([tl.x, tl.y, 1]);
                        br = $V([br.x, br.y, 1]);

                        tl = this.screen.x(tl);
                        br = this.screen.x(br);

                        self.img.css(self.zoom.view = from_rect({
                            left: tl.e(1),
                            top: tl.e(2),
                            right: br.e(1),
                            bottom: br.e(2)
                        }));

                        _apply.call(self);

                        return this;
                    }
                }
            })();

            self.matrix.reset();
            self.matrix.self = this;

            self.checkAvailability(o.availability);

            if (o.touch) {
                console.warn('MAP TOUCH ' + o.touch);
            }

            self
                ._listen('city-select', function(_, msg) {
                    self._push_undo();
                    self.product.show_city(msg);
                    self._clear()._redraw();
                })
                ._listen('send-magics-stuff', function(_, msg) {
                    // This will redraw the map and send an email with the grib, magml...
                    self.mail = true;
                    self._clear()._redraw();
                    self.mail = false;
                })
                ._listen('collect-report-info', function(_, msg) {
                    msg.chart = self;
                })

                ._listen('tool-clear-click', function() {
                    self.setProduct('empty');
                    o.messageBus.send('chart-availability-invalidated');
                })
                ._listen('tool-full-map', function() {
                    self.setProjection(self.product.full_map_projection_name());
                })
                ._listen('vertical-longitude-select', function(_, pos) {
                    self.verticalLongitude(pos.lon);
                })
                ._listen('tool-zoom-in-click', function() {
                    self.zoomScale(2);
                })
                ._listen('tool-zoom-out-click', function() {
                    self.zoomScale(0.5);
                })
                ._listen('tool-drag-zoom-select', function(_, cursor) {
                    self._set_mode('zoom', cursor);
                })
                ._listen('tool-drag-pan-select', function(_, cursor) {
                    self._set_mode('pan', cursor);
                })
                ._listen('tool-select-select', function(_, cursor) {
                    self._set_mode('select', cursor);
                })
                ._listen('tool-vertical-select', function(_, cursor) {
                    self._set_mode('vertical', cursor);
                })
                ._listen('chart-frame-loaded', function(_, time, result) {
                    if (result) {
                        if (result.width)
                            img.css({width: result.width, height: result.height});
                        else if (result.magics)
                            img.css({width: result.magics.projection.img_width, height: result.magics.projection.img_height})
                        else
                            self._chart._viewport && img.css(self._chart._viewport());
                    }
                })
                ._listen('slider-time-change', function(_, dates) {
                    self._set_time(dates);
                })
                ._listen('layer-visibility-change', function(_, msg) {
                    self._push_undo();
                    self.product.layerVisibility(msg.layer_id, msg.visibility);
                    self._clear()._redraw(function() {
                        self._notifyProductChange();
                    });
                })
                ._listen('layer-delete', function(_, msg) {
                    self._push_undo();
                    self.product.layerDelete(msg.layer_id);
                    self._clear()._redraw(function() {
                        self._notifyProductChange();
                    });
                });

            o.messageBus.listen(self, 'layer-options-change', function(_, msg) {
                self._push_undo();
                self.product.layerOptions(msg.layer_id, msg.options);
                self._clear()._redraw(function() {
                    self._notifyProductChange();
                });
            }).listen(self, 'layer-order-change', function(_, msg) {
                self._push_undo();
                self.product.layerOrder(msg.order);
                self._clear()._redraw(function() {
                    self._notifyProductChange();
                });
            }).listen(self, 'product-thumbnail-click', function(_, msg) {
                self.setProduct(msg.name, msg.use_product_projection);
            }).listen(self, 'projection-thumbnail-click', function(_, msg) {
                self.setProjection(msg.name);
            }).listen(self, 'layer-thumbnail-click', function(_, msg) {
                self.addLayers([msg.name]);
            }).listen(self, 'layer-dialogue-selection', function(_, msg) { //list is just array of names.
                self.addLayers(msg.names);
            }).listen(self, 'base-time-select', function(_, msg) {
                console.warn('base-time-select', msg);
                self._set_cycle(msg.cycle);
            }).listen(self, 'prev-step', function(_, msg) {
                if (self.time) {
                    var prev = self.time.previous(self.product._availability);
                    if (prev) self._set_time(prev);
                }
            }).listen(self, 'next-step', function(_, msg) {
                if (self.time) {
                    var next = self.time.next(self.product._availability);
                    if (next) self._set_time(next);
                }
            }).listen(self, 'prev-day', function(_, msg) {
                if (self.time) {
                    var prev_day = self.time.valid_time.getTime() - 1000 * 60 * 60 * 24;
                    for (var prev = self.time; prev; prev = prev.previous(self.product._availability)) {
                        if (prev.valid_time.getTime() <= prev_day) {
                            self._set_time(prev);
                            break;
                        }
                    }
                }
            }).listen(self, 'next-day', function(_, msg) {
                if (self.time) {
                    var next_day = self.time.valid_time.getTime() + 1000 * 60 * 60 * 24;
                    for (var next = self.time; next; next = next.next(self.product._availability)) {
                        if (next.valid_time.getTime() >= next_day) {
                            self._set_time(next);
                            break;
                        }
                    }
                }
            }).listen(self, 'first-step', function(_, msg) {
                var time = self.time.first(self.product._availability);
                if (time) self._set_time(time);
            }).listen(self, 'last-step', function(_, msg) {
                var time = self.time.last(self.product._availability);
                if (time) self._set_time(time);
            }).listen(self, 'prev-base-time', function(_, msg) {
                if (self.time) {
                    var prev = self.time.previousBaseTime(self.product._availability);
                    if (prev) self._set_time(prev);
                }
            }).listen(self, 'next-base-time', function(_, msg) {
                if (self.time) {
                    var next = self.time.nextBaseTime(self.product._availability);
                    if (next) self._set_time(next);
                }
            }).listen(self, 'best-base-time', function(_, msg) {
                if (self.time) {
                    var next = self.time.bestBaseTime(self.product._availability);
                    if (next) self._set_time(next);
                }
            }).listen(self, 'preferences-changed', function() {
                if (o.show_latlong) {
                    if ($.getUserPreference('show_latlong')) {
                        self.latlong.length || self.element.div().latlong({
                            chart: self
                        });
                    } else {
                        self.latlong.remove();
                        self.latlong = $();
                    }
                }

                var mousewheel;
                if ((mousewheel = $.getUserPreference('chart__enable_scrollwheel')) !== o.mousewheel) {
                    enable_mousewheel(mousewheel);
                }

                self._redraw();
            });

            if(self.product.api > 1.0){
                self.product.variables.time.subscribe(self, function(param) {
                    self.time = param.value;
                    this.prev && self._clear()._redraw();
                });
            }

            function enable_mousewheel(enabled) {
                if ((o.mousewheel = (enabled !== false))) {
                    self.element.mousewheel(function(e, delta) {
                        self.mouseWheel(e, delta);
                    });
                } else {
                    self.element.unmousewheel();
                }
            }

            if (o.mousewheel && this.element.mousewheel) {
                enable_mousewheel();
            }

            this.resize(self.options.size || {width: self.element.width(), height: self.element.height()}, function() {
                self._notifyProductChange();
            });

            self.latlong = o.show_latlong && ($.getUserPreference || $.getUserPreference('show_latlong'))
                ? this.element.div().latlong({chart: self})
                : $();

            self._mouse = {
                pan: {
                    start: [self._mouseStart_pan],
                    drag: [self._mouseDrag_pan]
                },
                zoom: {
                    start: [self._mouseStart_zoom],
                    drag: [self._mouseDrag_zoom]
                },
                select: {
                    start: [self._mouseStart_select],
                    drag: [self._mouseDrag_select]
                },
                vertical: {
                    start: [self._mouseStart_vertical],
                    drag: [self._mouseDrag_vertical]
                },
                wheel: [],
                pinch: []
            };
        },

        _set_time: function(time) {
            var self = this;

            if (!time) time = self.time.bestBaseTime(self.product.api > 1.0 ? self.product.variables.time.availability : self.product._availability);
            if (__DEV__ && !time.valid_time) return console.warn('chart._set_time: bad arg: time: ', time);

            self._clear();
            self.time = time;

            if(self.product.api > 1.0){
                self.product.variables.time.set(time);
            } else {
                this._redraw(
                    function() {
                        self.options.messageBus.send('chart-time-select', self.time);
                    },
                    new ChangeType().set('time')
                );
            }
        },

        _viewport: function() {
            // Note that getting the size from the current DOM is slow and unreliable and should not be used except at startup.
            // Use product.get_view() or _promises.size instead.

            var top = 0;
            var left = 0;
            var width = this.element.width();
            var height = this.element.height();
            var view_width = this.element.width();
            var view_height = this.element.height();

            var img = this.product.get_view({width: width, height: height});

            top = img.top;
            left = img.left;
            width = img.width
            height = img.height

            if (typeof top === 'string') {
                top = parseFloat(top);
            }

            if (typeof left === 'string') {
                left = parseFloat(left);
            }

            if (typeof height === 'string') {
                height = parseFloat(height);
            }

            if (typeof width === 'string') {
                width = parseFloat(width);
            }

            if (typeof view_width === 'string') {
                view_width = parseFloat(view_width);
            }

            if (typeof view_height === 'string') {
                view_height = parseFloat(view_height);
            }

            if (top < 0) top = -top;
            if (left < 0) left = -left;

            return {
                top: -top,
                left: -left,
                width: width,
                height: height,
                view_height: view_height,
                view_width: view_width
            };
        },

        _rect: function() {
            return to_rect(this._viewport());
        },

        /*
         *  If no size is specified it will be queried from the DOM
         *  If one dimension is specified, the existing value will be used for the other dimension.
         */
        resize: function(size, done) {
            var self = this;
            var _size = {};
            this._resize_orig = this.resize;

            this.resize = function(size, done) {
                size || (size = {});
                size.width || (size.width = self.element.width());
                size.height || (size.height = self.element.height());
                // size.width and size.height may be 0 here

                if ((size.width && (size.width !== _size.width)) || (size.height && (size.height !== _size.height))) {

                    size.width && (_size.width = size.width);
                    size.height && (_size.height = size.height);

                    if(_size.width && _size.height){
                        var left = 0;
                        var top = 0;
                        var right = left + _size.width;
                        var bottom = top + _size.height;
                        self.clip.css({clip: 'rect(' + top + 'px,' + right + 'px,' + bottom + 'px,' + left + 'px)'});

                        this._promises.size.resolve(_size);
                    }

                    if (self.options.expand_on_resize) {
                        var port = this._viewport();
                        self.product.projection(self.product.resize(port.width, port.height));
                    }

                    _size.width && _size.height && self._clear()._redraw(done, new ChangeType().set('size'));
                }
            }
            return this.resize(size, done);
        },

        _clear: function(done) {
            var self = this;

            self.animator && self.animator().clear();

            return this;
        },

        _redraw: function(done, _change_type) {
            var self = this, o = self.options;

            if(self.__redraw_queued) return;
            self.__redraw_queued = true;

            self._promises.size.done(function(size) {

                var change_type = _change_type || new ChangeType().set('all');

                delete self.zoom.view;
                self.times = null;

                self.product.get_chart({
                    json: {
                        product: self.product.as_json(size, self.time, self.mail, o.format),
                        animate: null
                    },
                    name: self.name,
                    tag: self.name,
                    cache: self.time ? 100000000 : undefined,
                    status: Spinner(self.clip),
                    priority: o.priority,
                    img: self.img,
                    logErrors: 'Get map',
                    statistics: true,
                    success: function(json, time) {

                        // This can happen during animations
                        if (!self.img.is(":visible")) {
                            self.clip.find('img').hide();
                            self.img.show();
                        }

                        __DEV__ && self.in_destroy && console.warn('chart._redraw: in destroy.');

                        self._bind_image();

                        //remove obsolete messages:
                        $('body > .inform-dialogue').remove();

                        self.time = time;

                        if (!self.product) {
                            return;
                        }

                        /*
                         *   Previously the image position was managed not by the redraw but by the
                         *   transformer. With rounding enabled, what assumptions can be made about
                         *   when rounding offsets change? The rounding is guaranteed to be the same
                         *   if the projection is unchanged.
                         */
                        /*
                        var viewport = self.product.get_view(json.result.magics);
                        self.img.css({top: -viewport.top, left: -viewport.left});
                        */

                        self.img.attr('name', self.product.name);

                        self.product.availability(self.time, function(avail) {
                            if (avail) {
                                if (avail.valid_time.index === undefined) {
                                    var dates = avail.valid_time.values;
                                    var current = avail.current;
                                    var nearest = current.nearest(dates);
                                    console.error('DATE MISMATCH', current, nearest);
                                }
                                o.messageBus.send('chart-product-availability', avail);
                            }
                        });

                        // Check if the time as been changed by the backend
                        if (json.in_time && json.out_time) {
                            var in_time = new Time(json.in_time);
                            var out_time = new Time(json.out_time);

                            if (!in_time.equals(out_time)) {

                                o.messageBus.send('chart-product-availability-mismatch', {
                                    current: in_time,
                                    nearest: out_time
                                });

                                var text = '<b>Not all the layers are available for the requested base time and step.</b>' + '\n\nThe base time and step have been automatically adjusted from {in_time} to {out_time}.\n\nYou may want to {undo} the last change.';

                                ECMWF.inform(self.element, {
                                    offset_top: 50,
                                    text: text,
                                    auto_dismiss: 30,
                                    width: Math.max(100, Math.min(500, self.element.width() - 60)),
                                    titles: {
                                        //dismiss : "OK",
                                        in_time: in_time.asText(),
                                        out_time: out_time.asText()
                                    },
                                    classes: {
                                        in_time: 'highlight',
                                        out_time: 'highlight'
                                    },
                                    actions: {
                                        undo: function() {
                                            Undo.last();
                                        }
                                    }
                                });
                            }
                        }

                        if(self.product.api > 1.0){
                            json.axis && $.each(json.axis, function(i, v) {
                                var param = self.product.variables[v.name];
                                param.is_subscribed(self) || param.subscribe(self, function(param) {
                                    console.log('--> ' + this.name + '=', param);
                                    self._clear()._redraw();
                                });

                                if (param.val_type === 'string' && parameter_changed(param, v)) {
                                    param.values = v.values;
                                    param.labels = v.labels;
                                    change_type.set('parameters');
                                }
                            });
                        }

                        if (self.product.data.product.widgets && !!~$.inArray(self.current_tool, ['select', null])) {
                            var mode = self.product.has_widget('clickable')
                                ? ['select', {cursor: 'pointer'}]
                                : [null, {cursor: 'default'}];

                            (mode[0] === self.current_tool) || self._set_mode.apply(self, mode);
                        }

                        o.messageBus.send('chart-content-update', {
                            product: self.product,
                            time: self.time,
                            changed: change_type
                        });

                        o.title.html(self.product.title(self.time));

                        o.messages && $.logMessage('Chart updated ' + (self.times ? self.times : ''));

                        if (done) done();

                        function parameter_changed(param, v) {
                             if (v.values.length !== param.values.length) return true;
                             for (var i in v.values) {
                                 if (v.values[i] !== param.values[i]) return true;
                             }
                             return false;
                        }
                    },

                    error: function(error) {
                        self.matrix.reset();

                        var text = '<h4>The chart could not be produced</h4>{error}\nYou may choose to {retry}, reset the {projection} ' +
                            (Undo.length() ? ', {revert} to the previous state ' : '') +
                            'or {clear} the map.';

                        ECMWF.inform(self.element, {
                            offset_top: 50,
                            level: 'error',
                            text: text,
                            width: Math.max(100, Math.min(500, self.element.width() - 60)),
                            actions: {
                                retry: function() {
                                    self._clear()._redraw(done);
                                },
                                projection: function() {
                                    self.options.messageBus.send('tool-full-map', {});
                                },
                                clear: function() {
                                    self.options.messageBus.send('tool-clear-click', {});
                                },
                                revert: function() {
                                    Undo.last();
                                }
                            },
                            titles: {
                                error: $.formatError(error)
                            },
                            classes: {
                                error: 'error'
                            }
                        });
                    }
                });

                self.__redraw_queued = false;
            });
        },

        _destroy: function() {
            var self = this, o = self.options;

            self.in_destroy = true;

            clearInterval(self.interval);

            o.messageBus.forget_me(self);

            o.show_latlong && self.latlong.remove();

            // warning! errors in these destroy methods are not reported!
            $.each(self.plugins, function(i, plugin){ plugin.destroy && plugin.destroy(); });
            $.each(self.plugins_v2, function(i, plugin){ plugin.destroy && plugin.destroy(); });

            this._mouseDestroy();
            self._clear();
            if (o.enable_undo) $.undoRemoveAll(this);

            self.clip.remove();

            if (self.product) {
                self.product.destroy(); // makes no difference to memory usage.
                self.product = null;
            }

            this.resize = this._resize_orig;

            $('body > .inform-dialogue').remove();
        },

        _listen: function(name, fn) {
            var self = this,
                o = self.options;

            //this creates a circular reference as the bus is a property of self
            o.messageBus.listen(self, name, fn);
            return self;
        },

        _constrain: function(pt) {
            var self = this;

            var pos = self.img.offset();

            var view = this._viewport();

            pt.x = Math.max(pt.x, pos.left);
            pt.y = Math.max(pt.y, pos.top);

            pt.x = Math.min(pt.x, pos.left + parseInt(view.width, 10));

            pt.y = Math.min(pt.y, pos.top + parseInt(view.height, 10));

            pt.x = Math.floor(pt.x + 0.5);
            pt.y = Math.floor(pt.y + 0.5);

        },

        aspect: function(start, end, port) {
            return;

            /*
            var self = this,
            o = self.options;

            var view = port ? port : this._viewport();
            var width = Math.abs(start.x - end.x);
            var height = Math.abs(start.y - end.y);

            var r = Math.min(width / view.width, height / view.height);

            width = Math.floor(view.width * r + 0.5);
            height = Math.floor(view.height * r + 0.5);

            end.x = (end.x > start.x) ? start.x + width : start.x - width;
            end.y = (end.y > start.y) ? start.y + height : start.y - height;
            */
        },

        _helper: function() {
            var self = this;

            if (!self.drag.helper) self.drag.helper = $('<div>').css({
                zIndex: 2,
                borderStyle: 'dotted',
                background: 'white',
                opacity: 0.7,
                borderWidth: '1px',
                position: 'absolute'
            }).appendTo($('body'));
            return self.drag.helper;
        },

        _mouseStart: function(event) {
            var self = this,
                o = self.options;

            if (!o.editable) return;

            self._clear();

            self.mouse_tool = event.shiftKey ? 'zoom' : self.current_tool;

            self.drag = {};

            return Utils.run.apply(self, [self._mouse[self.mouse_tool].start, [event]])[0];
        },

        _clearRequestQueue: function() {
            Ajaxqueue.cancelByTag(this.name);
        },

        /* Gesture for the wall controller */
        dragStartEvent: function(ev) {
            var self = this;
            if (self._pinch_gesture) {
                return;
            }
            self._drag_gesture = true;
            self.current_tool = 'pan';
            self._mouseStart(ev.center);
        },
        dragEvent: function(ev) {
            var self = this;
            if (!self._drag_gesture) {
                return;
            }
            self.current_tool = 'pan';
            self._mouseDrag(ev.center);
        },
        dragEndEvent: function(ev) {
            var self = this;
            if (!self._drag_gesture) {
                return;
            }
            self._drag_gesture = false;
            self.current_tool = 'pan';
            self._mouseStop(ev.center);
        },

        pinchStartEvent: function(ev) {
            var self = this;
            if (self._drag_gesture) {
                return;
            }
            self._pinch_gesture = true;
            console.log('pinchStartEvent');

            self.matrix.pinch_magics = self.matrix.magics;
            self.matrix.pinch_screen = self.matrix.screen;
        },

        pinchEvent: function(ev) {
            var self = this;

            if (self._drag_gesture) {
                return;
            }

            if(!self._pinch_gesture) {
                self.pinchStartEvent(ev);
            }


            self.pinch(ev.scale);
        },

        pinchEndEvent: function(ev) {
            var self = this;
            if (!self._pinch_gesture) {
                return;
            }
            self._pinch_gesture = false;
            console.log('pinchEndEvent');
        },

        /*===========================*/

        _mouseStart_pan: function(event) {
            var self = this;

            self._clearRequestQueue();

            var p = self.img.position();
            self.drag.start = {
                x: event.pageX,
                y: event.pageY
            };
            self.drag.position = {
                top: Math.floor(p.top),
                left: Math.floor(p.left)
            };
            self.latlong.hide();
            return true;
        },

        _mouseStart_zoom: function(event) {
            var self = this;

            self._clearRequestQueue();

            self.drag.start = {
                x: event.pageX,
                y: event.pageY
            };
            this._helper().show();
            return true;
        },

        _mouseStart_select: function(event) {
            return true;
        },

        _mouseStart_vertical: function(event) {
            return true;
        },

        _mouseDrag: function(event) {
            var self = this,
                o = self.options;

            if (!o.editable) return;

            return Utils.run.apply(self, [self._mouse[self.mouse_tool].drag, [event]])[0];
        },

        _mouseDrag_pan: function(event) {
            var self = this;
            var duration;

            var start = self.drag.start;
            var end = {
                x: event.pageX,
                y: event.pageY
            };

            self.img[duration ? 'animate' : 'css'](
                {
                    top: self.drag.position.top + end.y - start.y,
                    left: self.drag.position.left + end.x - start.x
                },
                duration
            );

            return false;
        },

        _mouseDrag_zoom: function(event) {
            var self = this;

            var start = self.drag.start;
            var end = {
                x: event.pageX,
                y: event.pageY
            };
            this._constrain(end);
            this.aspect(start, end);

            self.drag.helper.css({
                top: Math.min(start.y, end.y),
                left: Math.min(start.x, end.x),
                width: Math.abs(start.x - end.x),
                height: Math.abs(start.y - end.y)
            });
            self.latlong.latlong('update', event);
            return false;
        },

        _mouseDrag_select: function(event) {
            return false;
        },

        _mouseDrag_vertical: function(event) {
            return false;
        },

        _mouseStop: function(event) {
            var self = this,
                o = self.options;

            if (!o.editable) return;

            var proc = '_mouseStop_' + self.mouse_tool;

            this[proc](event);

            self.mouse_tool = null;
            return false;
        },

        _mouseStop_pan: function(event) {
            var self = this;

            var start = self.drag.start;
            var end = {
                x: event.pageX,
                y: event.pageY
            };

            var tm = $M([
                [1, 0, -(end.x - start.x)],
                [0, 1, -(end.y - start.y)],
                [0, 0, 1]
            ]);

            var ts = $M([
                [1, 0, (end.x - start.x)],
                [0, 1, (end.y - start.y)],
                [0, 0, 1]
            ]);

            self.matrix.magics = self.matrix.magics.multiply(tm);
            self.matrix.screen = self.matrix.screen.multiply(ts);

            self.matrix.apply();

            self.latlong.show();
            return false;
        },

        _mouseStop_zoom: function(event) {
            var self = this;

            var start = self.drag.start;
            var end = {
                x: event.pageX,
                y: event.pageY
            };

            this._constrain(end);
            this.aspect(start, end);

            var pos = self.img.offset();

            this.zoomRectangle(
                Math.min(start.x - pos.left, end.x - pos.left), Math.min(start.y - pos.top, end.y - pos.top), Math.max(start.x - pos.left, end.x - pos.left), Math.max(start.y - pos.top, end.y - pos.top)
            );

            self.drag.helper.hide();

            return false;
        },

        _mouseStop_select: function(event) {
            return false;
        },

        _mouseStop_vertical: function(event) {
            return false;
        },

        _setData: function(key, value) {
            $.widget.prototype._setData.apply(this, arguments);
        },

        _notifyProductChange: function() {
            var self = this,
                o = self.options;

            o.messageBus.send('chart-product-change', {
                product: self.product
            });
        },

        _set_cycle: function(cycle) {
            var self = this;

            console.warn('_set_cycle', arguments);

            /*
            if (self.cycle) {
                self._push_undo();
            }
            */

            console.warn('_set_cycle', cycle);
            self.cycle = cycle;

            self.product.cycle(cycle);
            self._clear()._redraw(function() {
                self._notifyProductChange();
            });
        },

        setProductAndProjection: function(product, projection) {
            var self = this;

            self._push_undo();
            if (product) {
                self.product.product(product);
            }
            if (projection) {
                self.product.projection(projection);
            }
            self._clear()._redraw(function() {
                self._notifyProductChange();
            });
        },

        redraw: function() {
            var self = this;
            self.in_destroy || self._redraw(function() {
                self._notifyProductChange();
            });
        },

        setProduct: function(x, use_product_projection, done) {
            var self = this;

            self._push_undo();

            self.product.product(x, use_product_projection);
            self._clear()._redraw(function() {
                if (done) {
                    done();
                }
                self._notifyProductChange();
            });
        },

        setZoomLevel: function(level) {
            var self = this;
            var lvl = self.zoom_level;
            self.zoom_level = level;
            self.zoomScale(level / lvl);
        },

        addLayers: function(names) {
            var self = this;
            self._push_undo();
            self.product.addLayers(names);
            self._clear()._redraw(function() {
                self._notifyProductChange();
            });
        },

        setProjection: function(x, done) {
            var self = this;
            this._push_undo();
            self.product.projection(x);
            this._clear()._redraw(done || $.noop);
            return this;
        },

        mouseWheel: function(event, delta) {
            var self = this,
                o = self.options;

            if (!o.editable) {
                return;
            }

            self._promises.size.done(function(viewport) {
                var offset = self.img.offset();

                var x = event.pageX - offset.left;
                var y = event.pageY - offset.top;

                // For now, center the zoom
                x = viewport.width / 2;
                y = viewport.height / 2;

                if (!self.zoom.view) {
                    self._clearRequestQueue();
                    self._clear();
                    self.zoom.view = viewport;
                    self.img.css(self.zoom.view);

                    self.matrix.delta = 0;
                    self.matrix.pinch_magics = self.matrix.magics;
                    self.matrix.pinch_screen = self.matrix.screen;
                }

                self.matrix.delta += delta;

                self.matrix.transform(x, y, 1 + self.matrix.delta / 10);

                Utils.run.apply(self, [self._mouse.wheel, [event, delta]]);
            });

            return this;
        },

        pinch: function(scale) {
            var self = this;

            self._promises.size.done(function(viewport) {
                var offset = self.img.offset();

                var x = /* event.pageX */ -offset.left;
                var y = /* event.pageY */ -offset.top;

                // For now, center the zoom
                x = viewport.width / 2;
                y = viewport.height / 2;

                if (!self.zoom.view) {
                    self._clearRequestQueue();
                    self._clear();
                    self.zoom.view = viewport;
                    self.img.css(self.zoom.view);
                }

                self.matrix.transform(x, y, scale);

                Utils.run.apply(self, [self._mouse.pinch, [scale]]);
            });

            return this;
        },

        _push_undo: function() {
            var self = this,
                o = self.options;

            if (!o.enable_undo) return;

            var save = self.product.get_save_state(true);
            var time = self.time;

            Undo.store(self, function() {

                var revert = self.product.get_save_state(true);
                var now = self.time;

                self.time = time;
                self.product = ProductFactory({owner: self, product: save});

                self._clear()._redraw(function() {
                    self._notifyProductChange();
                });

                // Return "redo" function
                return function() {

                    self.time = now;
                    self.product = ProductFactory({owner: self, product: revert});
                    //self.product.reset();
                    self._clear()._redraw(function() {
                        self._notifyProductChange();
                    });
                };
            });
        },

        getState: function(full) {
            var self = this;
            return self.product.get_save_state(full);
        },

        setState: function(state) {
            var self = this;
            self.product = ProductFactory({owner: self, product: state || {}});
            self._clear()._redraw(function() {
                self._notifyProductChange();
            });
        },

        getTime: function() {
            return this.time;
        },

        zoomRectangle: function(x0, y0, x1, y1) {
            var self = this;

            this._push_undo();

            self.product.zoom(x0, y0, x1, y1);
            self._transform_redraw();

            return this;
        },

        zoomScale: function(a) {
            var self = this;
            var viewport = self._viewport();

            this._push_undo();

            // zoom
            var m = $M([
                [1 / a, 0, 0],
                [0, 1 / a, 0],
                [0, 0, 1]
            ]);

            // Translate
            var t1 = $M([
                [1, 0, -viewport.width / 2],
                [0, 1, -viewport.height / 2],
                [0, 0, 1]
            ]);

            var t2 = $M([
                [1, 0, viewport.width / 2],
                [0, 1, viewport.height / 2],
                [0, 0, 1]
            ]);

            self.transformMap(t2.x(m).x(t1));

            return this;
        },

        verticalLongitude: function(lon) {

            var self = this;

            this._push_undo();

            self.product.vertical_longitude(lon);
            this._clear()._redraw();
        },

        panMap: function(dx, dy) {
            var self = this;

            this._push_undo();

            self.product.pan(dx, dy);
            this._clear()._redraw();

            return this;
        },

        _transform_redraw: function() {
            var self = this;

            this._clear()._redraw(function(){
                self.matrix.reset();
                self.img.css($.extend({top: 0, left: 0}, self.product.size()));
            });
        },

        transformMap: function(matrix) {
            var self = this;

            this._push_undo();

            self.product.transform(matrix);
            self._transform_redraw();

            return this;
        },

        showFrame: function(time, img) {
            var self = this,
                o = self.options;

            self.time = time;
            self._show_frame(img);
            o.messageBus.send('chart-time-select', self.time);
        },

        makeFrame: function() {
            var self = this;

            return $('<img>')
                // .hide()
                .addClass('frame')
                .appendTo(self.clip)
                .click(function(event) {
                    return self._on_click(event);
                });
        },

        fetchFrame: function(time, img, success, error) {
            var self = this,
                o = self.options;

            self.product.get_chart({
                json: {
                    product: self.product.as_json(this._viewport(), time, undefined, o.format),
                    animate: null
                },
                cache: 24 * 60 * 60,
                img: img,
                logErrors: 'Load frame',
                statistics: true,
                success: success,
                error: error
            });
        },

        notifyNotLatest: function(no_latest) {
            var self = this,
                o = self.options;

            if (o.show_title_overlay) {
                self.title_overlay.pulsate(no_latest);
            }
        },

        showLatestCycle: function() {
            var self = this;

            self.product.clear_availability();
            self._set_time(null);
        },

        _get_availability: function() {
            var self = this, o = self.options;

            self.product.newAvailability(function(updated, avail) {
                o.messageBus.send('chart-product-new-availability', {
                    updated: updated,
                    availability: avail
                });
                $.logInfo('Product availability updated at ' + new Date(updated * 1000));

                self.product.availability(self.time, function(avail) {

                    if (avail && avail.valid_time.index !== undefined) {
                        o.messageBus.send('chart-product-availability', avail);
                    }
                });
            });
        },

        getAvailability: function() {
            var self = this
            setTimeout(function(){ self._get_availability(); }, 3000);
            return this;
        },

        checkAvailability: function(on) {
            var self = this, o = self.options;

            if (on) {
                if (self.interval) {
                    //__DEV__ && console.warn('CHART checkAvailability already active:', self.name);
                    self.interval = clearInterval(self.interval);
                }

                self.interval = setInterval(function(){ self._get_availability(); }, 60 * 1000);
            } else {
                self.interval = clearInterval(self.interval);
            }

            o.availability = on;

            return this
        },

        _show_frame: function(img) {
            var self = this;

            // hide and update the image that was previously visible
            // this.frame_prev = this.frame_prev || this.img;
            // this.frame_prev.hide();
            // this.frame_prev = img.show();
            this.frame_prev = img;

            // hide _all_ images, including the one that should be shown
            self.clip.find('img').hide();

            img.show();
            if (self.title) {
                self.title.html(self.product.title(self.time));
            }
        },

        _set_mode: function(mode, cursor) {
            var self = this,
                o = self.options;
            this.img.css(cursor);
            this.current_tool = mode;
            o.messageBus.send('chart-mode', mode);
        },

        _bind_image: function() {
            var self = this,
                o = self.options;
            if (o.touch) {
                self.img.off().on('touchstart', function(e) {
                    var te = e.originalEvent;
                    te.preventDefault();

                    if (te.touches.length === 3) {
                        te.preventDefault();
                        self.options.messageBus.send('tool-full-map', {});
                        return false;
                    }

                    if (te.touches.length === 1) {
                        self.touch = {
                            pageX: te.touches[0].pageX,
                            pageY: te.touches[0].pageY
                        };
                        return self._mouseStart(self.touch);
                    }
                    return false;

                }).on('touchmove', function(e) {
                    var te = e.originalEvent;
                    te.preventDefault();
                    if (te.touches.length === 1) {
                        self.touch = {
                            pageX: te.touches[0].pageX,
                            pageY: te.touches[0].pageY
                        };
                        return self._mouseDrag(self.touch);
                    }
                    return false;
                }).on('touchend', function(e) {
                    var te = e.originalEvent;
                    te.preventDefault();
                    // All fingers are up
                    if (te.touches.length === 0 && self.touch) {
                        var x = self._mouseStop(self.touch);
                        delete self.touch;
                        return x;
                    }
                    return false;
                });

                if (o.gesture) {
                    self.img.on('gesturestart', function(e) {
                        var te = e.originalEvent;
                        te.preventDefault();
                        return false;
                    })

                    .on('gesturechange', function(e) {
                        var te = e.originalEvent;
                        te.preventDefault();
                        self.pinch(te.scale);
                        return false;
                    }).on('gestureend', function(e) {
                        var te = e.originalEvent;
                        te.preventDefault();
                        return false;
                    });
                }
            } else {

                self.img.off().mousedown(function(event) {
                    event.preventDefault();
                }).mouseup(function(event) {
                    event.preventDefault();
                }).mousemove(function(event) {
                    o.show_latlong && self.latlong.latlong('update', event);
                    event.preventDefault();
                }).mouseenter(function(event) {
                    self.latlong.show();
                }).mouseleave(function(event) {
                    if (!self.mouse_tool) {
                        self.latlong.hide();
                    }
                }).click(function(event) {
                    return self._on_click(event);
                }).on('dblclick', function(event) {
                    var o = self.img.offset();
                    var x = event.pageX - o.left;
                    var y = event.pageY - o.top;

                    var viewport = self.product.get_view();

                    var dx = x - viewport.width / 2;
                    var dy = y - viewport.height / 2;

                    self.panMap(dx, dy);
                });
            }
        },

        _on_click: function(event) {
            var self = this,
                o = self.options;

            if (self.current_tool === 'select') {
                var offset = self.img.offset();
                var x = event.pageX - offset.left;
                var y = event.pageY - offset.top;

                self.options.messageBus.send('latlon-location-select', $.extend({},
                    self.product.position(x, y), {
                        pageX: event.pageX,
                        pageY: event.pageY,
                        x: x,
                        y: y,
                        from: 'chart'
                    }
                ));
            }
            if (self.current_tool === 'vertical') {
                o = self.img.offset();
                x = event.pageX - o.left;
                y = event.pageY - o.top;

                var pos = self.product.position(x, y);
                pos.pageX = event.pageX;
                pos.pageY = event.pageY;
                pos.x = x;
                pos.y = y;

                self.options.messageBus.send('vertical-longitude-select', pos);
            }
        },

        /*
         *  Instantiate a plugin on a running chart instance
         */
        add_plugin: function(name, plugin) {
            if(!$.findWhere(this.plugins, {name: name})){
                var plugin = new chart_plugins[name](this);
                plugin && (plugin.name = name, this.plugins.push(plugin));
                return true;
            }
            return false;
        }
    };

    // touch can be set even if main input device is mouse (eg Chrome 5 on Windows), so mouse is always initialised for now.
    var touch = false && !! ($.fn.browserTouchSupport.touches && (navigator.userAgent.match(/iPad/i) !== null));
    var gesture = false && !! ($.fn.browserTouchSupport.gestures);

    if (touch) {
        $.widget('ui.chart', ECMWF.Chart.Prototype);
    } else {
        $.widget('ui.chart', $.ui.mouse, ECMWF.Chart.Prototype);
    }

    var chart_plugins = {};
    $.chartRegisterPlugin = function(name, plugin) {
        chart_plugins[name] = plugin;
    };

    $.chartGetPlugins = function() {
        return chart_plugins;
    };

    function ChartRenderer() {
        this.set_img = function(img) {};
    }

    function DomChartRenderer(img, get_size, get_viewport) {
        this.img = img;

        this.set_img = function(url) {
            this.img
                .attr('src', url)
                .css($.extend({}, get_viewport(), get_size()));
        };
    }
    DomChartRenderer.prototype = new ChartRenderer();

    function ChangeType() {
        this.set = function(change_type) {
            this['_' + change_type] = true;
            return this;
        };
        this.get = function(change_type) {
            return this._all || this['_' + change_type];
        };
    }

    function to_rect(c) {
        return {
            top: c.top,
            left: c.left,
            bottom: c.top + c.height,
            right: c.left + c.width
        };
    }

    function from_rect(c) {
        return {
            top: c.top,
            left: c.left,
            width: c.right - c.left,
            height: c.bottom - c.top
        };
    }
});

