/*
 *  (C) Copyright ECWMF
 *
 *  This software is licensed under the terms of the Apache Licence Version 2.0
 *  which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 *  In applying this licence, ECMWF does not waive the privileges and immunities
 *  granted to it by virtue of its status as an intergovernmental organisation nor
 *  does it submit to any jurisdiction.
 *
 *  ECMWF Chart parameter
 *  ---------------------
 *
 *  Three types of observable chart parameters are provided for the following value types:
 *     1) a simple string value
 *     2) an ECMWF Time value chosen from the Availability list.
 *     3) a 'dynamic' value - TODO DOCUMENT ME
 */

define(['jquery', 'observable', 'time'], function($, Observable, Time) {

	/*
	 *   Note that before availability is retrieved, this parameter will have an empty value.
	 */
	var _Time = function(time) {
		this.name = 'time';
		this.title = 'Time';
		this.type = 'time';
		this.val_type = 'time';
		this._listeners = []; // note the _listeners object in the prototype is shared and should not be used by Parameters.

		this.set = function(value, availability) {
			// TODO use this.time_factory(value);
			if(typeof value === 'string'){
				value = this.from_string(value);
			}
			if(value.constructor === Array){
				value = new Time(value);
			}

			if(availability){
				var previous_availability = this.availability;
				this.availability = availability;
				this.previous_availability = previous_availability;
				this.availability_changed = this._availability_changed();
			}
			else this.availability_changed = false;

			if(value && this.availability){
				if(typeof value.index(this.availability) === 'undefined'){
					console.warn('cannot set time: not valid for this product', [value.base_time.format('yyyymmddHH'), value.step, value.valid_time.format('yyyymmddHH')]);
					return;
				}
			}

			// handle the case where the list has changed but the value has not
			// TODO handle list changes through a separate mechanism (to prevent duplicate requests)
			if(this.previous_availability && this.availability_changed && this.is_current(value)){
				this.emit(this.value);
			}

			return _Time.prototype.set.call(this, value);
		}

		/*
		 *	 Set next valid time
		 */
		this.next = function() {
			var self = this;

			var values = $.map(this.availability, function(v){ if(v.base_time.getTime() === self.value.base_time.getTime()) return v; });
			var index = this.value.index(values);

			(index < values.length - 1) && this.set(values[index + 1]);
		}

		/*
		 *	 Set previous valid time
		 */
		this.previous = function() {
			var self = this;

			var values = $.map(this.availability, function(v){ if(v.base_time.getTime() === self.value.base_time.getTime()) return v; });
			var index = this.value.index(values);

			(index > 0) && this.set(values[index - 1]);
		}

		/*
		 *	 Set next base time
		 */
		this.next_basetime = function() {
			var next = this.value.nextBaseTime(this.availability);
			next && this.set(next)
		}

		/*
		 *	 Set previous base time
		 */
		this.previous_basetime = function() {
			var prev = this.value.previousBaseTime(this.availability);
			prev && this.set(prev)
		}

		this.next_day = function() {
			var next_day = this.value.valid_time.getTime() + 1000 * 60 * 60 * 24;
			for (var next = this.value; next; next = next.next(this.availability)) {
				if (next.valid_time.getTime() >= next_day) {
					this.set(next);
					break;
				}
			}
		}

		this.previous_day = function() {
			var prev_day = this.value.valid_time.getTime() - 1000 * 60 * 60 * 24;
			for (var prev = this.value; prev; prev = prev.previous(this.availability)) {
				if (prev.valid_time.getTime() <= prev_day) {
					this.set(prev);
					break;
				}
			}
		}

		this.first_step = function() {
			var time = this.value.first(this.availability);
			time && this.set(time);
		}

		this.last_step = function() {
			var time = this.value.last(this.availability);
			time && this.set(time);
		}

		this.is_current = function(val) {
			return !!(this.value && (val.key === this.value.key));
		}

		this.to_string = function(val) {
			return this.value
				? ((this.availability && this.availability.length === 1)
					? this.value.base_time.format('yyyymmddHH')
					: [this.value.base_time.format('yyyymmddHH'), this.value.step, this.value.valid_time.format('yyyymmddHH')].join(','))
				: '';
		}

		this.from_string = function(str) {
			var a = str.split(',');
			if(a && (a.length === 1)){ // monthly forecasts without availability
				$.merge(a, [0, a[0]]);
			}
			if(a && (a.length === 3)){
				a[1] = +a[1];
				return new Time(a);
			}
			return str;
		}

		this.for_json = function() {
			return this.value && this.value.base_time ? [this.value.base_time.format('yyyymmddHH'), this.value.step, this.value.valid_time.format('yyyymmddHH')] : null;
		}

		this._availability_changed = function() {
			if(this.availability){
				if(!this.previous_availability || (this.previous_availability.length !== this.availability.length)) return true;

				for(var i in this.availability){
					if(this.availability[i].key !== this.previous_availability[i].key){
						return true;
					}
				}
			}
			return false;
		}

		/*
		 *  Find the best Time to use for the given base_time. Adjust the valid_time if needed.
		 */
		this.find_time_by_base_time = function(time) {
			var self = this;
			//console.log('find_time_by_base_time: looking for: ', [time.base_time.format('mmdd:HH'), time.valid_time.format('mmdd:HH')]);

			function closest() {
				for(var i in self.availability){
					if(self.availability[i].base_time.getTime() === time.base_time.getTime()){
						return self.availability[i];
					}
				}
			}

			var exact_match = $.findWhere(self.availability, {key: time.key});
			//if(exact_match) console.log('find_time_by_base_time: exact_match: ', [exact_match.base_time.format('mmdd:HH'), exact_match.step, exact_match.valid_time.format('mmdd:HH')]);

			return exact_match || closest();
		}

		this.index = function() {
			return this.value.index(this.availability);
		}

		this.by_month = function() {
			var self = this;
			var months = [];
			var month = [];

			function flush(){
				months.push(month);
				month = [];
			}

			var m_ = this.labels[0].split(' ')[2];
			$.each(this.values, function(i) {
				var s = self.labels[i].split(' ');
				var m = s[2];
				if(m !== m_){
					flush();
					m_ = m;
				}
				month.push(i);
			});
			flush();

			return months;
		}

		// TODO try and move this up into the Time constructor
		this.time_factory = function(time) {
			if(!time) return time;

			return typeof time === 'string'
				? this.from_string(time)
				: (time.constructor === Array
					? new Time(time)
					: time);
		}

		this.value = this.time_factory(time);
	}
	_Time.prototype = new Observable();

	var _String = function(v) {
		this.val_type = 'string';
		this._listeners = [];
		$.extend(this, v);
		(typeof this.value === 'undefined') && (this.value = this.values[0]);

		this.to_string = function(val) {
			return this.value;
		}

		this.for_json = function() {
			return (this.value + '').replace('=', '%3D');
		}
		
		this.set = function(value, values, labels) {
			if ((typeof value === 'number') || (this.values && typeof this.values[0] === 'number'))
				return console.error('string parameter cannot have non string value:', this.name), this;

			if(values) this.values = values;
			if(labels) this.labels = labels;

			if(this.values && $.inArray(value, this.values) < 0 ){
				console.warn('invalid parameter value: ', value, this.values);
				value = this.prev || this.values[0];
				return this;
			}

			return _String.prototype.set.call(this, value);
		}
	}
	_String.prototype = new Observable();

	var _Dynamic = function(v) {
		this.val_type = 'string';
		this._listeners = [];
		$.extend(this, v);
		(typeof this.value === 'undefined') && (this.value = this.values[0]);

		this.to_string = function(val) {
			return this.value;
		}

		this.for_json = function() {
			return (this.value + '').replace('=', '%3D');
		}
		
		this.set = function(value) {
			return _Dynamic.prototype.set.call(this, value);
		}
	}
	_Dynamic.prototype = new Observable();

	return {
		string: _String,
		time: _Time,
		dynamic: _Dynamic
	}
});
