/**
 *  ECMWF Bus
 *  (C) Copyright ECMWF http://www.ecmwf.int
 *
 *  This software is licensed under the terms of the Apache Licence Version 2.0
 *  which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 *  In applying this licence, ECMWF does not waive the privileges and immunities
 *  granted to it by virtue of its status as an intergovernmental organisation nor
 *  does it submit to any jurisdiction.
 */

define(['jquery', 'core/clone'], function($, clone) {

    'use strict';

    var count = 0;

    function Bus(name, elem, parent) {
        count += 1;
        this.parent = parent;
        this.id = count;
        this.past = {};
        this.elem = elem;
        this.name = name;
        this.active = true;
        this.listeners = new ListenerList();
    }

    $.extend(Bus.prototype, {

        enable: function() {
            this.active = true;
        },

        disable: function() {
            this.active = false;
        },

        /*
         *  'object' is a user reference that can be used when unbinding. It is otherwise not used internally.
         */
        listen: function(name, callback, object, no_past) {
            var self = this;

            if (typeof name !== 'string') {
                var tmp = name;
                name = callback;
                callback = object;
                object = tmp;
            }

            self.elem.on('bus-' + name, callback);

            //console.log("LISTEN TO [" , name, "] on [", self.name , "]");

            self.listeners.listeners.push([object, name, callback]);

            if (!no_past && self.past[name]) {
                console.log('PAST ' + name, self, self.past[name]);
                callback(null, self.past[name]);
            }

            if (self.parent) {
                // TODO: implement the unbind...
                self.parent.listen(name, callback, self);
            }

            return self;
        },

        listen_once: function(name, callback, object) {
            var self = this;

            if (typeof name !== 'string') {
                var tmp = name;
                name = callback;
                callback = object;
                object = tmp;
            }

            var cb = function _listen_once(a, b) {
                self.forget(name, cb);
                callback(a, b);
            };
            if (object) self.listen(object, name, cb);
            else self.listen(name, cb);

            return self;
        },

        forward: function(name1, name2, msg) {
            var self = this;

            self.listen(name1, function(e, m) {
                if ($.isFunction(msg)) {
                    self.send(name2, msg(m));
                } else {
                    self.send(name2, msg ? msg : m);
                }
            });
            return self;
        },

        forget_me: function(object) {
            var self = this;

            $.each(this.listeners.listeners, function(_, item) {
                if (item[0] === object) {
                    self.forget(item[1], item[2]);
                }
            });
            this.listeners.remove_object(object);
            return self;
        },

        forget: function(name, handler) {
            var self = this;

            if (!name) {
                console.trace();
                return console.warn('Bus.forget: name missing');
            }
            if (handler) {
                if (!handler.guid) console.warn('Bus.forget: bad handler arg: no guid: h=', handler); // possibly anonymous fn

                var n_handlers = function() {
                    var events = $._data(self.elem[0], 'events');
                    if (!events) return 0;
                    var my_events = events['bus-' + name];
                    return my_events ? my_events.length : 0;
                };

                var n = n_handlers();
                self.elem.off('bus-' + name, handler);

                if (n_handlers() === n) {
                    if(true === false) console.warn('Bus.forget: no handler unbound. event=' + name + ' n_handlers=' + n);

                    /* -- uncomment for debugging info
                    var events = $._data(self.elem[0], 'events');
                    var my_events = events['bus-' + name];
                    for(var i in my_events){
                        var _handler = my_events[i].handler;
                        console.log('   ', my_events[i], _handler ? _handler.guid : null, handler.guid);
                        if(_handler.guid === handler.guid){
                            console.log('           handler found');
                        }
                    }
                    */
                }
                this.listeners.remove_by_name(name, handler);
            } else {
                self.elem.off('bus-' + name, handler);
                if (self.past && self.past[name]) {
                    delete self.past[name];
                }
                this.listeners.remove_by_name(name);
            }
            if (this.parent) {
                //this is needed to free resources, however it may commonly remove handlers for more than the intended object if it has the same handler uid.
                this.parent.forget(name, handler);
            }
            return self;
        },

        forgetAll: function() {
            var self = this;
            var events = $._data(self.elem[0], 'events');
            for (var i in events) {
                self.elem.off(i);
                delete self.past[i];
            }
            this.listeners.clear();
        },

        send: function(name, msg) {
            var self = this;
            if (!this.active) return;

            var s;
            try {
                s = clone.serialise(msg);
                if (s && s.length > 80) {
                    s = s.substr(0, 77) + '...';
                }
            } catch (e) {
                s = e;
            }

            if (name !== 'chart-time-select' && name !== 'event-start') {
                console.info('BUS-SEND[' + (self.name || self.id) + '] ' + name + ' ' + (s ? s : ''));
            }


            self.past[name] = msg;

            self.elem.triggerHandler('bus-' + name, msg);

            return self;
        },

        // fn to use as an event handler to trigger a message send.
        sendfn: function(name) {
            var self = this;
            return function() {
                self.send(name);
            }
        },

        clear: function(name) {
            var self = this;

            if(name)
                self.past && self.past[name] && delete self.past[name];
            else
                self.past = {};
            return this;
        },

        destroy: function() {
            var self = this;

            self.forgetAll();
            delete self.past;
            self.elem.removeData('message-bus');
            self.elem = null;
            self.destroyed = true;

            if (self.parent) self.parent.forget_me(self);
        }
    });

	function ListenerList() {
		this.listeners = [];
		this.remove_object = function(object) {
			this.listeners = $.grep(this.listeners, function(a) {
				return a[0] !== object;
			});
		};
		this.remove_by_name = function(name, handler) {
			var filter = function(a) {
				return (a[1] !== name) || (a[2] !== handler);
			};
			if (handler) {
				this.listeners = $.grep(this.listeners, filter);
			} else {
				this.listeners = $.grep(this.listeners, function(a) {
					return a[1] !== name;
				});
			}
		};
		this.clear = function() {
			this.listeners = [];
		};
	}

	/*
	 *  elem:
	 *      - if not set, the GLOBAL bus is returned.
	 *      - if is a string, the bus attached to the dom element with that id is returned. A new element will be created if neccesary.
	 *      - if is an element, the bus for that element is returned.
	 *  parent_bus:
	 *      - optional bus object.
	 */
	var MessageBus = function(elem, parent_bus) {

		var tmp, bus, name = elem;

		if (typeof(elem) === 'string') {
			tmp = $('#' + elem);
			if (tmp.length) {
				elem = tmp;
			} else {
				elem = $("<div id='" + elem + "'/>").hide().appendTo('body');
			}
		}

		if (!elem) {
			name = 'GLOBAL';
			elem = $(document);
			parent_bus = null;
		} else {
			name = $(elem).attr('id');
		}

		bus = elem.data('message-bus');
		if (!bus) {
			bus = new Bus(name, elem, parent_bus);
			elem.data('message-bus', bus);
		}
		return bus;
	}

	$.fn.messageBus = function(parent_bus) {
		return MessageBus(this, parent_bus);
	};

	$.fn.bindBusMessage = function(bus, msg, action) {
		this.each(function() {
			var elem = $(this);
			bus.listen(msg, function(_, m) {
				elem.trigger(action, m);
			});
		});
		return this;
	};

	return MessageBus;
});
