/*jslint white: false, nomen: false */
/** @license
 *  ECMWF Time
 *  (C) Copyright ECMWF http://www.ecmwf.int
 *
 *  This software is licensed under the terms of the Apache Licence Version 2.0
 *  which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 *  In applying this licence, ECMWF does not waive the privileges and immunities
 *  granted to it by virtue of its status as an intergovernmental organisation nor
 *  does it submit to any jurisdiction.
 *
 */
/**
 *  The Time object represents the time specification for a chart.
 *
 *  Its main properties are:
 *      base_time   (UTCDate)
 *      valid_time  (UTCDate)
 *      step        (int)
 *
 *  Other properties:
 *      key         (string of form "201304080000_201304080300")
 *
 *  To instantiate:
 *      new Time(UTCDate base_time, int step, UTCDate valid_time)
 *      new Time(["201304080000", 21, "201304082100"])
 */

define(['jquery', 'utc-date', 'date-format'], function($, UTCDate) {

    function cycle2date(cycle) {

        if (cycle instanceof UTCDate) {
            return cycle;
        }

        var minute, hour, day, month, year;

        if (typeof(cycle) === 'string') {
            cycle = parseInt(cycle, 10);
        }

        if (cycle <= 9999999999) {
            cycle *= 100;
        }

        minute = cycle % 100;
        cycle = Math.floor(cycle / 100);
        hour = cycle % 100;
        cycle = Math.floor(cycle / 100);
        day = cycle % 100;
        cycle = Math.floor(cycle / 100);
        month = cycle % 100;
        cycle = Math.floor(cycle / 100);
        year = cycle;
        return new UTCDate(year, month - 1, day, hour, minute, 0);
    }

    function date2cycle(date) {
        return date.format('yyyymmddHHMM');
    }

    function date2title(date, include_year) {
        return include_year
            ? date.format('dddd d mmm yyyy, HH') + ' UTC'
            : date.format('dddd d mmm, HH') + ' UTC';
    }


    $.cycleToDate = cycle2date;
    $.dateToCycle = date2cycle;
    $.dateToTitle = date2title;


    function Time(a, b, c) {
        var base_time, step, valid_time;

        if (c !== undefined) {
            base_time = a;
            step = b;
            valid_time = c;
        } else if (a instanceof Array) {
            base_time = a[0];
            step = a[1];
            valid_time = a[2];
        } else {
            base_time = a.base_time;
            step = a.step;
            valid_time = a.valid_time;
            this.key = a.key;
        }

        this.base_time = cycle2date(base_time);
        this.step = step;
        this.valid_time = cycle2date(valid_time);

        var s = this;

        function make_key() {
            //optimised for speed. does not use Date.format()
            return s.base_time.yyyymmddHHMM() + '_' + s.valid_time.yyyymmddHHMM();
        }
        this.key = this.key || make_key();

        return this;
    }

    Time.prototype.equals = function(other) {
        return this.key === other.key;
    }

    Time.prototype.index = function(dates) {
        if (!dates) return -1;

        var self = this,
            result;

        $.each(dates, function(i, val) {
            if (val.key === self.key) {
                result = i;
                return false;
            }
        });
        //if (!result) console.error(this.key, $.map(dates, function(a) { return a.key; }));
        return result;
    };

    Time.prototype.nearest = function(dates) {
        var self = this,
            result = null,
            best, step;

        $.each(dates, function(i, val) {

            var delta = Math.abs(val.valid_time.getTime() - self.valid_time.getTime());
            if (result === null) {
                best = delta;
                step = val.step;
                result = val;
            } else {
                if (delta === best) {
                    if (val.step < step) {
                        best = delta;
                        step = val.step;
                        result = val;
                    }
                }

                if (delta < best) {
                    best = delta;
                    step = val.step;
                    result = val;
                }
            }
        });
        return result;
    }

    Time.prototype.previous = function(dates) {
        var self = this;

        var prev;
        $.each(dates, function(i, time) {
            if (time.base_time.getTime() === self.base_time.getTime()) {
                if ((time.valid_time.getTime() < self.valid_time.getTime()) && (!prev || time.valid_time.getTime() > prev.valid_time.getTime()))
                    prev = time;
            }
        });

        return prev;
    }

    Time.prototype.next = function(dates, wrap) {
        var self = this;

        var next;
        $.each(dates, function(i, time) {
            if (time.base_time.getTime() === self.base_time.getTime()) {
                if ((time.valid_time.getTime() > self.valid_time.getTime()) && (!next || (time.valid_time.getTime() < next.valid_time.getTime())))
                    next = time;
            }
        });

        if (wrap && !next) {
            next = self;
            $.each(dates, function(i, time) {
                if (time.base_time.getTime() === next.base_time.getTime()) {
                    if (time.valid_time.getTime() < next.valid_time.getTime()) {
                        next = time;
                    }
                }
            });
        }

        return next;
    }

    Time.prototype.first = function(dates) {
        var self = this;

        var first;
        $.each(dates, function(i, time) {
            if (time.base_time.getTime() === self.base_time.getTime()) {
                if (!first || (time.valid_time.getTime() < first.valid_time.getTime()))
                    first = time;
            }
        });

        return first;
    }

    Time.prototype.toJSON = function() {
        return {
            base_time: date2cycle(this.base_time),
            step: this.step,
            valid_time: date2cycle(this.valid_time)
        };
    }

    Time.prototype.last = function(dates) {
        var self = this;

        var last;
        $.each(dates, function(i, time) {
            if (time.base_time.getTime() === self.base_time.getTime()) {
                if (!last || (time.valid_time.getTime() > last.valid_time.getTime()))
                    last = time;
            }
        });

        return last;
    }

    Time.prototype.sameBaseTime = function(dates) {
        var self = this;

        var result = [];
        $.each(dates, function(i, time) {
            if (time.base_time.getTime() === self.base_time.getTime()) {
                result.push(time);
            }
        });

        return result;
    }

    Time.prototype.previousBaseTime = function(dates, is_step_fixed) {
        var self = this;

        var prev;
        $.each(dates, function(i, time) {
            if (time.valid_time.getTime() === self.valid_time.getTime()) {
                if ((time.base_time.getTime() < self.base_time.getTime()) && (!prev || time.base_time.getTime() > prev.base_time.getTime()))
                    prev = time;
            }
        });
        // Trying by fixing the step  10-days 15-daysi meteograms
        if (!prev && is_step_fixed) {
            $.each(dates, function(i, time) {
                if (time.step === self.step) {
                    if ((time.base_time.getTime() < self.base_time.getTime()) && (!prev || time.base_time.getTime() > prev.base_time.getTime()))
                        prev = time;
                }
            });
        }
        return prev;
    }

    Time.prototype.nextBaseTime = function(dates, is_step_fixed) {
        var self = this;

        var next;
        $.each(dates, function(i, time) {
            if (time.valid_time.getTime() === self.valid_time.getTime()) {
                if ((time.base_time.getTime() > self.base_time.getTime()) && (!next || time.base_time.getTime() < next.base_time.getTime()))
                    next = time;
            }
        });
        // Trying by fixing the step  10-days 15-daysi meteograms
        if (!next && is_step_fixed) {
            $.each(dates, function(i, time) {
                if (time.step === self.step) {
                    if ((time.base_time.getTime() > self.base_time.getTime()) && (!next || time.base_time.getTime() < next.base_time.getTime()))
                        next = time;
                }
            });
        }

        return next;
    }

    Time.prototype.bestBaseTime = function(dates) {
        var self = this, next;

        $.each(dates, function(i, time) {
            if (time.valid_time.getTime() === self.valid_time.getTime()) {
                if ((time.base_time.getTime() > self.base_time.getTime()) || (next && time.base_time.getTime() > next.base_time.getTime()))
                    next = time;
            }
        });

        if (!next && dates.length) {
            next = dates[0];
            $.each(dates, function(i, time) {

                if (time.base_time.getTime() > next.base_time.getTime()) {
                    next = time;
                }

                if (time.base_time.getTime() === next.base_time.getTime()) {
                    if (time.valid_time.getTime() < next.valid_time.getTime()) {
                        next = time;
                    }
                }
            });
        }

        return next;
    }

    Time.prototype.add = function(hours) {
        var verif = this.valid_time.addSeconds(hours * 60 * 60);
        return new Time(this.base_time, this.step + hours, verif);
    }

    Time.prototype.asText = function(valid) {
        if (valid) {
            return date2title(this.base_time) + ' T+' + this.step + " Valid: " + date2title(this.valid_time);
        } else {
            return date2title(this.base_time) + ' T+' + this.step;
        }
    }

    return Time;
});

