/*
 *  sizing
 *    -widget width is determined by the widgets' external container/manager.
 *    -widget height is determined by the widgets' content.
 *        This is set by specifying the css height of the 'widget-content' element.
 *        However, for performance reasons, the css should not be accessed directly and is stored as _height, accessed by the content_height() fn.
 *        Height changes are indicated to containers via the height-changed signal.
 */
define(['config', 'jquery', 'bus', 'user_info', 'utils', '../model', 'handler', 'throttle-debounce', 'pulsate'], function(ECMWF, $, Bus, User, Utils, Tabs) {

	var registry = {};

	var DashboardWidget = {
		registry: function() {
			return registry;
		},

		register: function(klass, widget) {
			registry[widget.factory.kind] = widget.factory;
			registry[widget.factory.kind].klass = klass;
			return widget;
		},

		findClass: function(kind) {
			var widget = registry[kind];
			return widget ? widget.klass : null;
		}
	}

	$.fn.findDashboardWidgetClass = function() {
		var klass;
		this.each(function() {
			var el = $(this);
			for (var i in registry) {
				if (el.hasClass(registry[i].klass)) klass = registry[i];
			}
		});
		__DEV__ && !klass && console.warn('findDashboardWidgetClass: not found! class="' + $(this).attr('class') + '"');
		return klass || registry['null'];
	}

	$.fn.findDashboardWidget = function (kind) {
		var widget;
		this.each(function () {
			var el = $(this);
			var widgetClass = DashboardWidget.findClass(kind);
			if(el.data('ui' + ECMWF.ucfirst(widgetClass))){
				widget = el[widgetClass];
				return false;
			}
		});
		return widget;
	}

	/*
	 *  emitted events:
	 *  	- 'content-change' - sent on local bus. Indicates a state change. Not to be used in connection with rendering changes. Currently this message has no listeners (Saving is done explicitly by the Widget by calling the dashboardset proxy).
	 *  	- 'contentChanged' - dom event. Also used as content-ready. Should not be used to initiate save(). Intended to be private.
	 *    - 'height-changed' - sent on local bus. This is the only rendering-related information that containers need be interested in. TODO verify that this is ok for transient changes such as when editing.
	 *
	 */
	$.widget('ui.dashboardWidget', $.ui.handler, {

		options: {
			name: null,
			title: 'Untitled',
			collapsed: false,
			collapsible: true,
			has_help: false,
			has_external: false,
			height: '0.5',
			fixed_size: true, // relates to height only
			removable: true,
			editable: true,
			movable: true,
			active: false,
			messageBus: Bus()
		},

		_init: function () {
			var self = this, o = this.options;

			self.handleSelector = '.widget-head';

			o.active = o.collapsed ? false : o.active;

			//remove later
			if (o.height === 'single') o.height = '0.5';
			if (o.height === 'double') o.height = '1.0';

			self._message_handler_init();

			setTimeout(function() {
				o.messageBus
					.listen('tab-deactivated', self._bind(self.deactivate))
					.listen('bring-up-to-date', function() {
						self.bring_up_to_date();
					});
			}, 0);

			self.element
			.addClass('widget dashboardWidget dashboard-widget ' + Utils.camel_to_dashes(self.widgetName))
			.attr({id: o.name})
			.div('.widget-head')
			.safeFadeIn()
			.h3(o.title)
				.attr('unselectable', 'on')
				.editable({
					success: function(_, text) {
						if(text !== o.title){
							o.title = text;
							self._emit_changed();
						}
					}
				})
				.parent()
				.map(function () {
					var head = self.head = $(this);

					var buttons = head.div('.buttons');

					if (o.removable) {
						buttons.div()
							.button({ icons: { primary: 'ui-icon-close' } })
							.hide()
							.mousedown(function (e) {
								e.stopPropagation();
							})
							.click(function () {
								//note: this confirmation will be removed once Undo is implemented.
								ECMWF.confirm({
									title: 'Remove widget',
									text: 'Widget <b>' + o.title + '</b> will be removed, ok?',
									yes: function() {
										$('.dashboard').dashboardSet('remove_widget', o._dashboard.name, o.name, function() {
											$.logMessage('Widget \'' + o.title + '\' deleted');
										});
									}
								});
								return false;
							})
							.find('.ui-icon').modeTooltip({content: 'Remove'});
					}

					if (o.collapsible) {

						(self.collapse_button = buttons.div())
							.button({
								icons: { primary: 'ui-icon-triangle-1-n' }
							})
							.mousedown(function (e) {
								e.stopPropagation();
							})
							.click(function () {
								var icon = self.collapse_button.find('.ui-icon'); icon.data('uiModeTooltip') && icon.modeTooltip('close');
								self.collapse(!o.collapsed)._emit_changed();
								return false;
							})
							.hide()
							.find('.ui-icon').modeTooltip({
								content: function () {
									return {
										text: o.collapsed ? 'Un-collapse' : 'Collapse'
									};
								}
							});
					}

				if (o.editable) {

					$('<div class="edit-box" style="display:none;">').append('<ul><li class="item"><label>Change the title?</label><input value="' + $('h3', this).text() + '"/></li>')
						.insertAfter($(self.handleSelector, head.parent()));

					buttons.div()
						.button({icons: {primary: 'ui-icon-pencil'}})
						.mousedown(function (e) {
							e.stopPropagation();
						})
						.click(function () {
							if (!self.editing) {
								self.editing = true;
								console.log('edit.click ON');
								self.edit();
								return false;
							} else {
								console.log('edit.click OFF');
								self.editbox.triggerHandler('submit');
								return false;
							}
						})
						.hide()
						.find('.ui-icon').modeTooltip({
							content: function () {
								return { text: self.editing ? 'Close edit' : 'Edit' };
							}
						});
				}

				/*
				 * The 'view in catalogue' button should be shown if:
				 *   - the product has the original_path property
				 *
				 * The 'open in eccharts' button should be shown if both:
				 *   - the user has the forecaster policy
				 *   - the chart is an eccharts chart (for the moment we assume that all interactive charts in the dashboard are eccharts charts)
				 */
				if (o.has_external && User.is_forecaster()) {
					buttons
						.div()
						.button({ icons: { primary: 'ui-icon-extlink' } })
						.click(function () {
							self.open();
							return false;
						})
						.hide()
						.find('.ui-icon').modeTooltip({ content: 'Open' });
				}

				if (o.original_path) {
					buttons
						.div()
						//.button({icons: {primary: 'ui-icon-link'}})
						.button({icons: {primary: 'ui-icon-extlink'}})
						.click(function () {
							self.view_in_catalogue();
							return false;
						})
						.hide()
						.find('.ui-icon').modeTooltip({content: 'Open'});
				}

				if (o.has_help) {
					buttons
						.div()
						.button({ icons: { primary: 'ui-icon-help' } })
						.hide()
						.find('.ui-icon')
						.modeTooltip({ content: 'Show help' });
				}

				self.hideable = head.find('.ui-button:hidden');

				if ((o.can_be_out_of_date || true)) {
					self.out_of_date_button = buttons.div()
						.button({ icons: { primary: 'ui-icon-arrowthickstop-1-n' } })
						.click(function () {
							self.bring_up_to_date();
							return false;
						})
						.hide()
						.find('.ui-icon').modeTooltip({ content: 'A newer cycle is available for this product. Click to update' })
						.parent();
				}

				o.movable && head
					.css({cursor: 'move'})
					.mousedown(function (e) {
						//$sortableItems.css({width:''});
					})
					.mouseup(function () {
						var parent = $(this).parent();
						if (parent.hasClass('dragging')) {
							$(o.columns).sortable('disable');
						}
					});

				buttons.find('.ui-button').addClass('fff');

				return $(this).parent();
			})

			self.editbox = self.element.div('.edit')
			self.content = self.element.div('.widget-content')
				.click(function () {
					console.log('content.click');
					return self.preview();
				});

			if (o.collapsed) self.collapse();

			var dashboard_size = {
				width: 0,
				height: 0
			};
			self._width;
			self.element.on('dashboardResized', function(_, size, wsize){
				if (size.width !== dashboard_size.width || size.height !== dashboard_size.height) {
					self.content.triggerHandler('resize', wsize);
					dashboard_size = size;
					self._width = wsize.width;
				}
			});

			self._emit_changed = $.throttle(1000, function () {
				var self = this;

				Tabs.Widget.save(self.serialize(), $.noop);
				self.get_bus().send('content-change');
			});

			var form_selector = '#' + o.name + ' .edit input, #' + o.name + ' .edit select, #' + o.name + ' .edit a';
			self
				._add_live(form_selector, 'focus', function() {
					$(this).parents('.form').parents('.ui-tabs-panel').find('.focused').removeClass('focused');
					//$(this).parents('tr:first').addClass('focused').parents('.dashboardWidget').addClass('focused');
					$(this).parents('tr').addClass('focused').parents('.dashboardWidget').addClass('focused');
				})
				._add_live(form_selector, 'blur', function() {
					$(this).parents('.form').find('.focused').removeClass('focused');
					$(this).parents('.dashboard-widget').removeClass('focused');
				});

			self.element.showOnMouseover({
				hideables: self.hideable
			});
		},

		destroy: function () {
			this.deactivate();
			this.element.removeClass('dashboardWidget dashboard-widget ' + Utils.camel_to_dashes(this.widgetName));
			this.element.clearQueue();
			this.element.removeData('fxqueue')
			this.head.off().remove();
			this.content.remove();
			this.element.find('> .edit').remove();
			this.element.find('> .edit-box').remove();
			this.editbox.remove();
			this.element.removeData('dashboardWidget')
			this._message_handler_destroy();
		},

		out_of_date: function (on) {
			this.out_of_date_button[on ? 'show' : 'hide']().find('.ui-icon').pulsate(on);
		},

		bring_up_to_date: function () {
			// To implement
		},

		share: $.noop,

		_notify_content_changed: function () {
			this.element.triggerHandler('contentChanged');
			this._update_height();
		},

		preview: function (enable) {
			var self = this, o = this.options;

			if (self._on_preview_open) {
				$('body').addClass('in-preview');

				var v_padding = 24;

				/* Sizing the window to the content is temporarily disabled

				// this assumes that content is scaleable
				// Otherwise we have no idea how big to make the window.
				var aspect_ratio = self.content_aspect_ratio();
				var height = $(window).height() * 0.9;
				var inner_height = height - 2 * v_padding;
				var width = aspect_ratio ? Math.min($(window).width() - 40, inner_height * aspect_ratio) : $(window).width() * 0.9;
				width = Math.max(640, width); // controls dont fit if too narrow.
				*/

				var width = $(window).width() * 0.75;
				var height = $(window).height() * 0.9;
				var inner_height = height - 2 * v_padding;

				$('#preview').dialog({
					resizable: false,
					autoOpen: false,
					modal: true,
					height: height,
					width: width,
					open: function () {
						var el = $(this);

						el.parent().addClass('ui-dialog-preview');
						el.h3(o.title);

						self._on_preview_open(el, inner_height, width);

						el.off('dialogclose').on('dialogclose', function () {
							self._on_preview_close(el);
							el.data('uiDialog') && el.dialog('destroy');
							el.empty();
							$('body').removeClass('in-preview');
						});
					}
				}).dialog('open');

				return false;
			}
			else {
				__DEV__ && console.log('widget type has no preview mode');
				return true;
			}
		},

		set_title: function (value) {
			var self = this, o = this.options;

			if (self.title_widget && (value !== self.title_widget.val())) {
				// edit mode
				self.title_widget.val(value);
			}

			if(o.title !== value) {
				o.title = value;
				self.element.find('h3').text(value);
				self._emit_changed();
			}
		},

		edit: function (enable, form) {
			var self = this, o = this.options;

			if (enable !== false) {

				var form_items = self._get_form();
				var state = self.get_state();

				function _on_title_widget(widget) {
					self.title_widget = widget;
					widget.change(function(){
						self.set_title(widget.val());
					});
				}

				o.fixed_size || form_items.unshift({
					type: 'choice',
					title: 'Height',
					name: 'height',
					values: ['0.5', '0.75', '1.0', '1.25', '1.5'],
					labels: {
						'0.5': '50%',
						'0.75': '75%',
						'1.0': '100%',
						'1.25': '125%',
						'1.5': '150%'
					},
					value: o.height,
					menu: true
				});

				form_items.unshift({
					type: 'string',
					'title': 'Title',
					name: 'title',
					value: o.title,
					widget: _on_title_widget
				});

				var first_section_n_rows = o.fixed_size ? 1 : 2;

				self.editbox
					.form({
						items: form_items,
						show_ok: false,
						show_cancel: false,
						ok: function (form) {
							console.log('editbox.ok form=', form);
							self.edit(false, form); // includes subclass functionality.
						},
						cancel: function () {
							console.log('dashboardWidget.cancel');
							self.editbox.form('destroy').remove();
							self.set_state(state);
							self.edit(false);
						}
					})
					.find('.form tr:nth-child(' + (first_section_n_rows) + ')').addClass('section-end').end()
					.find('.form tr:last').hide().css({border: 0}).prev().addClass('section-end').end(); //TODO why does form widget add empty row if no buttons?

				self.editbox.find('input:first').focus();
				self.element.addClass('editing');

				if(self.editbox.find('.form').width() > self.element.width()){
					self.editbox.find('.form').wrap('<div class="stand-alone-wrapper"><div></div></div>').addClass('stand-alone');
					/* mask interferes with Meteogram editing.
					self.editbox.find('.stand-alone-wrapper').backgroundMask({on_click: function(){
						self.editbox.form('destroy').remove();
						self.set_state(state);
						self.edit(false);
					}});
					*/
				}

				//o.messageBus.send('tab-deactivated'); //no, we cannot deactivate the tab because the animation should continue to run (WREP-125)
			} else {
				console.log('dashboardWidget.edit. OFF.');

				stop_editing(form);
			}

			self.head.find('h3').editable(enable !== false ? 'disable' : 'enable');

			function stop_editing(form) {
				//if editing was cancelled, form arg should be undefined.

				self.editbox.form('destroy');
				self.title_widget = null;
				self.element.removeClass('editing');

				self.editbox.find('.stand-alone-wrapper').remove();

				if (form) {

					var title = form.title;
					if (title !== o.title) {
						self.set_title(title);
					}

					var height = form.height;
					if (height !== o.height) {
						o.height = height;
						self._update_height().content.resize();
					}

					self.update_from_form(form);
					self._emit_changed();
				}
				self.editing = false;
			}
		},

		open: function (enable) {
			__DEV__ && console.log('widget type doesnt support opening.');
		},

		view_in_catalogue: function() {
			console.log('widget.view_in_catalogue: original_path=', this.options.original_path);

			window.location.href =
				ECMWF.fix_host.to_frontend(window.location.protocol + '//' + window.location.host)
					+ this.options.original_path + '?' + $.map(this._product.variables, function(v) { return v.name + '=' + v.to_string(); }).join('&')
			return this;
		},

		collapse: function (collapsed) {
			// note that the button api is not used to set the icon as the mouse events are lost
			var self = this, o = this.options;

			if ((o.collapsed = (collapsed !== false))) {
				self.collapse_button.find('.ui-icon').removeClass('ui-icon-triangle-1-n').addClass('ui-icon-triangle-1-s');
				self.content.hide();
				$('.inform-dialogue').remove();
				self.element.trigger('contentChanged');
			} else {
				self.collapse_button.find('.ui-icon').removeClass('ui-icon-triangle-1-s').addClass('ui-icon-triangle-1-n');
				o.active || self.activate();
				self.content.show();
				self.on_show();
			}
			return self;
		},

		collapsed: function () {
			return this.options.collapsed;
		},

		content_aspect_ratio: function() {
		},

		dont_serialize: function() {
			return ['has_help', 'messageBus', 'collapsible', 'has_external', 'klass', '_dashboard'];
		},

		serialize: function () {
			var self = this, o = this.options;

			var dont = {};
			$.each(self.dont_serialize(), function(_, x) {
				dont[x] = true;
			});

			var r = {};
			$.each(o, function(k, v) {if(!dont[k]) { r[k] = v; }});
			return r;
		},

		on_show: function () {
			//the widget is now visible and may need to update its display.
			return this;
		},

		_update_height: function () {
			var o = this.options;
			console.log(o.name + '._update_height:', this.content.height());

			var height = this.content.height();
			if(height !== this._height){
				this._height = height;
				this.get_bus().send('height-changed', this.widget_height());
			}
			return this;
		},

		content_height: function() {
			return this._height || 100;
		},

		//the total height of the widget, including headers and borders, but not external padding.
		widget_height: function () {
			var border_height = 2;
			var header_height = 24;
			return this.content_height() + header_height + border_height;
		},

		_get_form: function () {
			return [];
		},

		activate: function () {
			// The containing tab is activated and ready
			this.options.active = true;
		},

		deactivate: function () {
			// The containing tab is deactivated
			this.options.active = false;
		},

		update_from_form: function (form) {},

		_emit_dashboard_changed: $.throttle(1000, function () {
			var o = this.options;
			o._owner._emit_dashboard_changed(o._dashboard);
		}),

		get_state: function() {
			// To override
			return this.options.data;
		},

		set_state: function(state) {
			// To override
			return null;
		},

		get_bus: function() {
			return this.local_bus || this.options.messageBus;
		},

		_bind: function (fn) {
			var self = this;
			return function(){ return fn.apply(self, arguments); };
		}
	});

	$.widget('ui.nullDashboardWidget', $.ui.dashboardWidget, DashboardWidget.register('nullDashboardWidget', {
		factory: {
			title: 'New null widget',
			kind: 'null',
			data: {},
			is_private: true
		},
		_init: function() {
			this.d = this.element.div('null widget');
			$.ui.dashboardWidget.prototype._init.call(this);
		},
		destroy: function() {
			this.d.remove();
			$.ui.dashboardWidget.prototype.destroy.call(this);
		}
	}));

	$.fn.safeFadeIn = function(args) {
		return ECMWF.browser.msie ? this : this.css('opacity', 0).animate({opacity: 1});
	};

	return DashboardWidget;
});
