/*
 *  (C) Copyright ECWMF http://www.ecmwf.int
 *
 *  This software is licensed under the terms of the Apache Licence Version 2.0
 *  which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 *  In applying this licence, ECMWF does not waive the privileges and immunities
 *  granted to it by virtue of its status as an intergovernmental organisation nor
 *  does it submit to any jurisdiction.
 *
 *  Filterable list
 *  ---------------
 *  A list of objects that has filters applied.
 *  Use the subscribe method to be notified of filtering changes.
 *  Use the have.filtering promise to ensure that the list creation is complete.
 *
 *  To have the list be sorted, set the sort_by property of the list. This will
 *  assume that each list item has a key with this name.
 *
 *  Observables:
 *    - filter change - subscribe to the list itself
 *    - selection
 *    - on.items - this indicates that the list contents have completely changed,
 *        eg, when changing package
 *
 */
define(['jquery', 'observable', 'throttle-debounce'], function($, Observable) {

var List = function() {}

List.prototype = $.extend({
	list_init: function(){
		this.items = null    // all items
		this.filtered = null // array of indexes to the full list.

		this.filters = []

		// TODO rename these to added, removed
		this.changed = {
			to_show: [],
			to_hide: []
		}

		this.selection = $.extend(new Observable(), {always: true})
		this.on = {
			items: new Observable() // indicates list contents have completely changed, eg change of chart package.
		}

		this.have = {
			items: $.Deferred(),
			filtering: $.Deferred()
		}

		this.filter = $.debounce(10, this.bind(this.__filter));
	},

	get_filters: function() {
		return this.filters
	},

	// this must be called by users to indicate that data is ready
	on_load: function(items) {
		__DEV__ && !items && console.warn('on_load: no items');
		this.items = items;
		this.filter();
		this.have.items.resolve(items);
		this.on.items.emit();
	},

	__filter: function() {
		var self = this;

		this._need_filter = this._need_filter && clearTimeout(this._need_filter)
		var old_len = this.filtered ? this.filtered.length : -1 // -1 ensures change is emitted first time
		this.changed = {to_hide: [], to_show: []};
		var old = this.filtered;
		this.filtered = [];

		if(!this.items) return __DEV__ && console.warn('filtering empty list');

		var n_matches = 0;
		this.each(function(v, i){
			var visible = this.filter_pass;

			if((this.filter_pass = self.filter_item(this))){
				if(!visible) self.changed.to_show.push(this);
				self.filtered.push(i);
				this.filtered_idx = n_matches++;
			}
			else{
				if(visible) self.changed.to_hide.push(this);
			}

			// TODO if this turns out to be useful it should be optimised and managed by the facet widget.
			self.filter_pass_no_facets = self.filter_item_without_facets(this);
		});

		if(this.sort_by){
			this.filtered.sort(function(a, b){
				return this.items[a].data[this.sort_by] > this.items[b].data[this.sort_by];
			});
		}

		list_compare(this.filtered, old) || this.emit();

		// fix selection if invalid
		if($.inArray(+this.selection.value, this.filtered) < 0){
			this.selection.set((this.selection.always && this.filtered.length) ? this.filtered[0] : null);
		}

		this.have.filtering.resolve();
	},

	/*
	 * return true or false
	 */
	filter_item: function(item) {
		for(var i in this.filters){
			if((this.filters[i].disabled !== true) && !this.filters[i].fn(item)) return false;
		}
		return true;
	},

	filter_item_without_facets: function(item) {
		for(var i in this.filters){
			if((this.filters[i].name !== 'facets' && this.filters[i].disabled !== true) && !this.filters[i].fn(item)) return false;
		}
		return true;
	},

	lookup_filter_by_name: function(name) {
		for(var i in this.filters){
			var f = this.filters[i];
			if(f.name === name) return f;
		}
	},

	lookup_idx_by_name: function(name) {
		for(var i in this.items){
			if(this.items[i].name() === name) return +i;
		}
	},

	add_filters: function(filters) {
		var self = this;

		$.map(filters, function(f) {
			if($.isFunction(f.fn)){
				f.init && f.init();
				self.filters.push(f);
			}
		})

		// very low priority (its very likely we are not ready to filter yet)
		// Currently it is set to a very long time due to the slowness of find-product
		this._need_filter = setTimeout(function(){
			self.filter()
		}, 1000);
	},

	remove_filter: function(name) {
		var filter = this.lookup_filter_by_name(name);
		this.filters = $.map(this.filters, function(v){ if(v.name !== name) return v; });
		this.filter()
	},

	each: function(fn) {
		// TODO why is i a string!?
		for(var i in this.items) fn.call(this.items[i], this.items[i], +i);
	},

	each_by_facet: function(facet, fn1, fn2) {
		var self = this;
		var g = -1, ig = 0, gr = {};
		var other = {name: 'Other'};
		for(var i=0;i<this.filtered.length;i++){
			var item = self.items[this.filtered[i]];
			var group_name = (item.data.tags && item.data.tags[facet] || [])[0];
			if(!i || (group_name && group_name !== gr.name)){
				g++;
				gr.name = group_name || other.name;
				fn1(gr.name);
				ig = 0;
			}
			fn2(item, i, ig++, group_name, g);
		}
	}
}, new Observable())

function list_compare(a, b) {
	if(!b || (a.length !== b.length)) return false;
	for(var i in a){
		if(a[i] !== b[i]) return false;
	}
	return true;
}

return List;

});
