define(['config', 'jquery', 'ajax_queue', 'mixins', '../assigned_item', 'utils'], function(ECMWF, $, Ajaxqueue, mixins, AssignedItem){

/*
 *   managing updates:
 *      the basic job of updates is to map grid.filtered to screen items.
 *      startup: in order prevent unnecesary activity, mapping should not be done until either:
 *         1) _on_filter_change happens, or
 *         2) a timeout is reached.
 *
 *   update types:
 *      resize
 *           - resize calls _assign_fresh directly ?
 *      scroll
 *	         does its own assignment and does not trigger any further events.
 *      filter_change
 *          -> hide_show -> _assign_fresh
 *
 *   update tasks:
 *      -sizing: calculate number of visible rows and columns.
 *      -filtering: calculate number of visible items and update the 'filtered' array.
 *      -dom creation: update the number of dom elements
 *      -assignment: assign model items to dom elements.
 *      -follow-on tasks: update scrollbars
 */
$.widget('ui.viewBase', $.extend({

	options: {
		grid: null,
		view: null,      // used to get read/write access the model.
		click: $.noop,
		// TODO dblclick should be listed here?
		select: $.noop,
		progress: true
	},

	_create: function() {
		var self = this, o = self.options;

		this.item_options = {
			grid: o.grid,
			view: self,
			on_click: function(item) {
				var selected = self.element.find('.selected');
				var widget = self.element.find('[name="' + item.name() + '"]').parents('.' + self.ItemView.className);
				if(!widget.hasClass('selected')){
					self.select(item);
					o.select(item);
				}
				o.click(item);
			},
			on_dblclick: function(item) {
				o.dblclick(item);
			}
		}
	},

	_init: function() {
		var self = this, o = this.options;

		if(!o.grid) return console.warn('grid._init: no grid');
		if(!o.products.items) console.warn('grid._init: no items');

		this.visible_items = {}; // Items keyed by name
		this.assigned = [];      // AssignedItem's keyed by dom index.

		this.cache = {
			n_cells: null,        // the total number of dom elements. includes disabled items and overscan.
			max_height: o.grid.element.css('maxHeight') && parseInt(o.grid.element.css('maxHeight'), 10) - parseInt(o.grid.outer.css('marginTop'), 10) - parseInt(o.grid.outer.css('marginBottom'), 10)
		}
		if(this.cache.max_height){
			o.grid.outer.css({maxHeight: this.cache.max_height});
		}

		this._check_layout = this.debounce(this.__check_layout);
		this._set_scrollbar_size = $.debounce(400, this._set_scrollbar_size);

		self.have = {
			assignment: $.Deferred(),//.done(o.grid.options.on_view_ready),
			size: $.Deferred().done(o.grid.options.on_view_ready)
		}

		o.grid._update({sizing: true});

		o.grid.element.resize(function() {
			o.grid._update({sizing: true});
		});

		self.element.on('layout', function(){
			o.grid._update({sizing: true});
		});
	},

	_destroy: function() {
		this._frozen = false;
	},

	_assign_fresh: function() {
		var self = this, list = this.options.products;

		self.visible_items = {};

		var n = 0;
		self._each_allocated(function(item, m, d){
			self.visible_items[list.items[m].name()] = list.items[m];
			self._assign(d, m);
			n = d;
		});

		var d = list.filtered.length ? n + 1 : 0;
		//console.log('_assign_fresh: matches=' + list.filtered.length + ' visible=' + self.cache.n_cells + ' cleared=' + (self.cache.n_cells - d) + ' elems=', self.element.find('.thumbnail').length, self.element.find('.thumbnail:visible').length);

		//clear remaining dom elements:
		for(;d<self.cache.n_cells;d++){
			self._unassign(d);
			self.assigned[d].widget({item: null});
		}

		self._on_n_viewable_items_changed();

		self.have.assignment.resolve();

		self.options.progress && self._start_progress();
	},

	_start_progress: function() {
		var self = this;
		self._progress = self._progress || {bar: $([])};
		var percent = 0;

		setTimeout(function() {
			var n = self.n_visible_really() || self.n_viewable_items();
			var p = $.extend(self._progress, {
				n: 0,
				total: Math.min(self.n_viewable_items(), n) // account for allocated but not visiable rows.
			});

			if(self._interval) return;

			self._interval = setInterval(function() {
				var tot = p.total;
				for(var d=0;d<self.n_viewable_items();d++){
					var assigned = self.assigned[d];
					var data = assigned.dom.data('uiThumbItem');
					if(data){
						if(data.thumbnail_loading && assigned.in_view){
							tot--;
						}
					}
				}

				if(++p.n >= 25 || tot >= p.total){
					self._interval = clearInterval(self._interval);
					//console.log('progress done: (' + self._progress.n + ')', tot, ' / ', p.total);
					p.bar.removeClass('smooth-width').css({visibility: 'hidden', opacity: 0}).delayed(200, function(){ $(this).css({width: '0%'}); });
				}else{
					p.bar = self.element.parent().parent().find('.progress-bar').orelse(function(){ self.element.parent().parent().div('.progress-bar fade-in-out smooth-width'); });
					p.bar.css({visibility: 'visible', opacity: 0.5});
					setTimeout(function() {

						var pct = 100 * Math.max(0, tot / p.total);
						//console.log('     ', tot, ': ', pct + '% ', tot, '/', p.total);
						if(pct < percent){
							p.bar.removeClass('smooth-width');
							setTimeout(function() {
								p.bar.css({width: pct + '%'});
							}, 200);
						}
						else p.bar.addClass('smooth-width').css({width: pct + '%'});
						percent = pct;
					}, 100);
				}
			}, 2000);
		}, 2000);
	},

	_assign: function(dom_idx, model_idx) {
		var o = this.options;

		var ai = this.assigned[dom_idx];

		if(ai.model_idx !== model_idx){
			ai.assign({
				item:         o.products.items[model_idx],
				model_idx:    model_idx,
				filtered_idx: o.products.items[model_idx].filtered_idx
			});
		}
	},

	_is_assigned: function(item) {
		return typeof this.visible_items[item.name] !== 'undefined';
	},

	/*
	 *   This excludes items that are allocated but just outside the viewport.
	 *   Needs refactoring: the in_view property should perhaps be set at allocation time.
	 */
	n_visible_really: function() {
		var self = this;

		var h = self.element.parent().height();

		var v = 0;
		for(var d=0;d<self.n_viewable_items();d++){
			var assigned = self.assigned[d];
			var data = assigned.dom.data('uiThumbItem');
			if(data){
				var top = parseInt(assigned.dom.css('top'), 10) - self.element.parent().scrollTop();
				(assigned.in_view = top < h) && (v++);
			}
		}
		return v;
	},

	// args: lists of model items to show and hide
	hide_show_items: function(args) {
		var self = this, o = self.options;

		/*
		$.each(args.show, function(){
		});
		*/

		//console.log('viewBase.hide_show_items: to_show=' + ECMWF.n_properties(args.show) + ' to_hide=' + ECMWF.n_properties(args.hide));

		if(!($.isEmptyObject(args.show) && $.isEmptyObject(args.hide))){
			var to_unassign = [];
			$.each(args.hide, function(){
				var is_assigned = self._is_assigned(this);
				if(is_assigned){
					to_unassign.push(this);
				}
			});

			$.each(to_unassign, function(i, b) {
				// this will no longer work because the tag is postfixed by the item widget eventNamespace
				// but we dont seem to get here anyway
				Ajaxqueue.cancelByTag('grid-' + this.name());
			});

			if(o.products.filtered.length < self.cache.n_cells){
				self.element.parent().scrollTop(0);
			}

			//now we may have gaps - look from scratch at what needs to be shown
			self._assign_fresh();

			self._set_scrollbar_size();
		}

		o.grid.options.size_fits_content && self.resize_to_content();
	},

	/*
	 *  Override for view modes that do not have uniform row height
	 */
	natural_height: function() {
		return Math.ceil(this.options.products.filtered.length / this.n_columns) * (this.item_height + this.padding.y) - this.padding.y;
	},

	max_items: function() {
		return Math.floor(this.n_columns * this.cache.max_height / (this.item_height + this.padding.y));
	},

	resize_to_content: function() {
		var self = this, o = self.options, el = o.grid.element;

		if(o.products.filtered.length <= this.max_items()){
			var height = this.natural_height() + parseInt(o.grid.outer.css('marginTop'), 10) + parseInt(o.grid.outer.css('marginBottom'), 10) + o.grid.widgets.foot.outerHeight(true);
			var old_height = parseInt(el.css('height'), 10);
			if(height !== old_height){
				el.css({height: height}).attr({sizor: 'view_base.resize_to_content'});
				setTimeout(function(){ // TODO
					o.grid._update({sizing: true});
				}, 1000);
			}
		}else{
			// with a large number of items, the height is limited by max-height in css
			el.css({height: ''}).removeAttr('sizor');
			o.grid.container.parent().css({maxHeight: self.cache.max_height - o.grid.widgets.foot.outerHeight(true)}).attr({'max-sizor': 'resize_to_content'});
			setTimeout(function(){ // TODO
				o.grid._update({sizing: true});
			}, 500);
		}
	},

	scroll: function(top) {
		//this._check_layout();
		this.options.progress && this._start_progress();
	},

	select: function(item) {
		var self = this, products = this.options.products;

		products.have.filtering.done(function() {
			self.element.find('.selected').removeClass('selected');
			var ai = $.findWhere(self.assigned, {item: item}); // note that there is no reason why ai should be assigned
			ai && ai.widget('select');
		});
	},

	select_by_idx: function(model_idx) {
		var self = this, products = this.options.products;

		products.have.filtering.done(function() {
			var item = products.items[model_idx];
			if(item){
				self.select(item);
			}
		});
	},

	// helper for resize()
	__post_resize_assignment: function() {
		var self = this;

		for(var d=0;d<self.cache.n_cells;d++){
			self.assigned[d] = self.assigned[d] || new AssignedItem(self.ItemView.widgetName, self.item_options);
			// note that we create empty elements here which may never be assigned, it would be better to create the elements on demand
			self.assigned[d].dom = self.assigned[d].dom || self.element.div(self._new_element(d));
		}

		if(self.cache.n_cells < self._old_n_visible_items){
			for(;d<self._old_n_visible_items;d++){
				self._unassign(d);
			}
		}
	},

	add_remove_items: function(add, remove) {
		// the model list has changed. currently not used.
		var self = this, o = self.options;

		if(add){
			for(var i=0;i<self.cache.n_cells;i++){
				self.assigned[i].widget($.extend(self.item_options, {product: add[i]}));
				self.visible_items[add[i].name] = add[i];
			}
			o.grid._update({sizing: true});
		}
	},

	_unassign: function(dom_idx){
		var ai, self = this;

		if((ai = self.assigned[dom_idx])){
			ai.assign({model_idx: null, item: null});
		}
	},

	_m_is_assigned: function(model_idx){
		var self = this;

		for(var i in self.assigned){
			var ai = self.assigned[i];
			if(ai.model_idx === model_idx) return true;
		}
		return false;
	},

	_each_allocated: function(fn) {
		var list = this.options.products;

		var n = Math.min(this.cache.n_cells, list.filtered.length);
		for(var i=0;i<n;i++){
			var m = list.filtered[i];
			fn(list.items[m], m, i);
		}
	},

	n_viewable_items: function() {
		 return Math.min(this.cache.n_cells, this.options.products.filtered.length);
	},

	_on_n_viewable_items_changed: function() {
		this._set_scrollbar_size();
	},

	_get_filter_idx: function(model_idx){
		var self = this, o = self.options, list = o.products;

		var name = list.items[model_idx].name();
		for(var f in list.filtered){
			if(list.items[list.filtered[f]].name() === name) return f;
		}
		console.warn('viewBase._get_filter_idx: not found: model_idx=' + model_idx + ' ' + name);
		return -1;
	},

	freeze: function() {
		this._frozen = true;
	},

	/*
	 *  Set the inner height, ie the full height of all the scrollable elements
	 */
	_set_scrollbar_size: function() {
		var self = this, o = this.options;

		if(self.n_columns){
			var height = self.natural_height() + 20;
/* no! we need to set the full (inner) height here
			if(self.cache.max_height) height = Math.min(height, self.cache.max_height - o.grid.widgets.foot.outerHeight(true));
*/
			self.element.css({height: height})
				.attr({sizor: 'view_base.set_scrollbar_size'});
			//console.log('_set_scrollbar_size: n_viewable=', o.products.filtered.length, ' ncolumns=', self.n_columns, ' natural_height=', height, '/', self.natural_height(), o.grid.container);
			self.have.size.resolve();
			self.element.parent().perfectScrollbar('update');
		}
	},

	__check_layout: function() {

		function array_to_object_keys(array){
			var obj = {};
			$.each(array, function(i, v){ obj[v] = {}; });
			return obj;
		}

		// check the same item is not displayed twice
		// - 'title' (product.description) is not checked as it is not unique.
		// - unfortunately some products are in fact duplicates for access control reasons so this test is not very useful
		var attrs = ['name', 'src'];
		var assigned = array_to_object_keys(attrs);
		this.element.find('.thumbnail-img').each(function(i){
			var img = $(this);
			$.each(attrs, function(i, attr){
				var val = img.attr(attr);
				if(val){
					if((val.indexOf('default') > -1) || (val.indexOf('data:') > -1) || (val.indexOf('failed_') > -1)) return;
					if(assigned[attr][val]) console.error('duplicate assignment: ' + attr + ': ' + val);
					assigned[attr][val] = true;
				}
			});
		});
	}
}, mixins));

});
