/**
* Copyright (c) 2009 Chris Leonello
* This software is licensed under the GPL version 2.0 and MIT licenses.
*/
(function($) {
    /**
     * Class: $.jqplot.BoxPlotRenderer
     * jqPlot Plugin to draw Open Hi Low Close, Candlestick and Hi Low Close charts.
     * 
     * To use this plugin, include the renderer js file in 
     * your source:
     * 
     * > <script type="text/javascript" src="plugins/jqplot.BoxPlotRenderer.js"></script>
     * 
     * You will most likely want to use a date axis renderer
     * for the x axis also, so include the date axis render js file also:
     * 
     * > <script type="text/javascript" src="plugins/jqplot.dateAxisRenderer.js"></script>
     * 
     * Then you set the renderer in the series options on your plot:
     * 
     * > series: [{renderer:$.jqplot.BoxPlotRenderer}]
     * 
     * For OHLC and candlestick charts, data should be specified
     * like so:
     * 
     * > dat = [['07/06/2009',138.7,139.68,135.18,135.4], ['06/29/2009',143.46,144.66,139.79,140.02], ...]
     * 
     * If the data array has only 4 values per point instead of 5,
     * the renderer will create a Hi Low Close chart instead.  In that case,
     * data should be supplied like:
     * 
     * > dat = [['07/06/2009',139.68,135.18,135.4], ['06/29/2009',144.66,139.79,140.02], ...]
     * 
     * To generate a candlestick chart instead of an OHLC chart,
     * set the "candlestick" option to true:
     * 
     * > series: [{renderer:$.jqplot.BoxPlotRenderer, rendererOptions:{candleStick:true}}],
     * 
     */
    $.jqplot.BoxPlotRenderer = function(){
        // subclass line renderer to make use of some of it's methods.
        $.jqplot.LineRenderer.call(this);
        
        // prop: bodyWidth
        // width of the candlestick body.  Default will auto calculate
        // based on plot width and number of candlesticks displayed.
        this.bodyWidth = null;
        // prop: fillColor
        // fillColor of the box.  Default is series color.
        this.fillColor = null;
            };
    
    $.jqplot.BoxPlotRenderer.prototype = new $.jqplot.LineRenderer();
    $.jqplot.BoxPlotRenderer.prototype.constructor = $.jqplot.BoxPlotRenderer;
    
    // called with scope of series.
    $.jqplot.BoxPlotRenderer.prototype.init = function(options) {
        // prop: lineWidth
        // Width of the hi-low line and open/close ticks.
        
        $.jqplot.LineRenderer.prototype.init.call(this, options);
        // set the yaxis data bounds here to account for hi and low values
        var db = this._yaxis._dataBounds;
        var d = this._plotData;
        // if data points have less than 5 values, force a hlc chart.
        if (d[0].length < 8) {
            
        }
        else {
            for (var j=0; j<d.length; j++) { 
                if (d[0][1] < db.min || db.min == null) {
                	db.min = d[j][1];
                }
                if (d[j][7] > db.max || db.max == null) {
                	db.max = d[j][7];
                }             
            }
        }
        
    };
    
    // called within scope of series.
    $.jqplot.BoxPlotRenderer.prototype.draw = function(ctx, gd, options) {
        var d = this.data;
        var xmin = this._xaxis.min;
        var xmax = this._xaxis.max;
        // index of last value below range of plot.
        var xminidx = 0;
        // index of first value above range of plot.
        var xmaxidx = d.length;
        var xp = this._xaxis.series_u2p;
        var yp = this._yaxis.series_u2p;
        var i, prevColor, ops, b, h, w, a, points;
        var o;
        var r = this.renderer;
        var opts = (options != undefined) ? options : {};
        var shadow = (opts.shadow != undefined) ? opts.shadow : this.shadow;
        var fill = (opts.fill != undefined) ? opts.fill : this.fill;
        var fillAndStroke = (opts.fillAndStroke != undefined) ? opts.fillAndStroke : this.fillAndStroke;
        r.bodyWidth = (opts.bodyWidth != undefined) ? opts.bodyWidth : r.bodyWidth;
        r.fillColor = (opts.fillColor != undefined) ? opts.fillColor : r.fillColor;
        ctx.save();
        //if (this.show)
         {
            var x, min, max, v10, v90, v25, v50, v75;
            // need to get widths based on number of points shown,
            // not on total number of points.  Use the results 
            // to speed up drawing in next step.
            for (var i=0; i<d.length; i++) {
                if (d[i][0] < xmin) {
                    xminidx = i;
                }
                else if (d[i][0] < xmax) {
                    xmaxidx = i+1;
                }
            }
            if (r.bodyWidth == null) {
                r.bodyWidth = Math.min(20, ctx.canvas.width/(xmaxidx - xminidx)/2);
            }
            
            for (var i=xminidx; i<xmaxidx; i++) {
                x = xp(d[i][0]);
                min = yp(d[i][1]);
                v10 = yp(d[i][2]);
                v25 = yp(d[i][3]);
                v50 = yp(d[i][4]);
                v75 = yp(d[i][5]);
                v90 = yp(d[i][6]);
                max = yp(d[i][7]);
                o = {};               
                w = r.bodyWidth;
                
                a = x - (w/2);
                ops = $.extend(true, {}, opts, o);
                
                
                r.shapeRenderer.draw(ctx, [[x, min], [x, v10]], ops); 
                r.shapeRenderer.draw(ctx, [[x, max], [x, v90]], ops); 
                this.renderer.shadowRenderer.draw(ctx, [[x, min], [x, v10]], ops); 
                this.renderer.shadowRenderer.draw(ctx, [[x, max], [x, v90]], ops);
               
                
                s = (w-1.5);
                
                o = {};
                o.fillRect = true;
                o.fillAlpha = 10;
                o.fillStyle = r.fillColor;
                ops = $.extend(true, {}, opts, o);
                
                
                r.shapeRenderer.draw(ctx, [x-s/4, v75, s/2, v90-v75], ops); 
                r.shapeRenderer.draw(ctx, [x-w/4, v10, s/2, v25-v10], ops);
                r.shapeRenderer.draw(ctx, [x-w/2, v25, w, v75-v25], ops);
                
                 o = {};
                 o.strokeRect = true;
                 
                  ops = $.extend(true, {}, opts, o);
                 r.shapeRenderer.draw(ctx, [x-w/4, v75, w/2, v90-v75], ops); 
                r.shapeRenderer.draw(ctx, [x-w/4, v10, w/2, v25-v10], ops);
                r.shapeRenderer.draw(ctx, [x-w/2, v25, w, v75-v25], ops);
                this.renderer.shadowRenderer.draw(ctx, [[x+w/4, v10], [x+w/4, v25]], ops);
                this.renderer.shadowRenderer.draw(ctx, [[x+w/2, v25], [x+w/2, v75]], ops);
                this.renderer.shadowRenderer.draw(ctx, [[x+w/4, v75], [x+w/4, v90]], ops);
                 
                 o = {};
                
                 
                  ops = $.extend(true, {}, opts, o);
                  r.shapeRenderer.draw(ctx, [[x-w/2, v50], [x+w/2, v50]], ops);
            }
                                
        }
        
        ctx.restore();
    };  
    
    $.jqplot.BoxPlotRenderer.prototype.drawShadow = function(ctx, gd, options) {
        // This is a no-op, shadows drawn with lines.
    };
    
    // called with scope of plot.
    $.jqplot.BoxPlotRenderer.checkOptions = function(target, data, options) {
        // provide some sensible highlighter options by default
        // These aren't good for hlc, only for ohlc or candlestick
        //if (!options.highlighter) {
            options.highlighter = {
                showMarker:false,
                tooltipAxes: 'y',
                yvalues: 7,
                formatString:'<table class="jqplot-highlighter"><tr><td>date:</td><td>%s</td></tr><tr><td>open:</td><td>%s</td></tr><tr><td>hi:</td><td>%s</td></tr><tr><td>low:</td><td>%s</td></tr><tr><td>close:</td><td>%s</td></tr><tr><td>hi:</td><td>%s</td></tr><tr><td>low:</td><td>%s</td></tr><tr><td>close:</td><td>%s</td></tr></table>'
            };
       // }
    };
    
	//$.jqplot.preInitHooks.push($.jqplot.BoxPlotRenderer.checkOptions);
    
})(jQuery);    
