/**
 *  ECMWF Chart Handler
 *
 *  Manages instantiation and updating of different types of chart widgets.
 *  This is all it does. It does not handle other page level objects.
 *
 *  (C) Copyright ECMWF http://www.ecmwf.int
 *  This software is licensed under the terms of the Apache Licence Version 2.0
 *  which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 *  In applying this licence, ECMWF does not waive the privileges and immunities
 *  granted to it by virtue of its status as an intergovernmental organisation nor
 *  does it submit to any jurisdiction.
 *
 */
define(['jquery', 'utils', 'bus', 'unit', 'page/query', 'chart-app/router', 'observable', 'container', 'element'], function($, Utils, Bus, Units, qs, router, Observable, Container, _, undefined){

	var handlers = {
		widgets: {
			chart: null
		},
		state: {
			product: null,
			params: {}
		},
		on: {
			size_change: $.noop
		},

		product_change: new Observable(),

		// TODO dont use from outside
		set_state: function(state) {
			$.extend(handlers.state, state);
		},

		pre_build: function(options) {
			var self = this;
			this.el = options.el;

			handlers.widgets.controls = {
				outer:  self.el.show().find('.chart-controls').orelse(function(){ return self.el.div('.chart-controls'); }),
				rhs:    self.el.find('.chart-controls > .bar-2').orelse(function(){ return self.el.find('.chart-controls').div('.bar-2'); }),
				params: self.el.find('.chart-controls > .chart-parameter-bar').orelse(function(){ return self.el.find('.chart-controls').div('.chart-parameter-bar'); })
			}
		},

		basic: new (function BasicHandler() {
			this.scrollable = null;
			this.bus = null;

			this.build = function(product, state_, options, done) {
				var self = this;

				require(['www_chart/model/legacy_products', 'basic_chart/model/product', 'basic_chart/parameter_select', 'basic_chart/tools', /*'www_chart/canvas_renderer', */'basic_chart/plugins/time_navigator', 'basic_chart/plugins/fail', 'basic_chart', 'basic_chart/parameter_controls/factory'], function(_products, BasicInstance, ParameterSelect, Tools, /*Canvas, */TimeNavigator, Fail){
					self.BasicInstance = BasicInstance;
					self.Products = _products;

					handlers.state.list_product = product;

					handlers.pre_build.call(self, options);

					self.scrollable = options.el.div('.scrollable')

					self.plugins = [
							{mixin: ParameterSelect, options: {controls: handlers.widgets.controls.params}},
							{mixin: Tools,         options: {parent: handlers.widgets.controls.rhs}},
							//{mixin: Canvas,        options: {}},
							{mixin: TimeNavigator},
							{mixin: Fail}
					];

					handlers.widgets.chart = self.scrollable
						.div('.www-chart-container')
							.div('.www-chart-image-wrapper')
							.attr({id: 'basic-chart'});

					self.bus = handlers.widgets.chart.messageBus()
						.listen(options.el, 'chart-content-update', function(_, content) {
							handlers.state.product = product = content.product;

							handlers.widgets.controls.outer.css({minHeight: ''});
							options.el.css({minHeight: ''});

							/*
							 *  The path is stored to detect whether route changes are initiated by the the application or the browswer.
							 *
							 *  TODO change requests should all be as similar as possible. This may mean that the chart parameter changes
							 *  need to be made via the Router. But a conflicting requirement is that the url should not be updated
							 *  until after the plot request is complete as the returned data determines the url.
							 */

							handlers.state.params = Utils.map(handlers.state.product.variables, function(v){ return v.to_string(); });

							router.set_path(undefined, product.name(), qs.get_string($.extend({facets: router.q.facets}, handlers.state.params)), product.title_());
						})
						.listen(options.el, 'toggle-parameter-animate', function(_, parameter) {
							require(['www_chart/parameter_animator'], function(animator) {
								animator
									.chart(handlers.widgets.chart.data('ui-basicChart')) // TODO only set once
									.parameter(parameter)
									.toggle();
							});
						})
						.listen(options.el, 'size-changed', function(_, size) {
							if(size && handlers.widgets.chart){
								handlers.on.size_change(size);
							}
						})
						.listen('new-history-state', function(){
							console.log('chart_handler: got new history state');
							router.add_state();
						});

					setTimeout(function(){ self.post() }, 0);

					done && done();
				});
			},
			this.update = function(product) {
				var self = this;
				handlers.widgets.controls.outer.css({minHeight: handlers.widgets.controls.outer.find('.ui-button:first').height() + 3});

				delete self.instance;
				self.bus.clear();

				handlers.state.list_product = product;

				this.post();
			},
			this.post = function() {
				// common to both build() and update()

				var self = this;
				var product = handlers.state.list_product;

				if(handlers.widgets.chart){
					var height = this.el.height();
					this.el.css({minHeight: height});
				}

				self.instance = this.BasicInstance.get_instance({product: product.data, modifiable: true, priority: 'high'})

				if(self.instance.data.img){
					$.each(handlers.state.params, function(i, v) {
						if(self.instance.variables[i]){
							self.instance.variables[i].set(v);
						}else{
							console.log('ChartHandler: ignoring inapropriate variable:', i);
						}
					});
				}else{
					self.instance.set_parameters(handlers.state.params);
				}

				if(handlers.widgets.chart.basicChart('instance')){
					handlers.widgets.chart.basicChart('setProduct', {instance: self.instance});
				}else{
					handlers.widgets.chart.basicChart({
						instance: self.instance,
						plugins: self.plugins
					});
					Bus().send('chart-added');
				}

				router.set_path(undefined, product.name(), qs.get_string($.extend({facets: router.q.facets}, handlers.state.params)), product.long_title());

				handlers.product_change.emit();
			},
			this.cleanup = function() {
				handlers.widgets.chart = null;
				this.scrollable.remove(); delete this.scrollable;
				delete this.instance;
				this.bus.clear();
			}
		})(),

		interactive: new (function InteractiveHandler() {
			var self = this;

			this.build = function(product, state, options, done) {
				Units.load.then(function() {

					require(['chart/animator'], function() { // chart modules must be loaded before the chart but are not a hard dependency.
// basic_chart/tools ? move to chart ?
					require(['basic_chart/tools', 'chart/plugins/parameter_select', 'chart/plugins/clickable', 'chart/latlon', 'time-navigator', 'chart/plugins/legend', 'chart/plugins/www_title', 'chart/plugins/meteogram'], function(Tools, ParameterSelect, Clickable) {
						options.el.show();

						handlers.pre_build.call(this, options);

						handlers.widgets.chart = options.el.div({css: options.size}).attr({sizor: 'chart_handler'});

						var chart = handlers.widgets.chart.chart({
							product: product.name(),
							size: options.size,
							messageBus: self.bus = handlers.widgets.chart.messageBus(),
							availability: false,
							editable: false,
							enable_undo: false,
							expand_on_resize: false,
							show_latlong: true,
							values: state,
							plugins: [
								{mixin: ParameterSelect, options: {controls: handlers.widgets.controls.params}},
								{mixin: Tools, options: {parent: '.chart-controls .bar-2'}},
								{mixin: Clickable}
							]
						}).chart('instance');

						handlers.state.product = chart.product;

						//Currently interactive charts are only clickable, set mouse pointer correctly
						// TODO move out of chart_handler
						chart._set_mode('select', {cursor: 'pointer'});

						// TODO for consistency, time_navigator should be a chart plugin
						if($('#time-navigator').length){
							console.warn('FIXME unexpected time-navigator found - removing...');
							$('#time-navigator').remove();
						}
						handlers.widgets.time_navigator = options.el.div('#time-navigator').timeNavigator({messageBus: self.bus, show_bt: false});

						self.bus
							.listen_once(options.el, 'chart-content-update', function(_, content) {
								// note that the time navigator is sized by the product and not by the window.
								handlers.widgets.time_navigator.show()
									.find('.containr').triggerHandler('layout', options.size)

								// temporary fix for legend plugin
								// TODO define the behaviour of this message and if it should be fired for new charts, ensure that the chart widget does this
								self.bus.send('chart-product-change', {product: chart.product})
							})
							.listen(options.el, 'chart-content-update', function(_, content) {
								var product = handlers.state.product = content.product;

								var product_name = product.name; // TODO make this compatible with basicChart.
								var product_title = product.title(); // TODO make this compatible with basicChart.
								handlers.state.params = Utils.map(product.variables, function(v){ return v.to_string(); });
								router.set_path(undefined, product_name, qs.get_string($.extend({facets: router.q.facets}, handlers.state.params)), product_title);
							})
							.listen('tool-print-select', function() {
								handlers.state.product.www_print(handlers.state.product.www_title, function(result) {
									location.href = result.url;
								});
							});

						Bus().send('chart-added');
						self.bus.send('product-swapped'); // fake message for Tools

						self.post();

						done && done();
					})});
				})
			},

			this.update = function(product) {
				handlers.widgets.chart && handlers.widgets.chart.chart('setProduct', product.name());
				handlers.state.list_product = product;

				self.post();
			},

			this.post = function() {
				handlers.product_change.emit();
			},

			this.cleanup = function() {
				handlers.widgets.chart.remove();
				handlers.widgets.chart = null;

				handlers.widgets.time_navigator.remove();
				delete handlers.widgets.time_navigator;
			}
		})(),

		current: null
	}

	return handlers;
});
