/*
 *  Basic usage:
 *
 *  require(['test-lib'], function(Tests) {
 *    new Tests([
 *       // list of tests to run. An instance of Test is created for each item.
 *    ])
 *    .run();
 *  });
 *
 *  Each test must have a 'start' method
 *  and can be stopped by calling methods including:
 *    - finish
 *    - fail
 *    - assert
 *    - wait_for
 *
 *   See Eccharts dashboard for usage examples
 */
define(['jquery', 'ajax_queue', 'async', 'bus', 'notify'], function($, Ajaxqueue, async, Bus) {

	var buttons = {};

	/*
	 *  Test - base object providing various utilities for component or application testing.
	 */
	window.Test = function(name, el, start, stop){
		this.name = name;
		this._start = start;
		this._stop = stop || $.noop;
		this.el = el;
		this.init(name);
		this.bus = Bus();
		this.status = $('#status').orelse(function(){ return el.div({id: 'status'}); });
		this.iter = 0;

		Ajaxqueue.options.on_error = this.bind(function(e){
			this.fail(e)
		})

		this.series = function(fnlist, done){
			// the test will finish after the series unless done returns false.

			var test = this;

			async.series(fnlist, function(err){
				if (err) console.error('******** Test.series.finished: ERROR:', err);
				else console.log('series finished ok.');
				var finish = true;
				done && (finish = done(err));
				if(finish !== false) test.finished();
				fnlist.splice(0, fnlist.length); // clear the list
			});
		}

		// As above but fn items must return a promise instead of using a callback function
		// The functions are run imediately, but the deferred's are not run
		// until their turn. For this to work, tests must not do any work outside of
		// the deferred.
		this.pseries = function(fnlist, done){
			var test = this;

			function chain_promises(list) {
				return list.reduce(function(chain, fn, idx){
					if(chain && !chain.then) return test.fail('previous test did not return a promise');
					return chain ? chain.then(fn) : fn()
				}, null);
			}

			chain_promises(fnlist)
				.then(function() {
					console.log('*************** series done');
					var finish = true;
					done && (finish = done(err));
					if(finish !== false) test.finished();
				},
				function(){
					console.log('TODO promise series fail');
				})
		}
	}

	$.extend(Test.prototype, {
		init: function(name) {
			var self = this;

			this.start_button = buttons[name] = self.el.p().a(name).attr('href', '#start')
				.click(function(){
					$(this).hide();
					for(var b in buttons) buttons[b].hide();

					if(self.running) return self.stop();

					self.start.call(self);
					self.stop_button.show();
					return false;
				});

			this.stop_button = this.el.p().a('stop ' + name).attr('href', '#stop').hide()
				.click(function(){
					$(this).hide();
					self.start_button.show();
					for(var b in buttons) buttons[b].show();
					self.stop();
					return false;
				});
		},

		finished: function(msg){
			var self = this;

			self._clear_timeout();

			self.running && self.status.text('finished. ' + (msg || ''));

			self.stop_button.hide();
			self.start_button.show();
			for(var b in buttons) buttons[b].show();

			self.stop();
		},

		assert: function(a, msg){
			if(typeof a === 'string' || (typeof a === 'object' && a.jquery)){
				var obj = (typeof a === 'string') ? $(a + ':visible') : a;
				if(obj.length !== 1){
					this.fail('fail. ' + (msg || 'assertion'));
					throw new Error((msg || 'assertion'));
				}
			} else if(a !== true){
				this.fail('fail. ' + (msg || 'assertion'));
				throw new Error((msg || 'assertion'));
			}
			return a;
		},

		assert_fn: function(fn, msg){
			return function(){
				return $.Deferred(function(df){
					var a = fn();
					if(a !== true){
						this.fail('fail. ' + (msg || 'assertion'));
						throw new Error((msg || 'assertion'));
					}
					df.resolve();
				}).promise();
			}
		},

		assert_no_element: function(sel, msg) {
			if($(sel).length !== 0){
				this.fail('fail. ' + (msg || 'selector: ' + sel));
				throw new Error((msg || 'selector: ' + sel));
			}
			return this;
		},

		not_element: function(sel) {
			return function(){ return $(sel).length === 0 };
		},

		set_timeout: function(length){
			var self = this;

			this._timeout_length = length;
			this._clear_timeout();
			this._timeout = setTimeout(function(){
				self._timeout = 0;
				var msg = (Test.timeout_info ? (typeof Test.timeout_info === 'function' ? Test.timeout_info() : Test.timeout_info) : '(' + self._timeout_length + ')');
				if(Array.isArray(msg)) msg.shift('Test.timeout'); else msg = 'Test.timeout ' + msg;
				self.fail(msg);
			}, length);

			return this;
		},

		_clear_timeout: function(){
			this._timeout = clearTimeout(this._timeout);
		},

		start: function(msg){
			this.running = true;
			this.set_timeout(12000)._start();
			return this;
		},

		stop: function(arg){
			if(this.running){
				this._clear_timeout();
				this._stop(arg);
				this.running = false;
			}
			return this;
		},

		fail: function(msg){
			console.warn(Array.isArray(msg) ? msg.join('\n') : msg);
			this.el.p('.error fail').text(Array.isArray(msg) ? msg.join('<br/>') : msg);
			this.stop({error: this.name + ': ' + msg});
		},

		_listen_once: function(message_name, fn, msg){
if(!message_name) console.warn('Test._listen_once._timeout_info: no message_name!');
			var self = this;
			this._received || (this._received = {});

			if(msg) self._set_timeout_info(msg);

			this._received[message_name] = false;
			self.bus.clear(message_name).listen_once(message_name, function(event, msg){
				self._received[message_name] = true;
				self._set_timeout_info(_timeout_info());
				fn && fn(event, msg);
			}, self);

			function _timeout_info(){
				var a = $.map(self._received, function(a, b){ if(a === false) return b; }).join(', ');
				return a;
			}

			return this;
		},

		_set_timeout_info: function(info){
			Test.timeout_info = info
				? (typeof info === 'function'
					? info
					: ('waiting for: ' + info + ' (' + this._timeout_length + ')'))
				: '';
		},

		_wait: function(done){
			return this._wait_for_listeners(done);
		},

		_wait_for_listeners: function(done){
			var self = this;
			return this.wait_for(function(){
				var complete = true;
				for(var i in self._received){
					//console.log('_wait_for_listeners: ', i, self._received[i]);
					if(self._received[i] !== true){
						complete = false;
						break;
					}
				}
				if(complete){
					//console.log('_wait_for_listeners: complete.', self._received);
					done && done();
				}
				return complete;
			})
		},

		// ** deprecated - use wait_for2 instead which is supposed to handle more than one simultaneous wait
		// note that arg1 can be a function (return true when ready)
		wait_for: function(selector, msg){
			var self = this;
			this._set_timeout_info(msg || '');
			return $.Deferred(function(df){
				var check = $.isFunction(selector)
					? selector
					: function(){
						var a = $(selector);
						return (a.length > 0);
					}
				var interval = setInterval(function(){
					if(!self.running) clearInterval(interval);

					if(check()){
						clearInterval(interval);
						df.resolve();
					}
				}, 250);
			}).promise();
		},

		// TODO work-in-progress
		wait_for2: function(selector, msg){
			var self = this;

			return $.Deferred(function(df){
				function stop(){
					clearInterval(interval);
					clearTimeout(t);
				}

				var t = setTimeout(function(){
					stop();
					var m = $.isFunction(msg) ? msg() : msg;
					self.fail('timeout: ' + m);
					df.reject(m);
				}, this._timeout_length || 20000);

				var check = $.isFunction(selector)
					? selector
					: function(){
						var a = $(selector);
						return (a.length > 0);
					}
				var interval = setInterval(function(){
					if(!self.running) stop();

					if(check()){
						stop();
						df.resolve();
					}
				}, 250);
			}).promise();
		},

		/*
		 *   Use this where you need a fn that returns a promise
		 */
		wait_for3: function(selector, msg){
			return this.wait_for2(selector, msg)
		},

		wait_for_remove: function(css, msg){
			this._set_timeout_info(msg || '');
			return Test.wait_for(function(){ return ! $(css + ':visible').length });
		},

		bind: function(fn, args){
			var self = this;
			return function(){ fn.apply(self, arguments); }
		}
	});

	Test.wait_for = function(selector){
		return $.Deferred(function(df){
			var check = $.isFunction(selector)
				? selector
				: function(){
					var a = $(selector);
					return (a.length > 0);
				}
			var interval = setInterval(function(){
				if(check()){
					clearInterval(interval);
					df.resolve();
				}
			}, 500);
		}).promise();
	}

	Test.sleep = function(){
		return $.Deferred(function(df){
			setTimeout(function(){
				df.resolve();
			}, 2000);
		});
	}

/*
 *  List of Test to be run in series.
 */
function Tests(list){
	this.tests = list;
}

$.extend(Tests.prototype, {
	run: function(){
		var self = this;
		setTimeout(function(){
			self.log({message: 'Running...'});

			$.each(self.dont_run, function(i, test){ self.tests.splice(self.tests.indexOf(self.byChildProp('name', test)), 1); });

			async.series($.map(self.tests, function(test, i){ return function(done){
					i && self.log({message: 'Running Test ' + (i + 1) + ' ' + test.name + '...'});
					Test.timeout_info = null;
					new Test(test.name, $(), test.start, done).start();
				}}),
				function(a){
					console.log('tests finished');
					if(a && a.error) self.log({message: a.error, kind: 'service-alert'});
					else self.log({message: 'All tests completed', kind: 'service-allclear'});
			});
		}, 0);
	},

	dont_run: function(items){
		this.dont_run = items;
		return this;
	},

	log: function(o){
		console.log('%c' + o.message, 'background-color:#fa6');
		$.notify({
			id: 1,
			story: 'test',
			title: 'Test',
			message: o.message,
			kind: (o.kind || 'service-news')
		});
	},

	byChildProp: function(prop, val){
		// do an array lookup for a member object with the named property with value val.
		var a;
		$.each(this.tests, function(i, b){
			if(b[prop] === val) a = b;
		});
		return a;
	}
});


ECMWF.product_fixture = {
	product: '2t_full',
	viewport: {top: 0, left: 0, width: 100, height: 100},
	use_product_projection: true,
	availability: true
};


Test.make_list = function(fn, n){
	var list = [];
	for(var i = 0; i < n; i++) list.push(fn);
	return list;
}

return Tests;

});
