/*
 * incoming signals:
 *    "chart-product-availability" { list: [ { base_time: Date, valid_time: Date, step: number }, { .... } ]}
 *    "chart-availability-invalidated"
 *    "chart-time-select"
 *    "animation-stop"
 *
 * outgoing signals:
 *    "slider-time-change" :  { time: { base_time : Date, valid_time: Date, step: number }}
 *    "slider-time-slide" - needed for advanced features, but not currently implemented here.
 *    "slider-flipped"
 *    via tools:
 *       "tool-animation-toggle"
 *       "tool-next-step"
 *       "tool-prev-step"
 */

import ECMWF from 'config'
import $ from 'jquery'
import Bus from 'bus'
import Time from 'time'
import dateFormat from 'date-format'
import Container from 'container'
import './canvas_slider'
import 'pulsate'
import 'tool-buttons/prev_step'
import 'tool-buttons/next_step'
import 'tool-buttons/animation_toggle'
import 'date-select'
import 'icon'
import './time_navigator.css'
import 'background_mask'
import 'tooltip'

var primary_axis_name   = 'valid_time';
var secondary_axis_name = 'base_time';

dateFormat.masks.ecmwfFull = 'dd mmmm yyyy, HH"Z"';

$.widget('ui.timeNavigator', {
	options: {
		messageBus : Bus(),
		show_bt    : true,
		sliderType : 'canvas',
		ready      : $.noop
	},

	_init: function() {
		var self = this, o = this.options;

		this.dates = null;
		this.current_idx = null;
		this.pos = {left: 2 * 25 + (o.show_bt ? 183 : 0) + 5, right: 32};

		if(!this.element.find('.time-nav').length){

			this.wrapper = this.element.div('.time-nav');

			this.b2 = this.wrapper.div('.float-left toolbar-container')
				.toolButton({
					tooltips: true,
					tools : [[
						['prev-step', function(){ delta_step(-1); }],
						['next-step', function(){ delta_step(+1); }]
				]]});

			this.secondary_label = (o.show_bt ? this.wrapper : $())
				.div('.label-align float-left')
					.div({className: 'v-center label', style: 'width:1.7em; padding-left:5px;'}).html('BT:')
						.modeTooltip({
							style: 'tooltip-elaborate',
							content: function(){
								if(self.secondary && self.secondary.dates.length){
									var d = self.secondary.dates[self.secondary.axis.index].base_time;
									var content = tooltip_text.base_time;
									return $.extend({}, content, {text: content.text + ': ' + d.format('ecmwfFull')});
								}
							}
						});

			(this.selector = o.show_bt ? this.wrapper.div('.float-left') : $())
				.dateSelect({
					mode: 'enhanced',
					position: {my: 'left bottom', at: 'left top'},
					select: function(date){
						self._send_time_change_from_date(date, self.secondary.axis_name);
					},
					on_open: function(){
						self.element.find('.mode-tooltip').modeTooltip('disable');
					},
					on_close: function(){
						self.element.find('.mode-tooltip').modeTooltip('enable');
					}
				})
				.modeTooltip({
					content: function(){ return tooltip_text.selector[self.is_latest ? 'latest' : 'not_latest'];}
				});

			var label_alignment2 = this.wrapper.div('.label-align float-left');
			this.primary_label = label_alignment2.div({'class': 'v-center label', style: 'width:1.5em; padding-left:10px;'})
				.text('VT:')
				.modeTooltip({
					content: function(){
						if(!self.dates) return;

						var d = self.dates[self.current_idx];
						var content = tooltip_text.valid_time;
						return $.extend({}, content, {text: content.text + ': T+' + d.step + ' ' + d.valid_time.format('ecmwfFull')});
					}
				});
			self.pos.left += 30;

			this.slider = this.wrapper.div('.canvas-wrapper')
				.canvasSlider({
					height:                 24,
					big_handle:             ECMWF.request.kiosk,
					on_select:              self._bind(self._send_time_change_from_date),
					range_selection: {
						enabled:             $.chartGetPlugins ? !!$.chartGetPlugins()['animationRangeController'] : true,
						on_select:           self._bind(self._send_animation_selection_change)
					}
				})
				.canvasSlider('instance');

			this.animation_toggle_button = $('<div class="animation-toggle toolbar-container float-right">')
				.toolButton({tools: [['animation-toggle']], align: 'right', messageBus: o.messageBus, tooltips: true})
				.appendTo(self.wrapper);

			this.slider.element.css(self.pos);

			this.wrapper.container({
				layout: function() {
					return {
						layout: function(size) {
							// we cannot use the passed-in size because in some cases the container parent is the window.
							size = {width: self.element.width()};
							self.slider.element.triggerHandler('container', {width: size && (size.width - self.pos.left - self.pos.right) || self.slider.element.width()});
						}
					}
				}
			});

			self._disable(); //widget cannot be used until the Availability is present.

			o.messageBus
				.listen(self, 'chart-time-select', on_map_step_change)

				.listen(self, 'chart-product-availability', function(event, msg){
					self._update(msg);
				})

				.listen_once(self, 'chart-product-availability', function(event, msg){
					self._update_first(msg);
				})

				.listen(self, 'chart-availability-invalidated', function(event, msg){
					console.log('TimeNavigator.on_chart_availability_invalidated');
					self._disable(msg);
				})

				.listen(self, 'chart-frame-loaded', on_frame_loaded)
				.listen(self, 'chart-animation-load-complete', on_load_complete)
				.listen(self, 'chart-animation-selection-change', on_animation_selection_change);

			Bus()
				.clear('animation-start')
				.listen(self, 'animation-start', on_animation_start)
				.listen(self, 'animation-stop', on_animation_stop);
		}

		function delta_step(delta){
			var old_index = self.current_idx;

			self.current_idx += delta;
			if(self.current_idx < 0)                     self.current_idx = 0;
			if(self.current_idx > self.dates.length - 1) self.current_idx = self.dates.length - 1;

			if(self.current_idx != old_index){
				o.messageBus.send('slider-time-change', self.dates[self.current_idx]);
			}
		}

		function on_map_step_change(event, time){
			self.slider.setVal(time[self.primary_axis_name]);
			self.current_idx = time.index(self.dates);
			self.update_button_states();
		}

		function __on_animation_play_state_changed(playing){
			setTimeout(function(){
				self.animation_toggle_button.toolbar('set_state', 'animation-toggle', playing ? 'on' : 'off');
			}, 100);
		}

		function on_animation_start(event, msg){
			//note: the animator supplies a framelist for one or more charts, but the slider only shows the dates for a single chart.

			var loaded = {};
			o.messageBus.send('collect-animation-loaded-state', loaded);

			self.slider.animating(true, loaded);
			if($._false) self.slider.option('trail', true);

			__on_animation_play_state_changed(true);
		}

		function on_animation_stop(event, msg){
			self.slider.animating(false);
			__on_animation_play_state_changed(false);

			if($._false) self.slider.option('trail', false);
		}

		function on_frame_loaded(event, msg){
			self.slider.frameLoaded(msg);
		}

		function on_load_complete(event, msg){
			self.slider.animationLoadComplete(msg);
		}

		function on_animation_selection_change(event, msg){
			self.slider.setAnimationSelection({start: (msg.start && msg.start.valid_time), end: (msg.end && msg.end.valid_time)});
		}
	},

	_destroy: function(){
		this.options.messageBus.forget_me(this);
		Bus().forget_me(this);
		this.wrapper.empty();
	},

	_update: function(msg){
		var self = this, o = self.options;

		function validate(msg){
			var axis_values = msg[primary_axis_name].values
			if(!(axis_values[0] instanceof Time)) return false;
			var refp;
			for(var i in axis_values){
				var date = axis_values[i].valid_time;
				if(refp && ((date - refp) < -0)){
					setTimeout(function(){$.logError('TimeNavigator.verify_dates: valid_time (' + i + ') out of range! ' + date.format('mmmdd HH:MM') + ' should be later than ' + (refp && refp.format('mmmdd HH:MM')))}, 1000);
					return false;
				}
				refp = date;
			}

			refp = null;
			axis_values = msg[secondary_axis_name].values;
			//axis_values.length || console.warn('TimeNavigator.verify_dates: no secondary axis dates');
			for(i in axis_values){
				date = axis_values[i].base_time;
				if(refp && ((date - refp) < -0)){
					(function(){
						setTimeout(function(){$.logWarning('TimeNavigator.verify_dates: ' + secondary_axis_name +' (' + i + ') out of range! ' + date.format('mmmdd HH:MM') + ' should be later than ' + refp.format('mmmdd HH:MM'))}, 1000);
					})();
					return false;
				}
				refp = date;
			}
			return true;
		}

		if(!validate(msg)) return;

		this.last_msg = msg;

		this.primary_axis_name   = this.flipped ? secondary_axis_name : primary_axis_name;
		var _secondary_axis_name = this.flipped ? primary_axis_name   : secondary_axis_name;

		this.dates       = msg[this.primary_axis_name].values;
		this.current_idx = msg[this.primary_axis_name].index;
		this.ref_date    = msg[this.primary_axis_name].values[0].base_time;

		this.primary = {
			axis       : msg[this.primary_axis_name]
		}

		this.secondary = {
			axis_name  : _secondary_axis_name,
			axis       : msg[_secondary_axis_name],
			dates      : msg[_secondary_axis_name].values,
			index      : msg[_secondary_axis_name].index
		}

		this._set_selectlist();
		(this.secondary.dates.length && o.show_bt) || (this.selector.hide(), this.secondary_label.hide());

		if(self.dates.length > 1)
			this._enable();
		else
			this._disable();

		self.slider.setDates({
			dates: (self.dates)
				? $.map(self.dates, function(a) { return a[self.primary_axis_name]})
				: null,
			index: self.current_idx,
			ref: self.ref_date,
			format: msg.format
		});
	},

	_update_first: function(msg){
		var self = this, o = self.options;
		o.ready();

		Bus().send('event-start', ['timeNavigator.ready']);
	},

	_set_selectlist: function(){
		var self = this;

		var primary_dates = $.map(this.primary.axis.values, function(a) { return a[self.primary_axis_name] });
		var secondary_dates = $.map(this.secondary.dates, function(a) { return a[self.secondary.axis_name] });
		var not_latest = (
			this.flipped ?
				(this.primary.axis.index != primary_dates.length - 1)
				: (this.secondary.index != secondary_dates.length - 1)
		);
		this.selector.dateSelect('setDates', secondary_dates, secondary_dates[this.secondary.index], (false && not_latest && !this.flipped));

		self.selector.find('.button div').pulsate(not_latest);

		/*
		if(not_latest){
			//this.icon.icon2('setIcon', 'ui-icon ui-icon-alert').show();
			//this.flipped ? this.primary_label.addClass('label-warn') : this.secondary_label.addClass('label-warn');
			//this.flipped ? this.secondary_label.removeClass('label-warn') : this.primary_label.removeClass('label-warn');
		}else{
			//this.icon.icon2('setIcon', 'ui-icon ui-icon-circle-check').show();
			//this.secondary_label.removeClass('label-warn');
			//this.primary_label.removeClass('label-warn');
		}
		*/
		this.is_latest = !not_latest;
	},

	_enable: function(){
		this.update_button_states();
		this.animation_toggle_button.toolbar('enable', 'animation-toggle');
		this.secondary_label.css('opacity', 1);
		this.primary_label.css('opacity', 1);

		this.selector.dateSelect('enable').modeTooltip('enable');
		this.slider.enable();
	},

	_disable: function(){
		this.b2.toolbar('disable', 'prev-step');
		this.b2.toolbar('disable', 'next-step');
		this.animation_toggle_button.toolbar('disable', 'animation-toggle');
		this.secondary_label.css('opacity', 0.4);
		this.primary_label.css('opacity', 0.4);

		this.selector.dateSelect('disable', true).modeTooltip('disable');
		this.slider.disable();

		this.selector.find('.button div').pulsate(false);
	},

	update_button_states: function(){
		var self = this;

		if(self.primary_axis_name){
			var length = self.last_msg[self.primary_axis_name].values.length;
			self.b2.toolbar((self.current_idx === 0) ? 'disable' : 'enable', 'prev-step');
			self.b2.toolbar((self.current_idx === length - 1) ? 'disable' : 'enable', 'next-step');
		}
	},

	_send_time_change_from_date: function(date, axis_name){
		var t = this._lookup_date(date, axis_name);
		if(t) this.options.messageBus.send('slider-time-change', t);
		else console.warn('TimeNavigator._send_time_change_from_date: lookup failed. date=', date, ' axis=' + axis_name);
	},

	_send_animation_selection_change: function(selection){
		//note: if the selection has been cleared, both t1 and t2 will be empty.
		var t1 = this._lookup_date(selection.start);
		var t2 = this._lookup_date(selection.end);
		this.options.messageBus.send('animation-set', {range: {start: t1, end: t2}});
	},

	_lookup_date: function(time, axis_name){
		if(!time) return;
		if(!axis_name) axis_name = this.primary_axis_name;

		var dates = this.last_msg[axis_name].values;

		for(var i in dates){
			var item = dates[i][axis_name];
			if(item.getTime() === time.getTime()){
				return dates[i];
			}
		}
	},

	_bind: function(fn){
		var self = this;
		return function(){
			fn.apply(self, arguments);
		};
	}
});

var tooltip_text = {
	selector: {
		latest:     {text: 'Use this menu to change the Base Time.', mode: 'message' },
		not_latest: {text: '<p>' + 'You do not currently have the latest Base Time selected.' + '</p><p>' + 'Use this menu to change the Base Time.' + '</p>', mode: 'warning'}
	},
	icon: {
		latest:     {text: 'Using latest Base Time', mode: 'message'},
		not_latest: {text: '<b>Warning</b> A more recent forecast is available.', mode: 'warning' }
	},
	base_time:     {text: 'Base Time',  mode: 'elaborate'},
	valid_time:    {text: 'Valid Time', mode: 'elaborate'}
}

