/*jslint white: false, nomen: false, onevar: false, forin: false, browser: true, devel: true */

/**
 *  ui.toolbar is a specialised container that holds and handles actions on a list of tools.
 *  - left and right packing are supported.
 *  - the toolbar itself draws the buttons, calls the callbacks and emits the signals for the tools.
 *  - radio-groups are supported.
 *  - toggle buttons are supported (enable by setting toggleable to true in the tool).
 */

/*
 *  dependencies:
 *  note that contrary to what one might expect, tools depend on the
 *  toolbar and not the other way round due to tool self registration.
 */

var $ = require('jquery')
var Bus = require('bus')
require('./toolbar.css')
require('tooltip')

var makers = {};

function make_tool(toolbar, name, on_click) {
    return makers[name] ? makers[name](toolbar, on_click) : null;
}

$.widget('ui.toolbar', {

    options: {
        messageBus  : Bus(),
        tools       : [],
        align       : 'left',
        small       : false,
        tooltips    : false,
        after       : null // build the toolbar after the given dom element instead of as a child of the context.
    },

    _init: function() {
        var self = this;
        var o = this.options;

        var reversed  = (o.align === 'right');
        var align     = (o.align === 'right') ? 'right' : 'left';

        self.tools   = {};
        self.buttons = {};

        if(reversed){
            for(var i in o.tools){
                o.tools[i].reverse();
            }
            o.tools.reverse();
        }

        o.messageBus.listen('tool-enable', function(_, name) {
            console.log('NAM' + name);
            self.enable(name);
        });

        o.messageBus.listen('tool-disable', function(_, name) {
            console.log('NAM' + name);
            self.disable(name);
        });

        o.messageBus.listen('tool-select', function(e, name) {
            if(self.buttons[name]){
                self.buttons[name].click();
            }
        });

        var div = $('<div>').addClass(o.small ? 'toolbar-small' : 'ec-toolbar');

        for(i in o.tools){
            var group = o.tools[i];
            var is_last = (i == o.tools.length - 1);
            var ul = div.ul('.' + align + '-aligned' + (is_last ? ' last' : ''));
            var selection_done = false;
            for(var g in group){
                //each tool is either specified as a string, or an array of string and callback.
                var tool_name = (typeof group[g] == 'string') ? group[g] : group[g][0];
                var callback = (typeof group[g] == 'string') ? undefined : group[g][1];

                var tool = make_tool(self, tool_name, callback);
                if(!tool) continue;

                self.tools[tool_name] = tool;

                var css_class = tool.name();
                var li     = $('<li>').addClass('sprite-outer float-' + align).appendTo(ul);
                var button = $('<div>').addClass('sprite ' + css_class + ' button float-' + align).appendTo(li);
                if(o.small) button.addClass('ui-icon ui-icon-circle-' + css_class);

                self.buttons[tool_name] = button;

                (function(t) {
                    var css_class = t.name();
                    var b = button
                        .click(function(event) {
                            if(!t.disabled){
                                t.selectable && (t.selected = !t.selected);
                                on_click.apply(this, [event, t]);
                                t.selectable && (t.selected ? b.addClass('selected').parent().addClass('selected') : b.removeClass('selected').parent().removeClass('selected'));
                            }

                            if(o.tooltips && self.element.modeTooltip){
                                button.data('ui-modeTooltip') && button.modeTooltip('close');
                            }
                            return false;
                        })
                        .mouseover(function(event) {
                            if(!self.tools[css_class].disabled){
                                b.addClass('hover');
                            }
                        })
                        .mouseout(function(event) {
                            b.removeClass('hover');
                        })
                        .mousemove(function(event) {
                            return false;
                        });

                    var l = li
                        .mouseover(function(event) {
                            l.addClass('over');
                        })
                        .mouseout(function(event) {
                            l.removeClass('active over');
                        })
                        .mousedown(function(event) {
                            l.addClass('active');
                            return false;
                        })
                        .mouseup(function(event) {
                            l.removeClass('active');
                            return false;
                        });

                    if(tool.disabled)
                        self.disable(tool.name());

                    if(tool.selectable){
                        if(tool.settings){
                            //if the individual tool has a user_setting, its state is not grouped.
                            /*jsl:pass*/
                        }else if(reversed || !selection_done){
                            setTimeout(function(){             //give the parent time to bind the signals.
                                on_click.apply(b, [null, t]);  //set the default selection for the group.
                            }, 100);
                            selection_done = true;
                        }
                    }

                    if(o.tooltips && self.element.modeTooltip){
                        button.modeTooltip({
                            content: function(){ return t.disabled ? null : {text: t.tool_tip(), mode: 'message'}}
                        });
                    }
                })(tool);
            }
        }
        div.clear();

        if(o.after)
            this.element.find(o.after).after(div);
        else
            this.element.append(div);

        function on_click(event, tool){
            tool.click.apply(this, [event, o.messageBus]);
            if(tool.selectable) {
                reset_buttongroup(this);
                $(this)
                    .addClass('selected ' + css_class + '-selected')
                    .parent().addClass('selected');
            }
            if(tool.toggleable) {
                $(this)
                    .parent().toggleClass('on');
            }
        }

        function reset_buttongroup(sibling){
            $(sibling).parent().parent()
                .find('.sprite').removeClass('selected').end()
                .find('.sprite-outer').removeClass('selected');
        }
    },

    enable : function(tool_name) {
        if(this.buttons[tool_name])
            this.buttons[tool_name].parent().removeClass('ui-state-disabled');
        if(this.tools[tool_name])
            this.tools[tool_name].disabled = false;
    },

    disable : function(tool_name) {
        if(this.buttons[tool_name])
            this.buttons[tool_name].parent().addClass('ui-state-disabled');
        if(this.tools[tool_name])
            this.tools[tool_name].disabled = true;
    },

    set_state : function(tool_name, state) {
        if(state !== 'on' && state !== 'off'){ console.warn('bad arg: state: ' + state); return; }

        if(state === 'on'){
            this.buttons[tool_name].parent().addClass('on');
        }else{
            this.buttons[tool_name].parent().removeClass('on');
        }
    },

    destroy : function() {
        var self = this;
        for(var i in self.tools) {
            self.tools[i].destroy();
        }
        self.tools   = undefined;
        self.buttons = undefined;
    }
});

module.exports = {
	register: function(name, maker) {
		makers[name] = maker;
	}
}
