/*
 *  ECWMF Products
 *  --------------
 *
 *  A generic collection object for chart products.
 *
 *  Please note that this conforms to the Product Collection Interface (see documentation). Do not modify the API.
 *
 */
define(['config', 'jquery', 'ajax_queue', 'storage', 'bus', 'filterable-list', 'collection', 'mixins', 'utils', 'throttle-debounce'], function(ECMWF, $, Ajaxqueue, localStorage, Bus, List, Collection, Mixins){

	var settings = {
		score: ECMWF.options.product_scoring || false,   // reject misbehaving products
		cache_expire_mins: ECMWF.options.thumbnail_cache || 10
	}

	var mixins = {
		validate: function(tests) {
			for(var t in tests) if(!tests[t][1]()){
				var message = tests[t][0];
				console.warn('validate: "' + message + '" failed.');
				this.error_msg = "test '" + message + "' failed";
				return false;
			}
			return true;
		}
	}

	// TODO merge with collection/item.js
	var ChartListItem = function(item) {
		if(!item) return console.warn('ChartListItem: no data');

		this.data = item;

		// tags are forced to be arrays for efficient filtering.
		for(var tags in item.tags){
			var t = item[tags];

			if(t) this.data.tags[tags] = [t]; // for 'title'. can probably be remove later.
			if(typeof item.tags[tags] === 'string') this.data.tags[tags] = [item.tags[tags]];
		}

		this.thumbnail = function(args, done) {
			done = done || $.noop;

			var tn = this.cache.thumbnail.call(this);
			if(tn){
				return done(tn);
			}

			Thumbnail_queue.add(this, args, done);
		}

		/*
		 *   Temporary method TODO integrate with ChartListItem.thumbnail
		 */
		this.large_thumbnail = function(args, done) {
			var viewport = {width: 480}
			var json = this.has_widget('interactive')
				? {product: this.name(), viewport: $.extend(viewport, {height: 270})}
				: {product: {name: this.name()}, viewport: $.extend(viewport, {thumbnail: true})}
			Ajaxqueue($.extend({
				url        : '/plots/get-map/',
				json       : {product: json},
				tag        : 'grid-large-' + this.name(),
				priority   : products.thumbnail_priority,
				logErrors  : 'Get thumbnail (' +  this.title_() + ')',
				complete: function(result){
					done(result)
				}
			}, args));
		}

		this.cache = {
			thumbnail: function(val) {
				return this.cache._prop('thumbnail', val);
			},

			score: function(val) {
				return settings.score
					? (this.cache._prop('score', val) || 0) // TODO is the zero needed?
					: 0;
			},

			_prop: function(name, val) {

				this._item = this._item || products._cache(this.product.name()) || this._empty();
				if(val){
					this._item[name] = val;
					products.store(this.product.name(), this._item);
				}else{
					products.expired(this._item) && this._clear();
				}

				if(this._item[name]){
					return this._item[name];
				}
			},

			_empty: function() {
				return this._item = {timestamp: new Date().getTime()};
			},

			_clear: function() {
				this._empty();
				products.store(this.product.name(), this._item);
			},

			_item: null,
			product: this
		}
	}

	ChartListItem.prototype = $.extend({
		name: function() {
			return this.data.name;
		},

		title_: function() {
			return this.data.title;
		},

		long_title: function() {
			return this.data.long_title || this.data.title;
		},

		keywords: function(order) {
			return '';
		},

		has_widget: function(widget) {
			return !!~$.inArray(widget, this.data.widgets)
		}
	}, mixins)

	// In the absence of generic ajax bundling, thumbnail requests are bundled together here
	Thumbnail_queue = {
		max_bundle: 1,
		items: [],
		in_progress: [],
		timer: null,
		add: function(item, args, done) {
			var self = this, q = Thumbnail_queue, lists;

			for(var ai in lists = [q.items, q.in_progress]){
				for(var i in lists[ai]){
					var l = lists[ai][i];
					if(item.name() === l.item.name()){
						return l.done.push(done);
					}
				}
			}

			q.items.push({item: item, args: args, done: [done]});
			q.timer = q.timer || setInterval(function() {
				if(q.items.length){
					do q._fetch(q.items.splice(0, q.max_bundle)); while(q.items.length >= q.max_bundle);
				}else{
					q.timer = clearInterval(q.timer);
				}
			}, 100);
		},
		_fetch: function(items) {
			var self = this, q = Thumbnail_queue;

			q.in_progress = q.in_progress.concat(items);

			var args = items[0].args; // args (eg, size) are assumed to be common to all requests
			var first = items[0].item;
			var names = $.map(items, function(v){ return v.item.name() });

			function _fail(error) {
				$.each(items, function(i, item){
					var product = item.item;
					settings.score && product.cache.score.call(product, product.cache.score.call(product) - 1);
					$.each(item.done, function(i, done){ done(null, null, error || 'failed'); });
				});

				if(false && __SERVER__) throw Error(error); // we now trap all ajax errors in the ajax_queue, so this may no longer be needed
			}

			Ajaxqueue($.extend(
				{
					url        : '/backend/product-thumbnail-inline/',
					json       : {name: names},
					tag        : 'grid-' + names.join(','),
					priority   : products.thumbnail_priority,
					logErrors  : 'Get thumbnail (' +  first.title_() + ')',
					complete: function(result){
						var indexes = [];
						for(var i in items){
							var item = items[i].item
							for(var j in q.in_progress){
								var l = q.in_progress[j];
								if(l.item.name() === item.name()){
									indexes.push(+j)
									break
								}
							}
						}
						indexes = indexes.sort(function(a, b){ return b - a });
						for(i in indexes){
							q.in_progress.splice(indexes[i], 1);
						}

						$.each(items, function(i, item) {
							var product = item.item;
							if(product.validate([
								['has result', function(){ return result; }],
								['has url', function(){ return result.thumbnails || result.url; }]
							])){
								product.cache.thumbnail.call(product, result.url)
								$.each(item.done, function(j, done){ done(result.thumbnails[i].url, result.thumbnails[i].img); });
							}
							else _fail('validation failed');
						});
					},
					error: _fail
				},
				args)
			);
		}
	}

	ECMWF.options.timeout && $.extend(Ajaxqueue.options, {
		timeout: ECMWF.options.timeout * 1000
	});

	/*
	 *   To reduce file size, prerendered thumbnails are not included in the thumbnail json data.
	 *   They are added to the json data here
	 */
	$(function() {
		// TODO this MUST run before the first grid render - check this is _always_ so
		if(window._ECMWF && _ECMWF.all_products){
			$('.column-2 .thumbnail img').each(function() {
				var img = $(this)
				var src = img.attr('src')
				if(src){
					var name = img.attr('name')
					if(__DEV__){
						var p = $.findWhere(_ECMWF.all_products, {name: name})
						p && console.warn('thumbnail already present in _all_products');
					}
					_ECMWF.all_products.push({name: name, thumbnail: src})
				}
			})
		}
	})

	function Products(){
		this.query = {}
		this.thumbnail_priority = 'low'

		this.list_init()

		this.products = function() {
			var self = this;
			// TODO change so that it does not return a new promise each time
			return $.Deferred(function(df) {

				function have_products(result) {
					if(__SERVER__){
						if(!result.query_count) throw Error('No products')
					}

					if(window._ECMWF && _ECMWF.all_products){
						$.each(_ECMWF.all_products, function(i, v){
							var p = $.findWhere(result.records, {name: v.name})
							if(p){
								p.thumbnail = v.thumbnail
							}
						})
					}

					__DEV__ && Bus().send('event-start', ['start verification']);
					self._products = $.map(result.records, function(item){
						if(self.validate([
								['not empty', function(){ return item; }],
								['has name', function(){ return item.name; }],
								['has title', function(){ return item.title; }],
								['has no empty facets', function(){
									if(item.tags){
										for(var tags in item.tags){
											if(Array.isArray(item.tags[tags]) && !item.tags[tags].length){
												console.warn('product.validate: product error: ' + item.name + ': tag is empty array (' + tags + ')');
												delete item.tags[tags];
											}
										}
									}
									return true;
								}],
								['path has no trailing slash', function(){ if(Array.isArray(item.path))
									for(var i in item.path){
										var p = item.path[i];
										if((item.path[i] = item.path[i].replace(/\/$/, '')) !== p)
										console.warn('path has trailing slash: ', p, item.name);
									}
									return true;
								}],
								['has keywords', function(){ return item.keywords || (item.keywords = ' '); }] // temporary. should be server-side.
							])){
							var product = new ChartListItem(item);
							if((!__DEV__) || (product.cache.score.apply(product) > -1)) return product;
						}
					})

					if(self._products.length){
						var rejected = result.query_count - self._products.length
						console.log('n_products=' + self._products.length + (rejected ? ' (' + (rejected) + ' rejected)' : ''))
					}else{
						console.log('no products found for:', self.query)
					}

					self.packages.then(function() {
						df.resolve(self._products)
					});
				}

				if(window._ECMWF && _ECMWF.products){
					var public_packages = ['s2']
					if(!!~$.inArray(self.query.package, public_packages)){
						if(_ECMWF.all_products){
							$.each(_ECMWF.all_products, function(i, v){
								var p = $.findWhere(_ECMWF.products, {name: v.name})
								if(p){
									p.thumbnail = v.thumbnail
								}
							})
						}
						return have_products({records: []});
					}
				}

				Ajaxqueue({
					url: '/plots/find-product/',
					json: {query: self.query},
					complete: have_products
				});
			}).promise()
		}

		this.packages = $.Deferred(function(df){
			Ajaxqueue({
				url: '/plots/get-packages/',
				priority: 'very-high',
				complete: function(result){
					if(result.packages){
						products._packages = result.packages;
						products.hierarchy.update(result.categories);
						df.resolve(products._packages);
					}
				}
			});
		}).promise()

		this.hierarchy = {
			products: this,
			tree: {},

			update: function(categories) {
				var self = this;
				/*
				this.tree = {};
				$.each(this.products.products, function(i, p){
					for(var k in p.data.path){
						var path = p.data.path[k].split('/');
						var cursor = self.tree;
						for(var j in path){
							if(path[j]){
								cursor = cursor[path[j]] = cursor[path[j]] || {};
							}
						}
					}
				});
				*/
				this.tree = categories;

				// fill in missing categories TODO move to server.
				for(var i in this.products._packages){
					var pkg = this.products._packages[i];
					if(!$.findWhere(this.tree, {name: pkg.name})){
						this.tree.push({name: pkg.name, title: pkg.title});
					}
				}
			},

			children: function(path) {
				var node = this._child(path, 0, this.tree);
				return node ? node.children : [];
			},

			lookup: function(path) {
				if(!path || path === 'all') return {title: 'ECMWF charts'}; // special case because the tree does not currently have a root item

				$.isEmptyObject(this.tree) && console.warn('hierarchy.lookup: no package list');
				if(!Array.isArray(path)) path = path.replace(/\/#.*/, '').split('/');
				return this._child(path, 0, this.tree) || $.findWhere(products._packages, {name: ECMWF.options.default_package})
			},

			_child: function(path, p, tree){
				for(var i in tree){
					var node = tree[i];
					if(node.name === path[p]){
						if(p === path.length - 1){
							return node;
						}else{
							p++;
							return this._child(path, p, node.children);
						}
					}
				}
			},

			path_is_valid: function(path) {
				return true;
			}
		}

		this._cache =
			localStorage ? function(name) {
				if(!this.__cache){
					this.__cache = JSON.parse(localStorage.products || '{}');
				}
				return name ? this.__cache[name] : this.__cache;
			}
			: function(){ return {}; }

		this._store = $.debounce(2000, function() {
			localStorage.products = JSON.stringify(this.__cache);
		})

		this.store = localStorage ? function(name, item) {
			this.__cache[name] = item;
			this._store();
		} : $.noop;

		setTimeout(function clean_cache() {
			$.each(products._cache(), function(i, item) {
				if(products.expired(item)){
					delete products._cache()[i]
					products._store();
				}
			});
			//console.log('products.clean_cache: size=', ECMWF.n_properties(products._cache()));
		}, 5000);
	}

	Products.prototype = $.extend(new Collection(), {
		kind: 'product',

		expired: function(item) {
			return item && ((!item.timestamp) || (new Date().getTime() - item.timestamp > 1000 * 60 * settings.cache_expire_mins));
		}
	})

	var products;
	return products = new Products();
});
