/*
 *  A wintersmith Content plugin for prerendering a product browser page.
 *
 *  - inherits from the Page class.
 *  - is triggered for pages of type .browser.yaml
 *
 *  config:
 *    backend_host is set in wintersmith config.js
 *
 *  TODO
 *    - thumbnails need to be optimised - use jpgs instead of png !!
 *    - inline javascript data:
 *      - remove unused properties
 *      - minimise (remove whitespace)
 *    - error handling
 *      - if the backend is not working properly, errors are not meaningfully reported. Perhaps do a preflight test to check the backend?
 *         -errors which are not trapped:
 *            -"validate: "has url" failed'
 *                 ---- an exception has been added. should abort now. TODO check
 *      - check for thumbnail errors:
 *           1) YES in-browser thumbnail loading does now cause script to fail
 *           2) YES get_cachable_data 1/2 DOES halt on timeout
 *           3) ??? get_cachable_data 2/2          ??
 *      - need to abort in case of js errors
 *
 *      - what should happen in the case of a persistent product failure?
 *
 *    - gtmetrix recommends moving our inline js containing the product information _after_ the js includes as it blocks loading of these files (blocks processing yes, but loading?)
 *
 */
(function() {
	var extend = function(child, parent) {
		for (var key in parent) {
			if (hasProp.call(parent, key)) child[key] = parent[key];
		}
		function ctor() { this.constructor = child; }
		ctor.prototype = parent.prototype;
		child.prototype = new ctor();
		child.__super__ = parent.prototype;
		return child;
	};
	var hasProp = {}.hasOwnProperty;

	const async = require('async')
	const fs = require('fs')
	const path = require('path')
	const url = require('url')
	const yaml = require('js-yaml')
	const request = require('sync-request')
	const jsdom = require('jsdom').jsdom
	const beautify = require('js-beautify').html

	process.on('unhandledRejection', function (err) {
		console.log(red('Page builder caught unhandledRejection error'));
		throw err;
	});

	process.on('uncaughtException', function(err) {
		console.log(red('Page builder got uncaught exception'));
		throw err;
	});

	module.exports = function(env, callback) {
		var view = function(env, locals, contents, templates, callback) {
			var self = this;
			var exclude = [
				//'forecasts/charts/s2s/index.html',
				//'forecasts/charts/mofc_multi/index.html',

				'forecasts/charts/all/index.html',
				'forecasts/charts/catalogue/index.html',
				'forecasts/charts/obstat/index.html',
				'forecasts/charts/long-range/index.html',
				'forecasts/charts/seasonal/index.html',
				'forecasts/charts/catalogue/w_t2m30mw.html',
				'forecasts/charts/catalogue/w_z500t850.html'
			]
			console.log(cyan('prerendering:'), this.filename, exclude.indexOf(this.filename) > -1 ? '[' + yellow('excluding') + ']' : '');

			var template = templates[path.normalize(this.template)];
			ctx = {page: this};
			env.utils.extend(ctx, locals);
			template.render(ctx, function(err, html){
				if(err) return console.warn('chart view:', err), callback(err, null);
				if(exclude.indexOf(this.filename) > -1) return callback(null, html);

				var output_file = env.config.output + this.filename
				fs.stat(output_file, (err, stats) => {
					if(!err){
						var mtime = stats.mtime
						var age = Math.round((new Date().getTime() - mtime.getTime()) / 1000);
						var age_str = Math.round(age/60) + ':' + ('0' + (age%60)).substring(0, 2) + ''
						const cache_time = 30 * 60 // 30 minutes

						if(age < cache_time){
							console.log(magenta('    using cached page ...') + ' age=' + age_str)
							// does the date get preserved? unexpectedly it apparently does
							fs.readFile(output_file, {encoding: 'utf-8'}, function(err, data){
								callback(null, data)
							})
							return;
						}
					}

					try {
						self.pre_render(html.toString(), function(e, html){
							callback(e, new Buffer(html || '', 'utf8'));
						});
					} catch(e){
						// because the prerender is asynchronous, its unlikely this will catch anything
						console.log('>>>>>>>>>>>>> caught exception from prerender NEVER GET HERE');
						callback(e, html);
					}
				})
			});
		}

		var ChartPage = (function(superClass) {
			extend(ChartPage, superClass);

			function ChartPage(filepath1, metadata1, markdown1) {
				console.log('      ChartPage: filepath=', filepath1.relative);
				var y = yaml.safeLoad(fs.readFileSync(filepath1.full, 'utf8'));;

				this.filepath = filepath1;
				this.metadata = {template: 'page.jade'};
				extend(this.metadata, metadata1);
				extend(this.metadata, y);
				this.markdown = 'Chart page';
			}

			/*
			 * At present jsdom does not handle navigation (such as setting window.location.href === "https://example.com/"). However, if you'd like to change the URL of an existing Window instance (such as for testing purposes), you can use the jsdom.changeURL method:
          *   jsdom.changeURL(window, "https://example.com/");
			 */

			ChartPage.prototype.pre_render = function(html, callback) {
				console.log('starting jsdom...');
				var self = this;

				html = html.replace(/<script.*\/script>/g, ''); // TODO why do scripts stop jsdom page loading?

				var vconsole = require('jsdom').createVirtualConsole()//.sendTo(console);
					.on('error', function() {
						console.log('page prerender caught jsdom error')
						console.log('\x1b[31m\033[1A')
						console.log.apply(console, arguments)
						console.log('\x1b[0m\033[1A')
						//throw Error(arguments[0])
					})
					.on('log', function() {
						console.log('\x1b[38;5;241m\033[1A')
						console.log.apply(console, arguments)
						console.log('\x1b[0m\033[1A')
					})
					.on('warn', function() {
						console.log('\x1b[33m\033[1A')
						console.log.apply(console, arguments)
						console.log('\x1b[0m\033[1A')
					})
					.on('info', function() {
						console.log.apply(console, arguments)
					});

				var window = jsdom(html, {
					features: {
						FetchExternalResources: ['script'], // this is the default
						ProcessExternalResources: ['script'],
						MutationEvents: '2.0'
					},
					//documentRoot: '/var/tmp/sdo/git/webdev/apps/frontend/static/frontend',
					url: 'http://' + env.config.locals.backend_host + '/' + this.filename.replace('index.html', ''),
					virtualConsole: vconsole,
					created: function(errors, window) {
						console.log('jsdom.created');
					},
					onload: function() {
						console.log('jsdom.onload');
					},
					done: function(errors, window) {
						console.log('jsdom.done');
					}
				}).defaultView;

				function on_error(e) {
					console.error(red('Error from jsdom: ', e.error.message));
					console.error('    ', e.error.stack);
					callback([e.error.message])
					window.close();
				}

				window.addEventListener('error', on_error)

				window.addEventListener('load',  function(){
					window.is_node = true;
					var document = window.document; // document is not available initially.
					//window.jQuery('script').load(() => { console.log('--> script loaded!', this.src); })

					var scriptEl = document.createElement('script');
					scriptEl.src = 'file://' + env.contentsPath + '/js/iso/product_browser.js';
					document.body.appendChild(scriptEl);

					function finished() {
						console.log('postprocessing...')

						if(false){
							var scripts = document.getElementsByTagName('script');
							if(scripts.length){
								console.log('scripts:');
								for(var i=0;i<scripts.length;i++) {
									if(scripts[i].src) console.log(i, scripts[i].src)
									else console.log(i, scripts[i].innerHTML)
								}
							}else{
								console.warn('no scripts');
							}
						}

						// these are all inline styles - it should be safe to delete them
						Array.prototype.slice.call(document.getElementsByTagName('style')).forEach((a, b) => {
							a.remove();
						});

						// remove in-progress jsonp calls and file:/// scripts
						// TODO is duplicated below - needs to be run twice?
						Array.prototype.slice.call(document.getElementsByTagName('script')).forEach((a) => {
							/*
							Array.prototype.slice.call(a.attributes).forEach((a) => {
								console.log('      - attr=' + a.nodeName + '=' + a.nodeValue);
							});
							console.log('     src=', a.attributes.src);
							*/
							if(a.attributes.src){
								if(!!~a.attributes.src.nodeValue.indexOf('callback=')){
									a.remove();
								}
								if(!!~a.attributes.src.nodeValue.indexOf('file://')){
									a.remove();
								}
							}
						});

						/*
						 *  Store product list and thumbnails
						 *
						 *  @param all  - thumbnails for all products (ignoring access control)
						 *  @param data - full product list (for public users?)
						 *
						 */
						window.get_cachable_data().then(function(all, data){
							if(all.length){
								// thumbnails for products which are not already rendered on the page
								// are added to the _ECMWF.all_products array.

								var filtered = []
								all.forEach((p) => {
									var el = document.querySelector('.thumbnail img[name="' + p.name + '"]');
									if(!el) filtered.push(p)
								})

								console.log('total products:', all.length);
								console.log('rendered products:', all.length - filtered.length);

								var script = document.createElement('script');
								script.innerHTML = '_ECMWF = window._ECMWF || {};\n_ECMWF.all_products = ' + JSON.stringify(filtered) + ';'
								document.head.appendChild(script);
							}
							else console.log('no _all_ products');

							function remove(selector) {
								var el = document.querySelectorAll(selector)[0];
								el && el.remove();
							}
							function remove_attr(selector, attr) {
								var el = document.querySelectorAll(selector)[0];
								el && el.removeAttribute(attr);
							}
							function remove_attr_(selector, attr) {
								document.querySelectorAll(selector).forEach((el) => {
									el.removeAttribute(attr);
								})
							}

							if(data.length){
								if(false){ // thumbnails are now retrieved separately
									var image_data = {}
									document.querySelectorAll('.thumbnail img').forEach((img, i) => {
										image_data[img.getAttribute('name')] = img.getAttribute('src')
										return img.getAttribute('name')
									});
									data.forEach((product) => {
										if(image_data[product.name]) product.thumbnail = image_data[product.name]
									})
								}

								var script = document.createElement('script');
								script.innerHTML = '_ECMWF = window._ECMWF || {};\n_ECMWF.products = ' + JSON.stringify(data) + ';'
								document.head.appendChild(script);

								/*
								// removing the src attribute causes the first render to be empty, so it is better to not remove the src.
								remove_attr_('.catalogue-grid .thumbnail img', 'src')
								*/

								remove_attr_('.catalogue-grid .thumbnail img', 'title') // keep file size down
								remove_attr_('.catalogue-grid .thumbnail img', 'alt')
							}else{
								console.log(red('failed to find cachable data'));
								return callback('failed to find cachable data');
							}

							remove('.catalogue-grid .ps-scrollbar-x-rail');
							remove('.catalogue-grid .ps-scrollbar-y-rail');
							remove('.thumb-progress-bar');

							document.querySelectorAll('script').forEach((el) => {
								var src = el.getAttribute('src');
								if(src){
									if(src.indexOf('googletagmanager') > 0 || src.indexOf('google-analytics') > 0){
										console.log(yellow('removing google analytics script...'));
										el.remove()
									}
									if(src.indexOf('callback=') > 0){
										el.remove()
									}
								}
							});

							// prevent the height of the grid being set to zero
							remove_attr('.catalogue-grid .scrollable', 'style');
							remove_attr('.catalogue-grid .scrollable', 'sizor');
							remove_attr('.catalogue-grid .scrollable', 'max-sizor');

							// change thumbnail layout to be based on floated elements instead of absolute position
							// (only required when not using the vgrid/simple layout)
							/*
							remove_attr_('.catalogue-grid .item', 'style');
							document.querySelectorAll('.catalogue-grid .grid-items').forEach((el) => {
								el.className += ' floated';
							})
							*/

							// set height of main-column-wrapper (because contents are position:absolute) - ideally we should measure content height
							/*
							document.querySelectorAll('.main-column-wrapper').forEach((el) => {
								el.style.height = '1000px';
							})
							*/

							var body_class = document.body.getAttribute('class');
							document.body.className = body_class.replace('not-logged-in', '');

							// we want to disable log messages now, but dont seem to be able
							/*
							var virtualConsole = require('jsdom').getVirtualConsole(window);
							console.log('virtualConsole=', virtualConsole);
							virtualConsole._events.error = function(){};
							virtualConsole._events.log = function(){};
							console.log('virtualConsole=', virtualConsole._events.Console);
							virtualConsole._events.Console = function dummy(){};
							console.log('virtualConsole=', virtualConsole._events.Console);
							*/

							var html = document.documentElement.outerHTML;
							window.removeEventListener('error', on_error);
							// put back scripts that were removed - TODO restore the ones which were removed instead of respecifying them here.
							html = '<!DOCTYPE html>\n' + html.replace(/<\/head>/, '<script src="/sites/default/files/js/ecmwf_widgets_options.js"></script><script src="/js/builds/common.js"></script>\n  <script src="/js/builds/browser-common.js"></script>\n  <script src="/js/builds/product_browser.js"></script>\n  </head>');
							callback(null, beautify(html, {indent_size: 2}));

							console.log('-----------------------------------------------');

							window.close();

						}).fail((reason) => {
							console.log(red('FAIL: ' + reason), 'type=', typeof reason);
							callback([reason])
						});
					}

					if(true){
						// wait for the thumbnails to load in the page
						var thumbnail_done_count = 0;
						var max_thumbnails = 64;
						new Promise((resolve, reject) => {
							var t1 = setTimeout(() => {
								console.log(self.filepath.relative + ': timeout (90s): n_thumbnails=' + thumbnail_done_count);
								t1 = clearTimeout(t1)
								t2 = clearInterval(t2)
								throw Error('Timeout waiting for in-page thumbnails')
							}, 90000); // this timeout needs to be quite long in case thumbnails are not cached server-side. TODO we must abort sooner if there are errors with any single thumbnail

							var t2 = setInterval(() => {
								var items = document.querySelectorAll('.column-2 .grid-items img')
								if(items.length){
									var el = items[thumbnail_done_count];
									var src = el && el.getAttribute('src')
									if(src && !~src.indexOf('/backend/get-thumbnail/product/')){
										if(src.substring(0, 5) === 'data:'){
											console.log('    thumbnail ' + thumbnail_done_count + ': src=', src.substring(0, 96))
										}else{
											console.log(orange('    warning') + ': thumbnail ' + thumbnail_done_count + ' not data image: src=', src, '(removing...)');
											el.removeAttribute('src'); // appears to work, but perhaps it will just get re-rendered?
										}
										thumbnail_done_count++;
										if(thumbnail_done_count >= Math.min(items.length, max_thumbnails)){
											console.log('page got all thumbnails (' + thumbnail_done_count + ')');
											t1 = clearTimeout(t1)
											t2 = clearInterval(t2)
											resolve();
										}
									}
								}
							}, 100);
						}).then(() => {
							finished();
						})
					}else{
						setTimeout(function(){ // TODO
							finished();
						}, 2000)
					}
				});
				/*
				window.addEventListener('DOMContentLoaded', function() {
					console.log('--> dom loaded!');
				});
				*/
			}

			/*
			 *  use a custom view that calls jsdom after the template is loaded
			 */
			ChartPage.prototype.getView = function() {
				return this.metadata.view || 'prerender';
			}

			ChartPage.prototype.getLocation = function(base) {
				var uri = this.getUrl(base);
				return uri.slice(0, +uri.lastIndexOf('/') + 1 || 9e9);
			}

			ChartPage.prototype.getHtml = function(base) {
				if (base == null) {
					base = env.config.baseUrl;
				}

				return this.markdown;
			}

			/*
			ChartPage.prototype.getUrl = function(base) {
				return 'forecasts/charts/medium/index.html';
			}
			*/
			ChartPage.prototype.getFilename = function() {
				filename = ChartPage.__super__.getFilename.call(this).replace(/.browser.html$/, '.html');
				return filename;
			}

			return ChartPage;

		})(env.plugins.Page);

		ChartPage.fromFile = function(filepath, callback) {
			return async.waterfall([
				function(callback) {
					return fs.readFile(filepath.full, callback);
				},
				function(buffer, callback) {
					return ChartPage.extractMetadata(buffer.toString(), callback);
				},
				(function(_this) {
					return function(result, callback) {
						var markdown = result.markdown;
						var metadata = result.metadata;
						var page = new _this(filepath, metadata, markdown);
						return callback(null, page);
					};
				})(this)
			], callback);
		};

		ChartPage.extractMetadata = function(content, callback) {
			var end, markdown, metadata = '', parseMetadata, result;

			function parseMetadata (source, callback) {
				var error, error1, lines, markerPad;
				if (!(source.length > 0)) {
					return callback(null, {});
				}
				try {
					return callback(null, yaml.load(source) || {});
				} catch (error1) {
					error = error1;
					if ((error.problem != null) && (error.problemMark != null)) {
						lines = error.problemMark.buffer.split('\n');
						markerPad = ((function() {
							var i, ref, results;
							results = [];
							for (i = 0, ref = error.problemMark.column; 0 <= ref ? i < ref : i > ref; 0 <= ref ? i++ : i--) {
								results.push(' ');
							}
							return results;
						})()).join('');
						error.message = "YAML: " + error.problem + "\n\n" + lines[error.problemMark.line] + "\n" + markerPad + "^\n";
					} else {
						error.message = "YAML Parsing error " + error.message;
					}
					return callback(error);
				}
			};

			markdown = content;
			if (content.slice(0, 3) === '---') {
				result = content.match(/^-{3,}\s([\s\S]*?)-{3,}(\s[\s\S]*|\s?)$/);
				if ((result != null ? result.length : void 0) === 3) {
					metadata = result[1];
					markdown = result[2];
				}
			} else if (content.slice(0, 12) === '```metadata\n') {
				end = content.indexOf('\n```\n');
				if (end !== -1) {
					metadata = content.substring(12, end);
					markdown = content.substring(end + 5);
				}
			}
			return async.parallel({
				metadata: function(callback) {
					return parseMetadata(metadata, callback);
				},
				markdown: function(callback) {
					return callback(null, markdown);
				}
			}, callback);
		};

		env.registerView('prerender', view);
		// ******************* temporary - keep them separate for debugging
		env.registerContentPlugin('browser', '**/*.browser.yaml', ChartPage);
		return callback();
	}

	function red(s) {
		return '\x1b[31m' + s + '\x1b[0m'
	}

	function blue(s) {
		return '\x1b[34m' + s + '\x1b[0m'
	}

	function cyan(s) {
		return '\x1b[36m' + s + '\x1b[0m'
	}

	function orange(s) {
		return '\x1b[33m' + s + '\x1b[0m'
	}

	function yellow(s) {
		return '\x1b[33;1m' + s + '\x1b[0m'
	}

	function magenta(s) {
		return '\x1b[35m' + s + '\x1b[0m'
	}

	function grey(s) {
		return '\x1b[37m' + s + '\x1b[0m'
	}
}).call(this);
